@@ -338,7 +338,7 @@ jump_protected(vm_t *vm, Bit16u cs_raw, Bit32u disp32)
         gate_cs_selector = ((gate_t *) &cache.desc)->selector;
         if ( IsNullSelector(gate_cs_selector) ) {
           monpanic(vm, "jump_pro: CS selector null\n");
-          exception(vm, ExceptionGP, 0x0000);
+          exception(vm, ExceptionGP, 0);
           }
 
         /* selector must be within its descriptor table limits
@@ -778,48 +778,46 @@ iret_protected(vm_t *vm)
   void
 call_protected(vm_t *vm, Bit16u cs_raw, Bit32u disp32)
 {
-  selector_t cs_selector;
-  descriptor_cache_t cs_cache;
+  descriptor_cache_t cache;
+  selector_t         selector;
   
+  selector.raw = cs_raw; 
   /* Opsize in effect for CALL is specified by the D bit for the
    * segment containing dest & by any opsize prefix.
    * For gate descriptor, deterermined by type of call gate:
    * 4=16bit, 12=32bit
    * count field: 16bit specifies #words, 32bit specifies #dwords
    */
-  cache_sreg(vm, SRegCS);
-  cache_sreg(vm, SRegSS);
 
   /* new cs selector must not be null, else #GP(0) */
-  if ( (cs_raw & 0xfffc) == 0 ) {
+  if ( IsNullSelector(selector)) {
     monpanic(vm, "call_pro: CS selector null\n");
     exception(vm, ExceptionGP, 0);
     }
 
-  cs_selector.raw = cs_raw;
 
   /* check new CS selector index within its descriptor limits,
    * else #GP(new CS selector) */
-  fetch_raw_descriptor(vm, cs_selector, &cs_cache.desc, ExceptionGP);
-  descriptor2cache(vm, &cs_cache);
+  fetch_raw_descriptor(vm, selector, &cache.desc, ExceptionGP);
+  descriptor2cache(vm, &cache);
 
   /* examine AR byte of selected descriptor for various legal values */
-  if (cs_cache.valid==0) {
+  if (cache.valid==0) {
     monpanic(vm, "call_pro: invalid CS descriptor\n");
     exception(vm, ExceptionGP, cs_raw & 0xfffc);
     }
 
-  if (cs_cache.desc.type & D_S) { /* normal segment */
+  if (cache.desc.type & D_S) { /* normal segment */
     Bit32u temp_ESP;
 
-    if ( !(cs_cache.desc.type & D_EXECUTE) ) {
+    if ( !(cache.desc.type & D_EXECUTE) ) {
       monpanic(vm, "call_pro: non executable segment\n");
       exception(vm, ExceptionGP, cs_raw & 0xfffc);
       }
 
-    if ( cs_cache.desc.type & D_CONFORM ) { /* Conforming code seg */
+    if ( cache.desc.type & D_CONFORM ) { /* Conforming code seg */
       /* DPL must be <= CPL, else #GP(code seg selector) */
-      if (cs_cache.desc.dpl > G_CPL(vm)) {
+      if (cache.desc.dpl > G_CPL(vm)) {
         monpanic(vm, "call_pro: cs.dpl > CPL\n");
         exception(vm, ExceptionGP, cs_raw & 0xfffc);
         }
@@ -827,15 +825,15 @@ call_protected(vm_t *vm, Bit16u cs_raw, Bit32u disp32)
     else { /* Non-conforming code segment */
       /* RPL must be <= CPL, else #GP(code seg selector) */
       /* DPL must be = CPL, else #GP(code seg selector) */
-      if ( (cs_selector.fields.rpl > G_CPL(vm)) ||
-           (cs_cache.desc.dpl != G_CPL(vm)) ) {
+      if ( (selector.fields.rpl > G_CPL(vm)) ||
+           (cache.desc.dpl != G_CPL(vm)) ) {
         monpanic(vm, "call_pro: cs.rpl > CPL\n");
         exception(vm, ExceptionGP, cs_raw & 0xfffc);
         }
       }
 
     /* segment must be present, else #NP(code seg selector) */
-    if (cs_cache.desc.p == 0) {
+    if (cache.desc.p == 0) {
       monprint(vm, "call_pro: cs.p = 0\n");
       exception(vm, ExceptionNP, cs_raw & 0xfffc);
       }
@@ -853,7 +851,7 @@ call_protected(vm_t *vm, Bit16u cs_raw, Bit32u disp32)
         }
 
       /* IP must be in code seg limit, else #GP(0) */
-      if (disp32 > cs_cache.limit_scaled) {
+      if (disp32 > cache.limit_scaled) {
         monpanic(vm, "call_pro: IP not in code seg limit\n");
         exception(vm, ExceptionGP, 0);
         }
@@ -869,7 +867,7 @@ call_protected(vm_t *vm, Bit16u cs_raw, Bit32u disp32)
         }
 
       /* IP must be in code seg limit, else #GP(0) */
-      if (disp32 > cs_cache.limit_scaled) {
+      if (disp32 > cache.limit_scaled) {
         monpanic(vm, "call_pro: IP not in code seg limit\n");
         exception(vm, ExceptionGP, 0);
         }
@@ -883,23 +881,24 @@ call_protected(vm_t *vm, Bit16u cs_raw, Bit32u disp32)
      * set RPL of CS to CPL
      * load eIP with new offset
      */
-    load_cs_pro(vm, cs_selector, &cs_cache, G_CPL(vm));
+    load_cs_pro(vm, selector, &cache, G_CPL(vm));
     G_EIP(vm) = disp32;
-    if (cs_cache.desc.d_b==0)
+    if (cache.desc.d_b==0)
       G_EIP(vm) &= 0x0000ffff;
     return;
     }
   else { /* gate & special segment */
     descriptor_cache_t  gate_cache, tss_cache;
     selector_t          gate_selector, tss_selector;
-    Bit32u new_EIP;
-    Bit16u dest_selector;
+    Bit32u gate_offset;
+    selector_t gate_cs_selector;
+    descriptor_cache_t  gate_cs_cache;
     Bit16u raw_tss_selector;
     Bit32u temp_eIP;
 
     /* 1 level of indirection via gate, switch gate & cs */
-    gate_cache    = cs_cache; /* remove this copy +++ */
-    gate_selector = cs_selector;
+    gate_cache    = cache; /* remove this copy +++ */
+    gate_selector = selector;
 
     switch (gate_cache.desc.type) {
       case 1: /* available 16bit TSS */
@@ -1011,324 +1010,484 @@ call_protected(vm_t *vm, Bit16u cs_raw, Bit32u disp32)
         return;
         break;
 
+#endif
       case  4: /* 16bit CALL GATE */
-      case 12: /* 32bit CALL GATE */
+	/* XXX NOT TESTED */
         /* call gate DPL must be >= CPL, else #GP(call gate selector)
          * call gate DPL must be >= RPL, else #GP(call gate selector)
          */
-        if ( (gate_descriptor.dpl < CPL) ||
-             (gate_descriptor.dpl < gate_selector.rpl) ) {
-          monpanic(vm, "call_pro: DPL < CPL or RPL\n");
-          exception(vm, ExceptionGP, gate_selector.value & 0xfffc, 0);
+        if (cache.desc.dpl < G_CPL(vm)) {
+	  monpanic(vm, "call_pro: gate.dpl < CPL\n");
+	  exception(vm, ExceptionGP, cs_raw & 0xfffc);
+	  }
+
+	if (cache.desc.dpl < selector.fields.rpl) {
+          monpanic(vm, "call_pro: gate.dpl < selector.dpl\n");
+          exception(vm, ExceptionGP, cs_raw & 0xfffc);
           }
 
         /* call gate must be present, else #NP(call gate selector) */
-        if (gate_descriptor.p==0) {
-          monpanic(vm, "call_pro: not present\n");
-          exception(vm, ExceptionNP, gate_selector.value & 0xfffc, 0);
+        if (cache.desc.p==0) {
+          monpanic(vm, "call_pro: gate.p == 0\n");
+          exception(vm, ExceptionNP, cs_raw & 0xfffc);
           }
 
         /* examine code segment selector in call gate descriptor */
 
-        if (gate_descriptor.type==4) {
-          dest_selector = gate_descriptor.u.gate286.dest_selector;
-          new_EIP = gate_descriptor.u.gate286.dest_offset;
-          }
-        else {
-          dest_selector = gate_descriptor.u.gate386.dest_selector;
-          new_EIP = gate_descriptor.u.gate386.dest_offset;
-          }
-
         /* selector must not be null else #GP(0) */
-        if ( (dest_selector & 0xfffc) == 0 ) {
-          monpanic(vm, "call_pro: selector in gate null\n");
-          exception(vm, ExceptionGP, 0, 0);
+	gate_cs_selector = ((gate_t *) &cache.desc)->selector;
+        if ( IsNullSelector(gate_cs_selector)) {
+          monpanic(vm, "call_pro: CS selector is null\n");
+          exception(vm, ExceptionGP, 0);
           }
 
-        parse_selector(dest_selector, &cs_selector);
-
         /* selector must be within its descriptor table limits,
          *   else #GP(code segment selector) */
-        fetch_raw_descriptor(&cs_selector, &dword1, &dword2,
+        fetch_raw_descriptor(vm, gate_cs_selector, &gate_cs_cache.desc,
           ExceptionGP);
-        parse_descriptor(dword1, dword2, &cs_descriptor);
+        descriptor2cache(vm, &gate_cs_cache);
 
         /* AR byte of selected descriptor must indicate code segment,
          *   else #GP(code segment selector)
-         * DPL of selected descriptor must be <= CPL,
-         * else #GP(code segment selector)
          */
-        if (cs_descriptor.valid==0 ||
-            cs_descriptor.segment==0 ||
-            cs_descriptor.u.segment.executable==0 ||
-            cs_descriptor.dpl > CPL) {
-          monpanic(vm, "call_pro: selected desciptor not code\n");
-          exception(vm, ExceptionGP, cs_selector.value & 0xfffc, 0);
+        if ( (gate_cs_cache.valid==0) ||
+             !(gate_cs_cache.desc.type & D_S) ||
+	     !(gate_cs_cache.desc.type & D_EXECUTE) ) {
+          monpanic(vm, "call_pro: AR byte: not a code segment.\n");
+          exception(vm, ExceptionGP, gate_cs_selector.raw & 0xfffc);
+          }
+
+        if (gate_cs_cache.desc.dpl > G_CPL(vm)) {
+	  monpanic(vm, "call_pro: conform: code seg des DPL > CPL.\n");
+	  exception(vm, ExceptionGP, gate_cs_selector.raw & 0xfffc);
+	  }
+
+	if (gate_cs_cache.desc.p==0) {
+          monpanic(vm, "call_pro: code seg not present.\n");
+          exception(vm, ExceptionNP, gate_cs_selector.raw & 0xfffc);
+          }
+
+	gate_offset = ((gate_t *) &cache.desc)->offset_low |
+          (((gate_t *) &cache.desc)->offset_high << 16);
+        if ( gate_offset > gate_cs_cache.limit_scaled ) {
+          monpanic(vm, "call_pro: IP > limit\n");
+          exception(vm, ExceptionGP, 0);
           }
 
         /* CALL GATE TO MORE PRIVILEGE
-         * if non-conforming code segment and DPL < CPL then
-         * ??? use gate_descriptor.dpl or cs_descriptor.dpl ???
+         * We have to do a stack switch.
          */
-        if ( (cs_descriptor.u.segment.c_ed==0)  &&
-             (cs_descriptor.dpl < CPL) ) {
-
-          Bit16u SS_for_cpl_x;
-          Bit32u ESP_for_cpl_x;
-          bx_selector_t   ss_selector;
-          bx_descriptor_t ss_descriptor;
-          unsigned room_needed;
+	if ( (gate_cs_cache.desc.dpl < G_CPL(vm)) &&
+	     !(gate_cs_cache.desc.type & D_CONFORM)) {
+          Bit16u SP_for_cpl_x;
+          selector_t   ss_selector;
+          descriptor_cache_t ss_cache;
+	  unsigned room_needed;
           Bit8u    param_count;
           Bit16u   return_SS, return_CS;
-          Bit32u   return_ESP, return_EIP;
+          Bit16u   return_SP, return_IP;
           Bit32u   return_ss_base;
           unsigned i;
           Bit16u   parameter_word[32];
           Bit32u   parameter_dword[32];
-          Bit32u   temp_ESP;
+          Bit16u   temp_SP;
 
           /* get new SS selector for new privilege level from TSS */
-          get_SS_ESP_from_TSS(cs_descriptor.dpl,
-                              &SS_for_cpl_x, &ESP_for_cpl_x);
+          get_SS_ESP_from_TSS(vm, gate_cs_cache.desc.dpl,
+                              &ss_selector, (Bit32u *)&SP_for_cpl_x);
 
-/* ??? use dpl or rpl ??? */
 
           /* check selector & descriptor for new SS:
            * selector must not be null, else #TS(0)
            */
-          if ( (SS_for_cpl_x & 0xfffc) == 0 ) {
-            monpanic(vm, "call_pro: new SS null\n");
-            exception(vm, ExceptionTS, 0, 0);
+          if ( IsNullSelector(ss_selector)) {
+            monpanic(vm, "call_pro: new SS selector is null\n");
+            exception(vm, ExceptionTS, 0);
             return;
             }
 
           /* selector index must be within its descriptor table limits,
            *   else #TS(SS selector)
            */
-          parse_selector(SS_for_cpl_x, &ss_selector);
-          fetch_raw_descriptor(&ss_selector, &dword1, &dword2,
-            BX_TS_EXCEPTION);
 
-          parse_descriptor(dword1, dword2, &ss_descriptor);
+          fetch_raw_descriptor(vm, ss_selector, &ss_cache.desc,
+            ExceptionTS);
+
+          descriptor2cache(vm, &ss_cache);
 
           /* selector's RPL must equal DPL of code segment,
            *   else #TS(SS selector)
            */
-          if (ss_selector.rpl != cs_descriptor.dpl) {
+          if (ss_selector.fields.rpl != gate_cs_cache.desc.dpl) {
             monpanic(vm, "call_pro: SS selector.rpl != CS descr.dpl\n");
-            exception(vm, ExceptionTS, SS_for_cpl_x & 0xfffc, 0);
+            exception(vm, ExceptionTS, ss_selector.raw & 0xfffc);
             return;
             }
 
           /* stack segment DPL must equal DPL of code segment,
            *   else #TS(SS selector)
            */
-          if (ss_descriptor.dpl != cs_descriptor.dpl) {
+          if (ss_cache.desc.dpl != gate_cs_cache.desc.dpl) {
             monpanic(vm, "call_pro: SS descr.rpl != CS descr.dpl\n");
-            exception(vm, ExceptionTS, SS_for_cpl_x & 0xfffc, 0);
+            exception(vm, ExceptionTS, ss_selector.raw & 0xfffc);
             return;
             }
 
           /* descriptor must indicate writable data segment,
            *   else #TS(SS selector)
            */
-          if (ss_descriptor.valid==0 ||
-              ss_descriptor.segment==0  ||
-              ss_descriptor.u.segment.executable ||
-              ss_descriptor.u.segment.r_w==0) {
-            bx_printf("call_pro: ss descriptor not writable data seg\n");
-            exception(vm, ExceptionTS, SS_for_cpl_x & 0xfffc, 0);
+          if (ss_cache.valid==0 ||
+              (ss_cache.desc.type & D_EXECUTE) ||
+              !(ss_cache.desc.type & D_WRITE)) {
+            monprint(vm, "call_pro: ss descriptor not writable data seg\n");
+            exception(vm, ExceptionTS, ss_selector.raw & 0xfffc);
             return;
             }
 
           /* segment must be present, else #SS(SS selector) */
-          if (ss_descriptor.p==0) {
+          if (ss_cache.desc.p==0) {
             monpanic(vm, "call_pro: ss descriptor not present.\n");
-            exception(vm, ExceptionSS, SS_for_cpl_x & 0xfffc, 0);
+            exception(vm, ExceptionSS, ss_selector.raw & 0xfffc);
             return;
             }
 
-          if ( cs_descriptor.u.segment.d_b )
-            /* new stack must have room for parameters plus 16 bytes */
-            room_needed = 16;
-          else
-            /* new stack must have room for parameters plus 8 bytes */
-            room_needed =  8;
+          /* get word count from call gate, mask to 5 bits */
+          param_count = ((gate_t *) &cache.desc)->count & 0x1f;
 
-          if (gate_descriptor.type==4) {
-            /* get word count from call gate, mask to 5 bits */
-            param_count = gate_descriptor.u.gate286.word_count & 0x1f;
-            room_needed += param_count*2;
-            }
-          else {
-            /* get word count from call gate, mask to 5 bits */
-            param_count = gate_descriptor.u.gate386.dword_count & 0x1f;
-            room_needed += param_count*4;
-            }
+          /* new stack must have room for parameters plus 8 bytes */
+          room_needed = param_count*2 + 8;
 
           /* new stack must have room for parameters plus return info
            *   else #SS(SS selector)
            */
 
-          if ( !can_push(&ss_descriptor, ESP_for_cpl_x, room_needed) ) {
-            bx_printf("call_pro: stack doesn't have room\n");
-            exception(vm, ExceptionSS, SS_for_cpl_x & 0xfffc, 0);
+          if ( !can_push(vm, &ss_cache, SP_for_cpl_x, room_needed) ) {
+            monprint(vm, "call_pro: stack doesn't have room\n");
+            exception(vm, ExceptionSS, ss_selector.raw & 0xfffc);
             return;
             }
 
           /* new eIP must be in code segment limit else #GP(0) */
-          if ( new_EIP > cs_descriptor.u.segment.limit_scaled ) {
+          if ( gate_offset > LimitOfDataDescriptor(gate_cs_cache.desc) ) {
             monpanic(vm, "call_pro: IP not within CS limits\n");
-            exception(vm, ExceptionGP, 0, 0);
+            exception(vm, ExceptionGP, 0);
             return;
             }
 
 
           /* save return SS:eSP to be pushed on new stack */
-          return_SS = BX_CPU_THIS_PTR sregs[BX_SEG_REG_SS].selector.value;
-          if (BX_CPU_THIS_PTR sregs[BX_SEG_REG_SS].cache.u.segment.d_b)
-            return_ESP = ESP;
-          else
-            return_ESP =  SP;
-          return_ss_base = BX_CPU_THIS_PTR sregs[BX_SEG_REG_SS].cache.u.segment.base;
+	  return_SS = vm->guest_cpu.selector[SRegSS].raw;
+
+          return_SP = G_SP(vm);
+          return_ss_base = vm->guest_cpu.desc_cache[SRegSS].base;
 
           /* save return CS:eIP to be pushed on new stack */
-          return_CS = BX_CPU_THIS_PTR sregs[BX_SEG_REG_CS].selector.value;
-          if ( cs_descriptor.u.segment.d_b )
-            return_EIP = EIP;
-          else
-            return_EIP =  IP;
+          return_CS = vm->guest_cpu.selector[SRegCS].raw;
+          return_IP = G_IP(vm);
 
 
-          if (gate_descriptor.type==4) {
-            for (i=0; i<param_count; i++) {
-              access_linear(return_ss_base + return_ESP + i*2,
-                2, 0, BX_READ, &parameter_word[i]);
-              }
-            }
-          else {
-            for (i=0; i<param_count; i++) {
-              access_linear(return_ss_base + return_ESP + i*4,
-                4, 0, BX_READ, &parameter_dword[i]);
-              }
-            }
+          for (i=0; i<param_count; i++) {
+            access_linear(vm, return_ss_base + return_SP + i*2,
+              2, 0, OP_READ, &parameter_word[i]);
+          }
 
           /* load new SS:SP value from TSS */
           /* load SS descriptor */
-          load_ss_pro(&ss_selector, &ss_descriptor, ss_descriptor.dpl);
-          if (ss_descriptor.u.segment.d_b)
-            ESP = ESP_for_cpl_x;
-          else
-            SP =  (Bit16u) ESP_for_cpl_x;
-
+          load_ss_pro(vm, ss_selector, &ss_cache, ss_cache.desc.dpl);
+          G_SP(vm) = SP_for_cpl_x;
+          
           /* load new CS:IP value from gate */
           /* load CS descriptor */
           /* set CPL to stack segment DPL */
           /* set RPL of CS to CPL */
-          load_cs_pro(&cs_selector, &cs_descriptor, cs_descriptor.dpl);
-          EIP = new_EIP;
+          load_cs_pro(vm, gate_cs_selector, &gate_cs_cache, gate_cs_cache.desc.dpl);
+          G_IP(vm) = gate_offset;
 
           /* push pointer of old stack onto new stack */
-          if (gate_descriptor.type==4) {
-            push_16(return_SS);
-            push_16((Bit16u) return_ESP);
-            }
-          else {
-            push_32(return_SS);
-            push_32(return_ESP);
-            }
+          push16(vm, return_SS);
+          push16(vm, return_SP);
 
           /* get word count from call gate, mask to 5 bits */
           /* copy parameters from old stack onto new stack */
-          if (BX_CPU_THIS_PTR sregs[BX_SEG_REG_SS].cache.u.segment.d_b)
-            temp_ESP = ESP;
-          else
-            temp_ESP =  SP;
+          temp_SP = G_SP(vm);
 
-          if (gate_descriptor.type==4) {
-            for (i=param_count; i>0; i--) {
-              push_16(parameter_word[i-1]);
-              }
+          for (i=param_count; i>0; i--) {
+            push16(vm, parameter_word[i-1]);
             }
-          else {
-            for (i=param_count; i>0; i--) {
-              push_32(parameter_dword[i-1]);
-              }
-            }
-
+          
           /* push return address onto new stack */
-          if (gate_descriptor.type==4) {
-            push_16(return_CS);
-            push_16((Bit16u) return_EIP);
-            }
-          else {
-            push_32(return_CS);
-            push_32(return_EIP);
-            }
+          push16(vm, return_CS);
+          push16(vm, return_IP);
 
           return;
           }
-
         /* CALL GATE TO SAME PRIVILEGE */
         else {
-          Bit32u temp_ESP;
+          Bit16u temp_SP;
 
-          if (BX_CPU_THIS_PTR sregs[BX_SEG_REG_SS].cache.u.segment.d_b)
-            temp_ESP = ESP;
-          else
-            temp_ESP = SP;
+          temp_SP = G_SP(vm);
 
-          if (gate_descriptor.type==12) {
-            if ( !can_push(&BX_CPU_THIS_PTR sregs[BX_SEG_REG_SS].cache, temp_ESP, 8) ) {
-              monpanic(vm, "call_pro: stack doesn't have room for 8 bytes\n");
-              exception(vm, ExceptionSS, 0, 0);
-              }
-            }
-          else {
-            /* stack must room for 4-byte return address
-             *   else #SS(0) */
-            if ( !can_push(&BX_CPU_THIS_PTR sregs[BX_SEG_REG_SS].cache, temp_ESP, 4) ) {
-              monpanic(vm, "call_pro: stack doesn't have room for 4 bytes\n");
-              exception(vm, ExceptionSS, 0, 0);
-              }
-            }
-
-          /* EIP must be within code segment limit, else #GP(0) */
-          if ( new_EIP > cs_descriptor.u.segment.limit_scaled ) {
-            monpanic(vm, "call_pro: IP not within code segment limits\n");
-            exception(vm, ExceptionGP, 0, 0);
-            }
-
-          if (gate_descriptor.type==12) {
-            /* push return address onto stack */
-            push_32(BX_CPU_THIS_PTR sregs[BX_SEG_REG_CS].selector.value);
-            push_32(EIP);
-            }
-          else {
-            /* push return address onto stack */
-            push_16(BX_CPU_THIS_PTR sregs[BX_SEG_REG_CS].selector.value);
-            push_16(IP);
+          if ( !can_push(vm, &vm->guest_cpu.desc_cache[SRegSS], temp_SP, 4) ) {
+            monpanic(vm, "call_pro: stack doesn't have room for ret addr (4 bytes)\n");
+            exception(vm, ExceptionSS, 0);
             }
+                    
+          /* push return address onto stack */
+          push16(vm, vm->guest_cpu.selector[SRegCS].raw);
+          push16(vm, G_IP(vm));
 
           /* load CS:EIP from gate
            * load code segment descriptor into CS register
            * set RPL of CS to CPL
            */
-          load_cs_pro(&cs_selector, &cs_descriptor, CPL);
-          EIP = new_EIP;
+          load_cs_pro(vm, gate_cs_selector, &gate_cs_cache, G_CPL(vm));
+          G_IP(vm) = gate_offset;
+
+          return;
+          }
+      case 12: /* 32bit CALL GATE */
+        /* call gate DPL must be >= CPL, else #GP(call gate selector)
+         * call gate DPL must be >= RPL, else #GP(call gate selector)
+         */
+        if (cache.desc.dpl < G_CPL(vm)) {
+	  monpanic(vm, "call_pro: gate.dpl < CPL\n");
+	  exception(vm, ExceptionGP, cs_raw & 0xfffc);
+	  }
+
+	if (cache.desc.dpl < selector.fields.rpl) {
+          monpanic(vm, "call_pro: gate.dpl < selector.dpl\n");
+          exception(vm, ExceptionGP, cs_raw & 0xfffc);
+          }
+
+        /* call gate must be present, else #NP(call gate selector) */
+        if (cache.desc.p==0) {
+          monpanic(vm, "call_pro: gate.p == 0\n");
+          exception(vm, ExceptionNP, cs_raw & 0xfffc);
+          }
+
+        /* examine code segment selector in call gate descriptor */
+
+        /* selector must not be null else #GP(0) */
+	gate_cs_selector = ((gate_t *) &cache.desc)->selector;
+        if ( IsNullSelector(gate_cs_selector)) {
+          monpanic(vm, "call_pro: CS selector is null\n");
+          exception(vm, ExceptionGP, 0);
+          }
+
+        /* selector must be within its descriptor table limits,
+         *   else #GP(code segment selector) */
+        fetch_raw_descriptor(vm, gate_cs_selector, &gate_cs_cache.desc,
+          ExceptionGP);
+        descriptor2cache(vm, &gate_cs_cache);
+
+        /* AR byte of selected descriptor must indicate code segment,
+         *   else #GP(code segment selector)
+         */
+        if ( (gate_cs_cache.valid==0) ||
+             !(gate_cs_cache.desc.type & D_S) ||
+	     !(gate_cs_cache.desc.type & D_EXECUTE) ) {
+          monpanic(vm, "call_pro: AR byte: not a code segment.\n");
+          exception(vm, ExceptionGP, gate_cs_selector.raw & 0xfffc);
+          }
+
+        if (gate_cs_cache.desc.dpl > G_CPL(vm)) {
+	  monpanic(vm, "call_pro: conform: code seg des DPL > CPL.\n");
+	  exception(vm, ExceptionGP, gate_cs_selector.raw & 0xfffc);
+	  }
+
+	if (gate_cs_cache.desc.p==0) {
+          monpanic(vm, "call_pro: code seg not present.\n");
+          exception(vm, ExceptionNP, gate_cs_selector.raw & 0xfffc);
+          }
+
+	gate_offset = ((gate_t *) &cache.desc)->offset_low |
+          (((gate_t *) &cache.desc)->offset_high << 16);
+        if ( gate_offset > gate_cs_cache.limit_scaled ) {
+          monpanic(vm, "call_pro: IP > limit\n");
+          exception(vm, ExceptionGP, 0);
+          }
+
+        /* CALL GATE TO MORE PRIVILEGE
+         * if non-conforming code segment and DPL < CPL then
+         * ??? use gate_descriptor.dpl or cs_descriptor.dpl ???
+         */
+	if ( (gate_cs_cache.desc.dpl < G_CPL(vm)) &&
+	     !(gate_cs_cache.desc.type & D_CONFORM)) {
+	  /* STACK SWITCH */
+          Bit32u ESP_for_cpl_x;
+          selector_t   ss_selector;
+          descriptor_cache_t ss_cache;
+	  unsigned room_needed;
+          Bit8u    param_count;
+          Bit16u   return_SS, return_CS;
+          Bit32u   return_ESP, return_EIP;
+          Bit32u   return_ss_base;
+          unsigned i;
+          Bit32u   parameter_dword[32];
+          Bit32u   temp_ESP;
+
+          /* get new SS selector for new privilege level from TSS */
+          get_SS_ESP_from_TSS(vm, gate_cs_cache.desc.dpl,
+                              &ss_selector, &ESP_for_cpl_x);
+
+
+          /* check selector & descriptor for new SS:
+           * selector must not be null, else #TS(0)
+           */
+          if ( IsNullSelector(ss_selector)) {
+            monpanic(vm, "call_pro: new SS selector is null\n");
+            exception(vm, ExceptionTS, 0);
+            return;
+            }
+
+          /* selector index must be within its descriptor table limits,
+           *   else #TS(SS selector)
+           */
+
+          fetch_raw_descriptor(vm, ss_selector, &ss_cache.desc,
+            ExceptionTS);
+
+          descriptor2cache(vm, &ss_cache);
+
+          /* selector's RPL must equal DPL of code segment,
+           *   else #TS(SS selector)
+           */
+          if (ss_selector.fields.rpl != gate_cs_cache.desc.dpl) {
+            monpanic(vm, "call_pro: SS selector.rpl != CS descr.dpl\n");
+            exception(vm, ExceptionTS, ss_selector.raw & 0xfffc);
+            return;
+            }
+
+          /* stack segment DPL must equal DPL of code segment,
+           *   else #TS(SS selector)
+           */
+          if (ss_cache.desc.dpl != gate_cs_cache.desc.dpl) {
+            monpanic(vm, "call_pro: SS descr.rpl != CS descr.dpl\n");
+            exception(vm, ExceptionTS, ss_selector.raw & 0xfffc);
+            return;
+            }
+
+          /* descriptor must indicate writable data segment,
+           *   else #TS(SS selector)
+           */
+          if (ss_cache.valid==0 ||
+              (ss_cache.desc.type & D_EXECUTE) ||
+              !(ss_cache.desc.type & D_WRITE)) {
+            monprint(vm, "call_pro: ss descriptor not writable data seg\n");
+            exception(vm, ExceptionTS, ss_selector.raw & 0xfffc);
+            return;
+            }
+
+          /* segment must be present, else #SS(SS selector) */
+          if (ss_cache.desc.p==0) {
+            monpanic(vm, "call_pro: ss descriptor not present.\n");
+            exception(vm, ExceptionSS, ss_selector.raw & 0xfffc);
+            return;
+            }
+
+          /* get word count from call gate, mask to 5 bits */
+          param_count = ((gate_t *) &cache.desc)->count & 0x1f;
+
+          /* new stack must have room for parameters plus 16 bytes */
+          room_needed = param_count*4 + 16;
+
+          /* new stack must have room for parameters plus return info
+           *   else #SS(SS selector)
+           */
+
+          if ( !can_push(vm, &ss_cache, ESP_for_cpl_x, room_needed) ) {
+            monprint(vm, "call_pro: stack doesn't have room\n");
+            exception(vm, ExceptionSS, ss_selector.raw & 0xfffc);
+            return;
+            }
+
+          /* new eIP must be in code segment limit else #GP(0) */
+          if ( gate_offset > LimitOfDataDescriptor(gate_cs_cache.desc) ) {
+            monpanic(vm, "call_pro: IP not within CS limits\n");
+            exception(vm, ExceptionGP, 0);
+            return;
+            }
+
+
+          /* save return SS:eSP to be pushed on new stack */
+	  return_SS = vm->guest_cpu.selector[SRegSS].raw;
+
+          return_ESP = G_ESP(vm);
+          return_ss_base = vm->guest_cpu.desc_cache[SRegSS].base;
+
+          /* save return CS:eIP to be pushed on new stack */
+          return_CS = vm->guest_cpu.selector[SRegCS].raw;
+          return_EIP = G_EIP(vm);
+
+
+          for (i=0; i<param_count; i++) {
+            access_linear(vm, return_ss_base + return_ESP + i*4,
+              4, 0, OP_READ, &parameter_dword[i]);
+          }
+
+          /* load new SS:SP value from TSS */
+          /* load SS descriptor */
+          load_ss_pro(vm, ss_selector, &ss_cache, ss_cache.desc.dpl);
+          G_ESP(vm) = ESP_for_cpl_x;
+          
+          /* load new CS:IP value from gate */
+          /* load CS descriptor */
+          /* set CPL to stack segment DPL */
+          /* set RPL of CS to CPL */
+          load_cs_pro(vm, gate_cs_selector, &gate_cs_cache, gate_cs_cache.desc.dpl);
+          G_EIP(vm) = gate_offset;
+
+          /* push pointer of old stack onto new stack */
+          push32(vm, return_SS);
+          push32(vm, return_ESP);
+
+          /* get word count from call gate, mask to 5 bits */
+          /* copy parameters from old stack onto new stack */
+          temp_ESP = G_ESP(vm);
+
+          for (i=param_count; i>0; i--) {
+            push32(vm, parameter_dword[i-1]);
+            }
+          
+          /* push return address onto new stack */
+          push32(vm, return_CS);
+          push32(vm, return_EIP);
+
+          return;
+          }
+        /* CALL GATE TO SAME PRIVILEGE */
+        else {
+          Bit32u temp_ESP;
+
+          if (vm->guest_cpu.desc_cache[SRegSS].desc.d_b)
+            temp_ESP = G_ESP(vm);
+          else
+            temp_ESP = G_SP(vm);
+
+          if ( !can_push(vm, &vm->guest_cpu.desc_cache[SRegSS], temp_ESP, 8) ) {
+            monpanic(vm, "call_pro: stack doesn't have room for ret addr (8 bytes)\n");
+            exception(vm, ExceptionSS, 0);
+            }
+                    
+          /* push return address onto stack */
+          push32(vm, (Bit32u) vm->guest_cpu.selector[SRegCS].raw);
+          push32(vm, G_EIP(vm));
+
+          /* load CS:EIP from gate
+           * load code segment descriptor into CS register
+           * set RPL of CS to CPL
+           */
+          load_cs_pro(vm, gate_cs_selector, &gate_cs_cache, G_CPL(vm));
+          G_EIP(vm) = gate_offset;
 
           return;
           }
 
-        monpanic(vm, "call_pro: call gate: should not get here\n");
-        return;
-
-#endif
       default:
         monprint(vm, "call_pro: type = %u\n",
-          (unsigned) cs_cache.desc.type);
+          (unsigned) cache.desc.type);
         monpanic(vm, "call_pro: type = %u\n",
-          (unsigned) cs_cache.desc.type);
+          (unsigned) cache.desc.type);
         return;
       }
     monpanic(vm, "call_pro: gate segment unfinished\n");
