@@ -24,6 +24,7 @@
 #include "monitor.h"
 
 #include <linux/config.h>
+#include <linux/modversions.h>
 #include <linux/module.h>
 #include <linux/types.h>
 #include <linux/kernel.h>
@@ -127,6 +128,12 @@ static int plex86_open(struct inode *, struct file *);
     static int plex86_mmap(struct inode * inode, struct file * file, struct vm_area_struct * vma);
 #endif
 
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,9)
+/* New License scheme */
+#ifdef MODULE_LICENSE
+MODULE_LICENSE("LGPL");
+#endif
+#endif
 
 
 
@@ -695,9 +702,9 @@ printk(KERN_WARNING "plex86: deactivateTimer returns %d.\n", ret);
       if (irqMsg.irq > 15)
         return -EINVAL;
       if (irqMsg.on)
-        picTriggerIRQ(vm, irqMsg.irq);
+        picRaiseIRQ(vm, irqMsg.irq);
       else
-        picUntriggerIRQ(vm, irqMsg.irq);
+        picLowerIRQ(vm, irqMsg.irq);
       return(0); /* OK. */
       }
 
@@ -1129,3 +1136,4 @@ hostprint(char *fmt, ...)
     printk(KERN_WARNING "%s\n", buffer);
     }
 }
+
