@@ -25,6 +25,10 @@
 void initIODevices(vm_t *vm);
 void picInit(vm_t *vm);
 void pitInit(vm_t *vm);
+#if 0
 unsigned picTriggerIRQ(vm_t *vm, unsigned irq);
 unsigned picUntriggerIRQ(vm_t *vm, unsigned irq);
+#endif
+unsigned picRaiseIRQ(vm_t *vm, unsigned irq);
+unsigned picLowerIRQ(vm_t *vm, unsigned irq);
 void miscIOInit(vm_t *vm);
