@@ -71,6 +71,7 @@ pitIORead(vm_t *vm, Bit32u address, unsigned io_len)
     default:
       monpanic(vm, "pit: unsupported io read from port %04x\n", address);
     }
+  return(0);
 }
 
   void
@@ -152,7 +153,7 @@ bx_printf("timer 0-2 mode control: comm:%02x mode:%02x bcd_mode:%u\n",
           vm->pit.s.timer[1].input_latch_value = 0;
           vm->pit.s.timer[1].input_latch_toggle = 0;
           vm->pit.s.timer[1].bcd_mode    = bcd_mode;
-          if ( mode!=2 || bcd_mode!=0 )
+          if ( (mode!=3 && mode!=2 && mode!=0) || bcd_mode!=0 )
             monpanic(vm, "pit: outp(43h): comm 7, mode %02x, bcd %02x unhandled\n",
               (unsigned) mode, bcd_mode);
           break;
@@ -171,7 +172,7 @@ bx_printf("timer 0-2 mode control: comm:%02x mode:%02x bcd_mode:%u\n",
           vm->pit.s.timer[2].input_latch_value = 0;
           vm->pit.s.timer[2].input_latch_toggle = 0;
           vm->pit.s.timer[2].bcd_mode    = bcd_mode;
-          if ( (mode!=3 && mode!=2) || bcd_mode!=0 )
+          if ( (mode!=3 && mode!=2 && mode!=0) || bcd_mode!=0 )
             monpanic(vm, "pit: outp(43h): comm Bh, mode %02x, bcd %02x unhandled\n",
               (unsigned) mode, bcd_mode);
           break;
@@ -485,9 +486,10 @@ pitStart(vm_t *vm, unsigned timerid)
   void
 pitTimerCallback(vm_t *vm, unsigned i)
 {
-  // xxx Fix this
+  // This is a hack to make the IRQ0 work
   if ( pitPeriodic(vm, PitPeriodicDelta) ) {
-    picTriggerIRQ(vm, 0);
+    picLowerIRQ(vm, 0);
+    picRaiseIRQ(vm, 0);
     }
 }
 
