@@ -1,4 +1,7 @@
-/*
+/************************************************************************
+ * $Id: decode.c,v 1.2 2002/07/21 23:14:55 DemonLord Exp $
+ ************************************************************************
+ *
  *  plex86: run multiple x86 operating systems concurrently
  *  Copyright (C) 1999-2000  Ulrich Weigand
  *
@@ -155,7 +158,7 @@ int AddrHasModRM[] = { FALSE, FALSE, TRUE, TRUE, TRUE, FALSE, TRUE, FALSE,
 #define IsOpSegment(op)   ( ((op) & OP_Segment) != 0 )
 #define SegName(op)       ( (op) & 0xffff )
 
-#define st(name)          ( OP_FPUStack | REG_ST##name  )
+#define st(name)          ( OP_FPUStack | P86_REG_ST##name  )
 #define IsOpFPUStack(op)  ( ((op) & OP_FPUStack) != 0 )
 #define FPUStackName(op)  ( (op) & 0xffff )
 
@@ -1601,32 +1604,32 @@ struct instruction EscAuxMap[] =
 int MapSeg[] = 
   { SEG_ES,  SEG_CS,  SEG_SS,  SEG_DS,  SEG_FS,  SEG_GS,   0,       0       };
 int MapControl[] = 
-  { REG_CR0, REG_CR1, REG_CR2, REG_CR3, REG_CR4, REG_CR5,  REG_CR6, REG_CR7 };
+  { P86_REG_CR0, P86_REG_CR1, P86_REG_CR2, P86_REG_CR3, P86_REG_CR4, P86_REG_CR5,  P86_REG_CR6, P86_REG_CR7 };
 int MapDebug[] = 
-  { REG_DR0, REG_DR1, REG_DR2, REG_DR3, REG_DR4, REG_DR5,  REG_DR6, REG_DR7 };
+  { P86_REG_DR0, P86_REG_DR1, P86_REG_DR2, P86_REG_DR3, P86_REG_DR4, P86_REG_DR5,  P86_REG_DR6, P86_REG_DR7 };
 int MapTest[] = 
-  { REG_TR0, REG_TR1, REG_TR2, REG_TR3, REG_TR4, REG_TR5,  REG_TR6, REG_TR7 };
+  { P86_REG_TR0, P86_REG_TR1, P86_REG_TR2, P86_REG_TR3, P86_REG_TR4, P86_REG_TR5,  P86_REG_TR6, P86_REG_TR7 };
 int MapMMX[] = 
-  { REG_MM0, REG_MM1, REG_MM2, REG_MM3, REG_MM4, REG_MM5,  REG_MM6, REG_MM7 };
+  { P86_REG_MM0, P86_REG_MM1, P86_REG_MM2, P86_REG_MM3, P86_REG_MM4, P86_REG_MM5,  P86_REG_MM6, P86_REG_MM7 };
 int MapXMMX[] = 
-  { REG_XMM0, REG_XMM1, REG_XMM2, REG_XMM3, REG_XMM4, REG_XMM5,  REG_XMM6, REG_XMM7 };
+  { P86_REG_XMM0, P86_REG_XMM1, P86_REG_XMM2, P86_REG_XMM3, P86_REG_XMM4, P86_REG_XMM5,  P86_REG_XMM6, P86_REG_XMM7 };
 
 int MapReg8[] = 
-  { REG_AL,  REG_CL,  REG_DL,  REG_BL,  REG_AH,  REG_CH,  REG_DH,  REG_BH  };
+  { P86_REG_AL,  P86_REG_CL,  P86_REG_DL,  P86_REG_BL,  P86_REG_AH,  P86_REG_CH,  P86_REG_DH,  P86_REG_BH  };
 int MapReg16[] = 
-  { REG_AX,  REG_CX,  REG_DX,  REG_BX,  REG_SP,  REG_BP,  REG_SI,  REG_DI  };
+  { P86_REG_AX,  P86_REG_CX,  P86_REG_DX,  P86_REG_BX,  P86_REG_SP,  P86_REG_BP,  P86_REG_SI,  P86_REG_DI  };
 int MapReg32[] = 
-  { REG_EAX, REG_ECX, REG_EDX, REG_EBX, REG_ESP, REG_EBP, REG_ESI, REG_EDI };
+  { P86_REG_EAX, P86_REG_ECX, P86_REG_EDX, P86_REG_EBX, P86_REG_ESP, P86_REG_EBP, P86_REG_ESI, P86_REG_EDI };
 
 int EffBase16[]  = 
-  { REG_BX,  REG_BX,  REG_BP,  REG_BP,  0,       0,       REG_BP,  REG_BX  };
+  { P86_REG_BX,  P86_REG_BX,  P86_REG_BP,  P86_REG_BP,  0,       0,       P86_REG_BP,  P86_REG_BX  };
 int EffIndex16[] = 
-  { REG_SI,  REG_DI,  REG_SI,  REG_DI,  REG_SI,  REG_DI,  0,       0       };
+  { P86_REG_SI,  P86_REG_DI,  P86_REG_SI,  P86_REG_DI,  P86_REG_SI,  P86_REG_DI,  0,       0       };
 
 int EffBase32[]  = 
-  { REG_EAX, REG_ECX, REG_EDX, REG_EBX, REG_ESP, REG_EBP, REG_ESI, REG_EDI };
+  { P86_REG_EAX, P86_REG_ECX, P86_REG_EDX, P86_REG_EBX, P86_REG_ESP, P86_REG_EBP, P86_REG_ESI, P86_REG_EDI };
 int EffIndex32[] = 
-  { REG_EAX, REG_ECX, REG_EDX, REG_EBX, 0,       REG_EBP, REG_ESI, REG_EDI };
+  { P86_REG_EAX, P86_REG_ECX, P86_REG_EDX, P86_REG_EBX, 0,       P86_REG_EBP, P86_REG_ESI, P86_REG_EDI };
 
 
 static int sign_extend( int val, int nFromSize, int nToSize )
@@ -1793,19 +1796,19 @@ int i386_decode( struct i386_context *ctx, struct i386_decode *dc )
         {
             base  = EffBase16 [rm];
             index = EffIndex16[rm];
-            if (base == REG_BP && !index && !mod) base = 0, mod = 2;
+            if (base == P86_REG_BP && !index && !mod) base = 0, mod = 2;
         }
         else
         {
             /* Get SIB byte if present */
-            if ((base = EffBase32[rm]) == REG_ESP)
+            if ((base = EffBase32[rm]) == P86_REG_ESP)
             {
                 if (!i386_decode_raw( ctx, &opcode, 1 )) goto fault;
                 base  = EffBase32 [ FieldSIBBase(opcode)  ];
                 index = EffIndex32[ FieldSIBIndex(opcode) ];
                 scale = index? 1 << FieldSIBScale(opcode) : 0;
             }
-            if (base == REG_EBP && !mod) base = 0, mod = 2;
+            if (base == P86_REG_EBP && !mod) base = 0, mod = 2;
         }
 
         /* Get displacement if present */
@@ -1940,7 +1943,7 @@ int i386_decode( struct i386_context *ctx, struct i386_decode *dc )
         case AddrMode_F:    /* flags register */
             if (da->type != OT_VALUE) goto fault;
             da->mode = OM_REGISTER;
-            da->u.reg = size == 2? REG_FL : REG_EFL;
+            da->u.reg = size == 2? P86_REG_FL : P86_REG_EFL;
             break;
 
         case AddrMode_C:    /* control register in reg field */
@@ -2021,8 +2024,8 @@ int i386_decode( struct i386_context *ctx, struct i386_decode *dc )
 
             if (dc->segment_override)
                 da->u.memory.segment = dc->segment_override;
-            else if (   base == REG_BP  || base == REG_SP
-                     || base == REG_EBP || base == REG_ESP)
+            else if (   base == P86_REG_BP  || base == P86_REG_SP
+                     || base == P86_REG_EBP || base == P86_REG_ESP)
                 da->u.memory.segment = SEG_SS;
             else
                 da->u.memory.segment = SEG_DS;
@@ -2041,13 +2044,13 @@ int i386_decode( struct i386_context *ctx, struct i386_decode *dc )
             else
                 da->u.memory.segment = SEG_DS;
 
-            da->u.memory.base = dc->address_size == 2 ? REG_SI : REG_ESI;
+            da->u.memory.base = dc->address_size == 2 ? P86_REG_SI : P86_REG_ESI;
             break;
 
         case AddrMode_Y:    /* memory addressed by ES:(E)DI */
             da->mode = OM_MEMORY;
             da->u.memory.segment = SEG_ES; /* no override possible */
-            da->u.memory.base = dc->address_size == 2 ? REG_DI : REG_EDI;
+            da->u.memory.base = dc->address_size == 2 ? P86_REG_DI : P86_REG_EDI;
             break;
         }
     }
@@ -2369,7 +2372,7 @@ int i386_decode_text_intel( struct i386_context *ctx, struct i386_decode *dc, ch
         switch (da->mode)
         {
         case OM_REGISTER:
-            assert( da->u.reg < REG_LAST );
+            assert( da->u.reg < P86_REG_LAST );
             strcat( string, RegNamesIntel[da->u.reg] );
             break;
 
@@ -2413,7 +2416,7 @@ int i386_decode_text_intel( struct i386_context *ctx, struct i386_decode *dc, ch
             /* Output segment override */
             if ( override )
             {
-                assert( addr->segment < REG_LAST );
+                assert( addr->segment < P86_REG_LAST );
                 sprintf( string+strlen(string), "%s:", RegNamesIntel[addr->segment] );
             }
 
@@ -2426,7 +2429,7 @@ int i386_decode_text_intel( struct i386_context *ctx, struct i386_decode *dc, ch
                 if (!first) strcat( string, "+" );
                 first = FALSE;
 
-                assert( addr->base < REG_LAST );
+                assert( addr->base < P86_REG_LAST );
                 strcat( string, RegNamesIntel[addr->base]);
             }
 
@@ -2436,7 +2439,7 @@ int i386_decode_text_intel( struct i386_context *ctx, struct i386_decode *dc, ch
                 if (!first) strcat( string, "+" );
                 first = FALSE;
 
-                assert( addr->index < REG_LAST );
+                assert( addr->index < P86_REG_LAST );
                 strcat( string, RegNamesIntel[addr->index] );
 
                 if ( addr->scale > 1 )
@@ -2769,7 +2772,7 @@ int i386_decode_text_att( struct i386_context *ctx, struct i386_decode *dc, char
         switch (da->mode)
         {
         case OM_REGISTER:
-            assert( da->u.reg < REG_LAST );
+            assert( da->u.reg < P86_REG_LAST );
             if( dc->instruction == INS_call || dc->instruction == INS_jmp )
                 strcat( string, "*" );
 
@@ -2803,7 +2806,7 @@ int i386_decode_text_att( struct i386_context *ctx, struct i386_decode *dc, char
             /* Output segment override */
             if ( override )
             {
-                assert( addr->segment < REG_LAST );
+                assert( addr->segment < P86_REG_LAST );
                 sprintf( string+strlen(string), "%s:", RegNamesATT[addr->segment] );
             }
 
@@ -2827,14 +2830,14 @@ int i386_decode_text_att( struct i386_context *ctx, struct i386_decode *dc, char
             /* Output base */
             if ( addr->base )
             {
-                assert( addr->base < REG_LAST );
+                assert( addr->base < P86_REG_LAST );
                 strcat( string, RegNamesATT[addr->base]);
             }
 
             /* Output index and scale */
             if ( addr->index )
             {
-                assert( addr->index < REG_LAST );
+                assert( addr->index < P86_REG_LAST );
                 sprintf( string+strlen(string), ",%s", RegNamesATT[addr->index] );
 
                 if ( addr->scale > 1 )
