@@ -1,4 +1,7 @@
-/*
+/************************************************************************
+ * $Id: curses.cc,v 1.3 2002/07/21 23:14:55 DemonLord Exp $
+ ************************************************************************
+ *
  *  Copyright 2000 by Josh Wilmes (josh@hitchhiker.org)
  *
  *  This library is free software; you can redistribute it and/or
@@ -254,6 +257,16 @@ bx_gui_c::specific_init(bx_gui_c *th, int argc, char **argv, unsigned tilewidth,
 #endif /* CURSES_COLOR */
 }
 
+// This is called whenever the mouse_enabled parameter changes.  It
+// can change because of a gui event such as clicking on the mouse-enable
+// bitmap or pressing the middle button, or from the control panel.
+// In all those cases, setting the parameter value will get you here.
+  void
+bx_gui_c::mouse_enabled_changed_specific (Boolean val)
+{
+  UNUSED(val);
+}
+
 
 // ::HANDLE_EVENTS()
 //
@@ -383,12 +396,13 @@ bx_gui_c::clear_screen(void)
   void
 bx_gui_c::text_update(Bit8u *old_text, Bit8u *new_text,
                       unsigned long cursor_x, unsigned long cursor_y,
-		      unsigned rows)
+                      Bit16u cursor_state, unsigned rows) 
 {
   UNUSED(old_text);
   UNUSED(new_text);
   UNUSED(cursor_x);
   UNUSED(cursor_y);
+  UNUSED(cursor_state);
 
   if (in_graphics_mode) {
     printf("Text mode has been re-entered - display is now available\n");
@@ -488,10 +502,11 @@ bx_gui_c::graphics_tile_update(Bit8u *tile, unsigned x0, unsigned y0)
 // y: new VGA y size (add headerbar_y parameter from ::specific_init().
 
   void
-bx_gui_c::dimension_update(unsigned x, unsigned y)
+bx_gui_c::dimension_update(unsigned x, unsigned y, unsigned fheight)
 {
   UNUSED(x);
   UNUSED(y);
+  UNUSED(fheight);
 }
 
 
