@@ -1,21 +1,33 @@
-/*
- *  Copyright 1994-2000 by Kevin P. Lawton.
- *
- *  This library is free software; you can redistribute it and/or
- *  modify it under the terms of the GNU Lesser General Public
- *  License as published by the Free Software Foundation; either
- *  version 2 of the License, or (at your option) any later version.
- *
- *  This library is distributed in the hope that it will be useful,
- *  but WITHOUT ANY WARRANTY; without even the implied warranty of
- *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- *  Lesser General Public License for more details.
- *
- *  You should have received a copy of the GNU Lesser General Public
- *  License along with this library; if not, write to the Free Software
- *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
- */
+/////////////////////////////////////////////////////////////////////////
+// $Id: x.cc,v 1.2 2002/08/12 18:43:18 DemonLord Exp $
+/////////////////////////////////////////////////////////////////////////
+// Based on Bochs file <Id: x.cc,v 1.44 2002/08/09 06:16:43 vruppert Exp >
+/////////////////////////////////////////////////////////////////////////
+//
+//  Copyright (C) 2002  MandrakeSoft S.A.
+//
+//    MandrakeSoft S.A.
+//    43, rue d'Aboukir
+//    75002 Paris - France
+//    http://www.linux-mandrake.com/
+//    http://www.mandrakesoft.com/
+//
+//  This library is free software; you can redistribute it and/or
+//  modify it under the terms of the GNU Lesser General Public
+//  License as published by the Free Software Foundation; either
+//  version 2 of the License, or (at your option) any later version.
+//
+//  This library is distributed in the hope that it will be useful,
+//  but WITHOUT ANY WARRANTY; without even the implied warranty of
+//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+//  Lesser General Public License for more details.
+//
+//  You should have received a copy of the GNU Lesser General Public
+//  License along with this library; if not, write to the Free Software
+//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 
+#define XK_PUBLISHING
+#define XK_TECHNICAL
 
 extern "C" {
 #include <X11/Xlib.h>
@@ -26,14 +38,14 @@ extern "C" {
 }
 
 #include "bochs.h"
+#include "keymap.h"
 #include "icon_bochs.h"
 
-
+#define LOG_THIS bx_gui.
 
 
 #define MAX_MAPPED_STRING_LENGTH 10
 
-
 /* These are used as arguments to nearly every Xlib routine, so it saves
  * routine arguments to declare them global.  If there were
  * additional source files, they would be declared extern there. */
@@ -50,6 +62,7 @@ static Window win;
 static GC gc, gc_inv, gc_headerbar, gc_headerbar_inv;
 static XFontStruct *font_info;
 static unsigned font_width, font_height;
+static unsigned font_height_orig = 16;
 static Bit8u blank_line[80];
 static unsigned dimension_x=0, dimension_y=0;
 
@@ -75,6 +88,8 @@ static void warp_cursor(int dx, int dy);
 static void disable_cursor();
 static void enable_cursor();
 
+static Bit32u convertStringToXKeysym (const char *string);
+
 struct {
   Pixmap bmap;
   unsigned xdim;
@@ -235,13 +250,47 @@ static void xkeypress(KeySym keysym, int press_release);
 #define ROUNDUP(nbytes, pad) ((((nbytes) + ((pad)-1)) / (pad)) * ((pad)>>3))
 
 
+#define MAX_VGA_COLORS 256
 
-unsigned long col_vals[256]; // 256 VGA colors
+unsigned long col_vals[MAX_VGA_COLORS]; // 256 VGA colors
 unsigned curr_foreground, curr_background;
 
 static unsigned x_tilesize, y_tilesize;
 
 
+// Try to allocate NCOLORS at once in the colormap provided.  If it can
+// be done, return true.  If not, return false.  (In either case, free
+// up the color cells so that we don't add to the problem!)  This is used
+// to determine whether Bochs should use a private colormap even when the
+// user did not specify it.
+static Boolean
+test_alloc_colors (Colormap cmap, Bit32u n_tries) {
+  XColor color;
+  unsigned long pixel[MAX_VGA_COLORS];
+  Boolean pixel_valid[MAX_VGA_COLORS];
+  Bit32u n_allocated = 0;
+  Bit32u i;
+  color.flags = DoRed | DoGreen | DoBlue;
+  for (i=0; i<n_tries; i++) {
+    // choose wierd color values that are unlikely to already be in the 
+    // colormap.
+    color.red   = ((i+41)%MAX_VGA_COLORS) << 8;
+    color.green = ((i+42)%MAX_VGA_COLORS) << 8;
+    color.blue  = ((i+43)%MAX_VGA_COLORS) << 8;
+    pixel_valid[i] = false;
+    if (XAllocColor (bx_x_display, cmap, &color)) {
+      pixel[i] = color.pixel;
+      pixel_valid[i] = true;
+      n_allocated++;
+    }
+  }
+  BX_INFO (("test_alloc_colors: %d colors available out of %d colors tried", n_allocated, n_tries));
+  // now free them all
+  for (i=0; i<n_tries; i++) {
+    if (pixel_valid[i]) XFreeColors (bx_x_display, cmap, &pixel[i], 1, 0);
+  }
+  return (n_allocated == n_tries);
+}
 
 
   void
@@ -251,7 +300,21 @@ bx_gui_c::specific_init(bx_gui_c *th, int argc, char **argv, unsigned tilewidth,
   unsigned i;
   int x, y;   /* window position */
   unsigned int border_width = 4;  /* four pixels */
+#ifndef PLEX86
+#if BX_CPU_LEVEL < 2
+  char *window_name = "Bochs 8086 emulator, http://bochs.sourceforge.net/";
+#elif BX_CPU_LEVEL == 2
+  char *window_name = "Bochs 80286 emulator, http://bochs.sourceforge.net/";
+#elif BX_CPU_LEVEL == 3
+  char *window_name = "Bochs 80386 emulator, http://bochs.sourceforge.net/";
+#elif BX_CPU_LEVEL == 4
+  char *window_name = "Bochs 80486 emulator, http://bochs.sourceforge.net/";
+#else
+  char *window_name = "Bochs Pentium emulator, http://bochs.sourceforge.net/";
+#endif
+#else
   char *window_name = GUI_WINDOW_TITLE;
+#endif
   char *icon_name = "Bochs";
   Pixmap icon_pixmap;
   XSizeHints size_hints;
@@ -266,13 +329,9 @@ bx_gui_c::specific_init(bx_gui_c *th, int argc, char **argv, unsigned tilewidth,
   unsigned long plane_masks_return[1];
   XColor color;
 
-
+  th->put("XGUI");
   UNUSED(th);
 
-if (bx_options.private_colormap) {
-  bx_printf("# 2WARNING: X11: private_colormap option not handled yet.\n");
-  }
-
   x_tilesize = tilewidth;
   y_tilesize = tileheight;
   bx_headerbar_y = headerbar_y;
@@ -286,8 +345,8 @@ if (bx_options.private_colormap) {
   /* connect to X server */
   if ( (bx_x_display=XOpenDisplay(display_name)) == NULL )
   {
-    bx_panic("%s: cannot connect to X server %s\n",
-        progname, XDisplayName(display_name));
+    BX_PANIC(("%s: cannot connect to X server %s",
+        progname, XDisplayName(display_name)));
   }
 
   /* get screen size from display structure macro */
@@ -308,8 +367,6 @@ if (bx_options.private_colormap) {
   dimension_x = columns * font_width;
   dimension_y = rows * font_height + headerbar_y;
 
-bx_printf("font_width = %u\n", (unsigned) font_width);
-bx_printf("font_height = %u\n", (unsigned) font_height);
 
   /* create opaque window */
   win = XCreateSimpleWindow(bx_x_display, RootWindow(bx_x_display,bx_x_screen_num),
@@ -328,12 +385,30 @@ bx_printf("font_height = %u\n", (unsigned) font_height);
   default_depth  = DefaultDepth(bx_x_display, bx_x_screen_num);
   default_visual = DefaultVisual(bx_x_display, bx_x_screen_num);
 
-  if (bx_options.private_colormap) {
+  if (!bx_options.Oprivate_colormap->get ()) {
+    default_cmap = DefaultColormap(bx_x_display, bx_x_screen_num);
+    // try to use default colormap.  If not enough colors are available,
+    // then switch to private colormap despite the user setting.  There
+    // are too many cases when no colors are available and Bochs simply
+    // draws everything in black on black.
+    if (!test_alloc_colors (default_cmap, 16)) {
+      BX_ERROR (("I can't even allocate 16 colors!  Switching to a private colormap"));
+      bx_options.Oprivate_colormap->set (1);
+    }
+    col_vals[0]  = BlackPixel(bx_x_display, bx_x_screen_num);
+    col_vals[15] = WhitePixel(bx_x_display, bx_x_screen_num);
+    for (i = 1; i < MAX_VGA_COLORS; i++) {
+      if (i==15) continue;
+      col_vals[i] = col_vals[0];
+    }
+  }
+
+  if (bx_options.Oprivate_colormap->get ()) {
     default_cmap = XCreateColormap(bx_x_display, DefaultRootWindow(bx_x_display),
                                    default_visual, AllocNone);
     if (XAllocColorCells(bx_x_display, default_cmap, False,
-                         plane_masks_return, 0, col_vals, 256) == 0) {
-      bx_panic("XAllocColorCells returns error.\n");
+                         plane_masks_return, 0, col_vals, MAX_VGA_COLORS) == 0) {
+      BX_PANIC(("XAllocColorCells returns error."));
       }
 
     win_attr.colormap = default_cmap;
@@ -341,7 +416,7 @@ bx_printf("font_height = %u\n", (unsigned) font_height);
 
     color.flags = DoRed | DoGreen | DoBlue;
 
-    for (i=0; i < 256; i++) {
+    for (i=0; i < MAX_VGA_COLORS; i++) {
       color.pixel = i;
       if (i==15) {
         color.red   = 0xffff;
@@ -356,21 +431,13 @@ bx_printf("font_height = %u\n", (unsigned) font_height);
       XStoreColor(bx_x_display, default_cmap, &color);
       }
     }
-  else {
-    default_cmap = DefaultColormap(bx_x_display, bx_x_screen_num);
-    col_vals[0]  = BlackPixel(bx_x_display, bx_x_screen_num);
-    col_vals[15] = WhitePixel(bx_x_display, bx_x_screen_num);
-    for (i = 1; i < 256; i++) {
-      if (i==15) continue;
-      col_vals[i] = col_vals[0];
-      }
-    }
 
   // convenience variables which hold the black & white color indeces
   black_pixel = col_vals[0];
   white_pixel = col_vals[15];
 
-  bx_printf("default_depth = %d\n", default_depth);
+  BX_INFO(("font %u wide x %u high, display depth = %d",
+		(unsigned) font_width, (unsigned) font_height, default_depth));
 
   //select_visual();
 
@@ -406,13 +473,13 @@ bx_printf("font_height = %u\n", (unsigned) font_height);
    * XTextProperty structures and set their other
    * fields properly. */
   if (XStringListToTextProperty(&window_name, 1, &windowName) == 0) {
-    bx_panic("%s: structure allocation for windowName failed.\n",
-        progname);
+    BX_PANIC(("%s: structure allocation for windowName failed.",
+        progname));
   }
 
   if (XStringListToTextProperty(&icon_name, 1, &iconName) == 0) {
-    bx_panic("%s: structure allocation for iconName failed.\n",
-        progname);
+    BX_PANIC(("%s: structure allocation for iconName failed.",
+        progname));
   }
 
   wm_hints.initial_state = NormalState;
@@ -445,7 +512,7 @@ bx_printf("font_height = %u\n", (unsigned) font_height);
 
   XSetState(bx_x_display, gc, white_pixel, black_pixel, GXcopy,AllPlanes);
 
-  XSetState(bx_x_display, gc_inv, black_pixel, white_pixel, GXcopy,AllPlanes);
+  XSetState(bx_x_display, gc_inv, black_pixel, white_pixel, GXinvert,AllPlanes);
 
   XSetState(bx_x_display, gc_headerbar, black_pixel, white_pixel, GXcopy,AllPlanes);
 
@@ -456,12 +523,12 @@ bx_printf("font_height = %u\n", (unsigned) font_height);
   XMapWindow(bx_x_display, win);
   XSync(bx_x_display, /* no discard */ 0);
 
-  bx_printf("waiting for MapNotify\n");
+  BX_DEBUG(("waiting for MapNotify"));
   while (1) {
     XNextEvent(bx_x_display, &report);
     if (report.type == MapNotify) break;
     }
-  bx_printf("MapNotify found.\n");
+  BX_DEBUG(("MapNotify found."));
 
 {
   char *imagedata;
@@ -475,19 +542,19 @@ bx_printf("font_height = %u\n", (unsigned) font_height);
              32,                     // # bits of padding
              0 );                    // bytes_per_line, let X11 calculate
   if (!ximage)
-    bx_panic("vga: couldn't XCreateImage()\n");
+    BX_PANIC(("vga: couldn't XCreateImage()"));
 
   imDepth = default_depth;
   imWide  = ximage->bytes_per_line;
   imBPP   = ximage->bits_per_pixel;
 
   imagedata = (char *) malloc( (size_t) (ximage->bytes_per_line * y_tilesize) );
-  if (!imagedata) bx_panic("imagedata: malloc returned error\n");
+  if (!imagedata) BX_PANIC(("imagedata: malloc returned error"));
 
   ximage->data = imagedata;
 
   if (imBPP < imDepth) {
-    bx_panic("vga_x: bits_per_pixel < depth ?\n");
+    BX_PANIC(("vga_x: bits_per_pixel < depth ?"));
     }
 
 }
@@ -501,20 +568,48 @@ bx_printf("font_height = %u\n", (unsigned) font_height);
 
 
   XFlush(bx_x_display);
+
+  // loads keymap for x11
+  if(bx_options.keyboard.OuseMapping->get()) {
+    bx_keymap.loadKeymap(convertStringToXKeysym);
+    }
 }
 
 
+// This is called whenever the mouse_enabled parameter changes.  It
+// can change because of a gui event such as clicking on the mouse-enable
+// bitmap or pressing the middle button, or from the control panel.
+// In all those cases, setting the parameter value will get you here.
+  void
+bx_gui_c::mouse_enabled_changed_specific (Boolean val)
+{
+  BX_DEBUG (("mouse_enabled=%d, x11 specific code", val?1:0));
+  if (val) {
+    BX_INFO(("[x] Mouse on"));
+    mouse_enable_x = current_x;
+    mouse_enable_y = current_y;
+    disable_cursor();
+    // Move the cursor to a 'safe' place
+    warp_cursor(warp_home_x-current_x, warp_home_y-current_y);
+  } else {
+    BX_INFO(("[x] Mouse off"));
+    enable_cursor();
+    warp_cursor(mouse_enable_x-current_x, mouse_enable_y-current_y);
+  }
+}
+
   void
 load_font(void)
 {
-
   /* Load font and get font information structure. */
-  if ((font_info = XLoadQueryFont(bx_x_display,"vga")) == NULL) {
-    bx_printf("# %s: Cannot open vga font\n", progname);
-    bx_panic("Could not open vga font\n");
+  if ((font_info = XLoadQueryFont(bx_x_display,"bochsvga")) == NULL) {
+    if ((font_info = XLoadQueryFont(bx_x_display,"vga")) == NULL) {
+      if ((font_info = XLoadQueryFont(bx_x_display,"-*-vga-*")) == NULL) {
+      BX_PANIC(("Could not open vga font. See docs-html/install.html"));
     }
+  }
+}
 }
-
 
   void
 bx_gui_c::handle_events(void)
@@ -564,17 +659,17 @@ bx_gui_c::handle_events(void)
       break;
 
     case ConfigureNotify:
-//bx_printf("ConfigureNotify Xevent\n");
+      BX_DEBUG(("ConfigureNotify Xevent"));
       show_headerbar();
       break;
 
     case ButtonPress:
       button_event = (XButtonEvent *) &report;
-//bx_printf("xxx: buttonpress\n");
+		BX_DEBUG(("xxx: buttonpress"));
       if (button_event->y < BX_HEADER_BAR_Y) {
-//bx_printf("xxx:   in headerbar\n");
+		BX_DEBUG(("xxx:   in headerbar"));
         if (mouse_update) {
-//bx_printf("xxx:   mouse_update=1\n");
+		  BX_DEBUG(("xxx:   mouse_update=1"));
           send_keyboard_mouse_status();
           mouse_update = 0;
           }
@@ -586,38 +681,26 @@ bx_gui_c::handle_events(void)
       current_x = button_event->x;
       current_y = button_event->y;
       mouse_update = 1;
-//bx_printf("xxx:   x,y=(%d,%d)\n", current_x, current_y);
+	  BX_DEBUG(("xxx:   x,y=(%d,%d)", current_x, current_y));
       switch (button_event->button) {
         case Button1:
-//bx_printf("xxx:   button1\n");
+		  BX_DEBUG(("xxx:   button1"));
           mouse_button_state |= 0x01;
           send_keyboard_mouse_status();
           mouse_update = 0;
           break;
         case Button2:
-//bx_printf("xxx:   button2\n");
+	      BX_DEBUG(("XXX:   button2"));
 
 	      // (mch) Hack for easier mouse handling (toggle mouse enable)
-	      mouse_handler();
-	      if (bx_options.mouse_enabled) {
-		    bx_printf("[x] Mouse enabled\n");
-		    mouse_enable_x = current_x;
-		    mouse_enable_y = current_y;
-		    disable_cursor();
-		    // Move the cursor to a 'safe' place
-		    warp_cursor(warp_home_x-current_x, warp_home_y-current_y);
-	      } else {
-		    bx_printf("[x] Mouse disabled\n");
-		    enable_cursor();
-		    warp_cursor(mouse_enable_x-current_x, mouse_enable_y-current_y);
-	      }
+	      toggle_mouse_enable();
 
           //mouse_button_state |= ;
           //send_keyboard_mouse_status();
           //mouse_update = 0;
           break;
         case Button3:
-//bx_printf("xxx:   button3\n");
+		  BX_DEBUG(("xxx:   button3"));
           mouse_button_state |= 0x02;
           send_keyboard_mouse_status();
           mouse_update = 0;
@@ -627,11 +710,11 @@ bx_gui_c::handle_events(void)
 
     case ButtonRelease:
       button_event = (XButtonEvent *) &report;
-//bx_printf("xxx: buttonrelease\n");
+//BX_INFO(("xxx: buttonrelease"));
       if (button_event->y < BX_HEADER_BAR_Y) {
-//bx_printf("xxx:   in headerbar\n");
+//BX_INFO(("xxx:   in headerbar"));
         if (mouse_update) {
-//bx_printf("xxx:   mouse_update=1\n");
+//BX_INFO(("xxx:   mouse_update=1"));
           send_keyboard_mouse_status();
           mouse_update = 0;
           }
@@ -643,22 +726,22 @@ bx_gui_c::handle_events(void)
       current_x = button_event->x;
       current_y = button_event->y;
       mouse_update = 1;
-//bx_printf("xxx:   x,y=(%d,%d)\n", current_x, current_y);
+//BX_INFO(("xxx:   x,y=(%d,%d)", current_x, current_y));
       switch (button_event->button) {
         case Button1:
-//bx_printf("xxx:   button1\n");
+//BX_INFO(("xxx:   button1"));
           mouse_button_state &= ~0x01;
           send_keyboard_mouse_status();
           mouse_update = 0;
           break;
         case Button2:
-//bx_printf("xxx:   button2\n");
+//BX_INFO(("xxx:   button2"));
           //mouse_button_state &= ~;
           //send_keyboard_mouse_status();
           //mouse_update = 0;
           break;
         case Button3:
-//bx_printf("xxx:   button3\n");
+//BX_INFO(("xxx:   button3"));
           mouse_button_state &= ~0x02;
           send_keyboard_mouse_status();
           mouse_update = 0;
@@ -669,14 +752,12 @@ bx_gui_c::handle_events(void)
     case KeyPress:
       key_event = (XKeyEvent *) &report;
       charcount = XLookupString(key_event, buffer, bufsize, &keysym, &compose);
-/*bx_printf("# >>> %s\n", XKeysymToString(keysym));*/
       xkeypress(keysym, 0);
       break;
 
     case KeyRelease:
       key_event = (XKeyEvent *) &report;
       charcount = XLookupString(key_event, buffer, bufsize, &keysym, &compose);
-/*bx_printf("# <<< %s\n", XKeysymToString(keysym));*/
       xkeypress(keysym, 1);
       break;
 
@@ -685,34 +766,34 @@ bx_gui_c::handle_events(void)
       current_x = pointer_event->x;
       current_y = pointer_event->y;
       mouse_update = 1;
-//bx_printf("xxx: motionNotify x,y=(%d,%d)\n", current_x, current_y);
+//BX_INFO(("xxx: motionNotify x,y=(%d,%d)", current_x, current_y));
       break;
 
     case EnterNotify:
       enter_event = (XEnterWindowEvent *) &report;
       prev_x = current_x = enter_event->x;
       prev_y = current_y = enter_event->y;
-//bx_printf("xxx: enterNotify x,y=(%d,%d)\n", current_x, current_y);
+//BX_INFO(("xxx: enterNotify x,y=(%d,%d)", current_x, current_y));
       break;
 
     case LeaveNotify:
       leave_event = (XLeaveWindowEvent *) &report;
       prev_x = current_x = -1;
       prev_y = current_y = -1;
-//bx_printf("xxx: LeaveNotify x,y set to -1\n");
+//BX_INFO(("xxx: LeaveNotify x,y set to -1"));
       break;
 
     case MapNotify:
       /* screen needs redraw, since X would have tossed previous
        * requests before window mapped
        */
-//bx_printf("xxx: mapnotify: found\n");
+//BX_INFO(("xxx: mapnotify: found"));
       //retval = 1;
       break;
 
     default:
 	  // (mch) Ignore...
-	  // bx_printf("xxx: default Xevent type\n");
+	  BX_DEBUG(("XXX: default Xevent type"));
       /* all events selected by StructureNotifyMask are thrown away here,
        * since nothing is done with them */
       break;
@@ -720,7 +801,7 @@ bx_gui_c::handle_events(void)
   } /* end while */
 
   if (mouse_update) {
-    //bx_printf("xxx: bottom, send status\n");
+    BX_DEBUG(("XXX: bottom, send status"));
     send_keyboard_mouse_status();
     }
 }
@@ -729,8 +810,8 @@ bx_gui_c::handle_events(void)
   void
 send_keyboard_mouse_status(void)
 {
-//bx_printf("xxx: prev=(%d,%d) curr=(%d,%d)\n",
-//  prev_x, prev_y, current_x, current_y);
+	BX_DEBUG(("XXX: prev=(%d,%d) curr=(%d,%d)",
+			prev_x, prev_y, current_x, current_y));
 
   if ( (prev_x!=-1) && (current_x!=-1) && (prev_y!=-1) && (current_y!=-1)) {
     int dx, dy;
@@ -740,7 +821,7 @@ send_keyboard_mouse_status(void)
     dy = -(current_y - prev_y - warp_dy);
     warp_cursor(warp_home_x-current_x, warp_home_y-current_y);
 
-//bx_printf("xxx: MOUSE_MOTION: dx=%d, dy=%d\n", (int) dx, (int) dy);
+//BX_INFO(("xxx: MOUSE_MOTION: dx=%d, dy=%d", (int) dx, (int) dy));
     pluginMouseMotion( dx, dy, mouse_button_state);
     //if (warped) {
     //  prev_x = current_x = -1;
@@ -775,133 +856,145 @@ xkeypress(KeySym keysym, int press_release)
 {
   Bit32u key_event;
 
+  /* Old (no mapping) behavior */
+  if(!bx_options.keyboard.OuseMapping->get()) {
 
-
-
-
-  // this depends on the fact that the X11 keysyms which
-  // correspond to the ascii characters space .. tilde
-  // are in consequtive order.
-  if ((keysym >= XK_space) && (keysym <= XK_asciitilde)) {
-    key_event = ascii_to_key_event[keysym - XK_space];
-    }
-  else switch (keysym) {
-    case XK_KP_1:
+    // this depends on the fact that the X11 keysyms which
+    // correspond to the ascii characters space .. tilde
+    // are in consequtive order.
+    if ((keysym >= XK_space) && (keysym <= XK_asciitilde)) {
+      key_event = ascii_to_key_event[keysym - XK_space];
+      }
+    else switch (keysym) {
+      case XK_KP_1:
 #ifdef XK_KP_End
-    case XK_KP_End:
+      case XK_KP_End:
 #endif
-      key_event = BX_KEY_KP_END; break;
+        key_event = BX_KEY_KP_END; break;
 
-    case XK_KP_2:
+      case XK_KP_2:
 #ifdef XK_KP_Down
-    case XK_KP_Down:
+      case XK_KP_Down:
 #endif
-      key_event = BX_KEY_KP_DOWN; break;
+        key_event = BX_KEY_KP_DOWN; break;
 
-    case XK_KP_3:
+      case XK_KP_3:
 #ifdef XK_KP_Page_Down
-    case XK_KP_Page_Down:
+      case XK_KP_Page_Down:
 #endif
-      key_event = BX_KEY_KP_PAGE_DOWN; break;
+        key_event = BX_KEY_KP_PAGE_DOWN; break;
 
-    case XK_KP_4:
+      case XK_KP_4:
 #ifdef XK_KP_Left
-    case XK_KP_Left:
+      case XK_KP_Left:
 #endif
-      key_event = BX_KEY_KP_LEFT; break;
+        key_event = BX_KEY_KP_LEFT; break;
 
-    case XK_KP_5:
-      key_event = BX_KEY_KP_5; break;
+      case XK_KP_5:
+#ifdef XK_KP_Begin
+      case XK_KP_Begin:
+#endif
+        key_event = BX_KEY_KP_5; break;
 
-    case XK_KP_6:
+      case XK_KP_6:
 #ifdef XK_KP_Right
-    case XK_KP_Right:
+      case XK_KP_Right:
 #endif
-      key_event = BX_KEY_KP_RIGHT; break;
+        key_event = BX_KEY_KP_RIGHT; break;
 
-    case XK_KP_7:
+      case XK_KP_7:
 #ifdef XK_KP_Home
-    case XK_KP_Home:
+      case XK_KP_Home:
 #endif
-      key_event = BX_KEY_KP_HOME; break;
+        key_event = BX_KEY_KP_HOME; break;
 
-    case XK_KP_8:
+      case XK_KP_8:
 #ifdef XK_KP_Up
-    case XK_KP_Up:
+      case XK_KP_Up:
 #endif
-      key_event = BX_KEY_KP_UP; break;
+        key_event = BX_KEY_KP_UP; break;
 
-    case XK_KP_9:
+      case XK_KP_9:
 #ifdef XK_KP_Page_Up
-    case XK_KP_Page_Up:
+      case XK_KP_Page_Up:
 #endif
-      key_event = BX_KEY_KP_PAGE_UP; break;
+        key_event = BX_KEY_KP_PAGE_UP; break;
 
-    case XK_KP_0:
+      case XK_KP_0:
 #ifdef XK_KP_Insert
-    case XK_KP_Insert:
-      key_event = BX_KEY_KP_INSERT; break;
+      case XK_KP_Insert:
 #endif
+        key_event = BX_KEY_KP_INSERT; break;
 
-    case XK_KP_Decimal:
+      case XK_KP_Decimal:
 #ifdef XK_KP_Delete
-    case XK_KP_Delete:
-      key_event = BX_KEY_KP_DELETE; break;
+      case XK_KP_Delete:
 #endif
+        key_event = BX_KEY_KP_DELETE; break;
 
 #ifdef XK_KP_Enter
-    case XK_KP_Enter:
-      key_event = BX_KEY_KP_ENTER; break;
+      case XK_KP_Enter:
+        key_event = BX_KEY_KP_ENTER; break;
 #endif
 
-    case XK_KP_Subtract: key_event = BX_KEY_KP_SUBTRACT; break;
-    case XK_KP_Add:      key_event = BX_KEY_KP_ADD; break;
+      case XK_KP_Subtract: key_event = BX_KEY_KP_SUBTRACT; break;
+      case XK_KP_Add:      key_event = BX_KEY_KP_ADD; break;
 
-    case XK_KP_Multiply: key_event = BX_KEY_KP_MULTIPLY; break;
-    case XK_KP_Divide:   key_event = BX_KEY_KP_DIVIDE; break;
+      case XK_KP_Multiply: key_event = BX_KEY_KP_MULTIPLY; break;
+      case XK_KP_Divide:   key_event = BX_KEY_KP_DIVIDE; break;
 
 
-    case XK_Up:          key_event = BX_KEY_UP; break;
-    case XK_Down:        key_event = BX_KEY_DOWN; break;
-    case XK_Left:        key_event = BX_KEY_LEFT; break;
-    case XK_Right:       key_event = BX_KEY_RIGHT; break;
+      case XK_Up:          key_event = BX_KEY_UP; break;
+      case XK_Down:        key_event = BX_KEY_DOWN; break;
+      case XK_Left:        key_event = BX_KEY_LEFT; break;
+      case XK_Right:       key_event = BX_KEY_RIGHT; break;
 
 
-    case XK_Delete:      key_event = BX_KEY_DELETE; break;
-    case XK_BackSpace:   key_event = BX_KEY_BACKSPACE; break;
-    case XK_Tab:         key_event = BX_KEY_TAB; break;
-    case XK_Return:      key_event = BX_KEY_ENTER; break;
-    case XK_Escape:      key_event = BX_KEY_ESC; break;
-    case XK_F1:          key_event = BX_KEY_F1; break;
-    case XK_F2:          key_event = BX_KEY_F2; break;
-    case XK_F3:          key_event = BX_KEY_F3; break;
-    case XK_F4:          key_event = BX_KEY_F4; break;
-    case XK_F5:          key_event = BX_KEY_F5; break;
-    case XK_F6:          key_event = BX_KEY_F6; break;
-    case XK_F7:          key_event = BX_KEY_F7; break;
-    case XK_F8:          key_event = BX_KEY_F8; break;
-    case XK_F9:          key_event = BX_KEY_F9; break;
-    case XK_F10:         key_event = BX_KEY_F10; break;
-    case XK_F11:         key_event = BX_KEY_F11; break;
-    case XK_F12:         key_event = BX_KEY_F12; break;
-    case XK_Control_L:   key_event = BX_KEY_CTRL_L; break;
-    case XK_Shift_L:     key_event = BX_KEY_SHIFT_L; break;
-    case XK_Shift_R:     key_event = BX_KEY_SHIFT_R; break;
-    case XK_Caps_Lock:   key_event = BX_KEY_CAPS_LOCK; break;
-    case XK_Num_Lock:    key_event = BX_KEY_NUM_LOCK; break;
-    case XK_Alt_L:       key_event = BX_KEY_ALT_L; break;
+      case XK_Delete:      key_event = BX_KEY_DELETE; break;
+      case XK_BackSpace:   key_event = BX_KEY_BACKSPACE; break;
+      case XK_Tab:         key_event = BX_KEY_TAB; break;
+      case XK_Return:      key_event = BX_KEY_ENTER; break;
+      case XK_Escape:      key_event = BX_KEY_ESC; break;
+      case XK_F1:          key_event = BX_KEY_F1; break;
+      case XK_F2:          key_event = BX_KEY_F2; break;
+      case XK_F3:          key_event = BX_KEY_F3; break;
+      case XK_F4:          key_event = BX_KEY_F4; break;
+      case XK_F5:          key_event = BX_KEY_F5; break;
+      case XK_F6:          key_event = BX_KEY_F6; break;
+      case XK_F7:          key_event = BX_KEY_F7; break;
+      case XK_F8:          key_event = BX_KEY_F8; break;
+      case XK_F9:          key_event = BX_KEY_F9; break;
+      case XK_F10:         key_event = BX_KEY_F10; break;
+      case XK_F11:         key_event = BX_KEY_F11; break;
+      case XK_F12:         key_event = BX_KEY_F12; break;
+      case XK_Control_L:   key_event = BX_KEY_CTRL_L; break;
+      case XK_Shift_L:     key_event = BX_KEY_SHIFT_L; break;
+      case XK_Shift_R:     key_event = BX_KEY_SHIFT_R; break;
+      case XK_Caps_Lock:   key_event = BX_KEY_CAPS_LOCK; break;
+      case XK_Num_Lock:    key_event = BX_KEY_NUM_LOCK; break;
+      case XK_Alt_L:       key_event = BX_KEY_ALT_L; break;
 
-    case XK_Insert:      key_event = BX_KEY_INSERT; break;
-    case XK_Home:        key_event = BX_KEY_HOME; break;
-    case XK_End:         key_event = BX_KEY_END; break;
-    case XK_Page_Up:     key_event = BX_KEY_PAGE_UP; break;
-    case XK_Page_Down:   key_event = BX_KEY_PAGE_DOWN; break;
+      case XK_Insert:      key_event = BX_KEY_INSERT; break;
+      case XK_Home:        key_event = BX_KEY_HOME; break;
+      case XK_End:         key_event = BX_KEY_END; break;
+      case XK_Page_Up:     key_event = BX_KEY_PAGE_UP; break;
+      case XK_Page_Down:   key_event = BX_KEY_PAGE_DOWN; break;
 
-    default:
-      bx_printf("# xkeypress(): keysym %x unhandled!\n", (unsigned) keysym);
-      return;
+      default:
+        BX_ERROR(( "xkeypress(): keysym %x unhandled!", (unsigned) keysym ));
+        return;
       break;
+      }
     }
+  else {
+   /* use mapping */
+   BXKeyEntry *entry = bx_keymap.getKeyXwin (keysym);
+   if (!entry) {
+     BX_ERROR(( "xkeypress(): keysym %x unhandled!", (unsigned) keysym ));
+     return;
+   }
+   key_event = entry->baseKey;
+ }
 
   if (press_release)
     key_event |= BX_KEY_RELEASED;
@@ -922,27 +1015,32 @@ bx_gui_c::clear_screen(void)
   void
 bx_gui_c::text_update(Bit8u *old_text, Bit8u *new_text,
                       unsigned long cursor_x, unsigned long cursor_y,
-                      unsigned nrows)
+                      Bit16u cursor_state, unsigned nrows)
 {
-  int font_height;
   unsigned i, x, y, curs;
   unsigned new_foreground, new_background;
   Bit8u string[1];
+  Bit8u cs_start, cs_end;
   unsigned nchars;
 
-  font_height = font_info->ascent + font_info->descent;
+  UNUSED(nrows);
+
+  cs_start = ((cursor_state >> 8) & 0x3f) * font_height / font_height_orig;
+  cs_end = (cursor_state & 0x1f) * font_height / font_height_orig;
 
   // Number of characters on screen, variable number of rows
-  nchars = 80*nrows;
+  nchars = columns*rows;
 
   // first draw over character at original block cursor location
-  if ( (prev_block_cursor_y*80 + prev_block_cursor_x) < nchars ) {
-    curs = (prev_block_cursor_y*80 + prev_block_cursor_x)*2;
+  if ( (prev_block_cursor_y*columns + prev_block_cursor_x) < nchars ) {
+    curs = (prev_block_cursor_y*columns + prev_block_cursor_x)*2;
     string[0] = new_text[curs];
+    if (string[0] == 0) string[0] = ' '; // convert null to space
+    XSetForeground(bx_x_display, gc, col_vals[new_text[curs+1] & 0x0f]);
     XSetBackground(bx_x_display, gc, col_vals[(new_text[curs+1] & 0xf0) >> 4]);
     XDrawImageString(bx_x_display, win,
       gc,
-      prev_block_cursor_x * font_info->max_bounds.width,
+      prev_block_cursor_x * font_width,
       prev_block_cursor_y * font_height + font_info->max_bounds.ascent + bx_headerbar_y,
       (char *) string,
       1);
@@ -963,12 +1061,12 @@ bx_gui_c::text_update(Bit8u *old_text, Bit8u *new_text,
 //XSetForeground(bx_x_display, gc, white_pixel);
 //XSetBackground(bx_x_display, gc, black_pixel);
 
-      x = (i/2) % 80;
-      y = (i/2) / 80;
+      x = (i/2) % columns;
+      y = (i/2) / columns;
 
       XDrawImageString(bx_x_display, win,
         gc,
-        x * font_info->max_bounds.width,
+        x * font_width,
         y * font_height + font_info->max_bounds.ascent + bx_headerbar_y,
         (char *) string,
         1);
@@ -982,19 +1080,37 @@ bx_gui_c::text_update(Bit8u *old_text, Bit8u *new_text,
   XSetBackground(bx_x_display, gc, black_pixel);
 
   // now draw character at new block cursor location in reverse
-  if ( (cursor_y*80 + cursor_x) < nchars ) {
-    string[0] = new_text[(cursor_y*80 + cursor_x)*2];
-    XDrawImageString(bx_x_display, win,
-      gc_inv,
-      cursor_x * font_info->max_bounds.width,
-      cursor_y * font_height + font_info->max_bounds.ascent + bx_headerbar_y,
-      (char *) string,
-      1);
+  if ( ( (cursor_y*columns + cursor_x) < nchars ) && (cs_start <= cs_end) ) {
+    for (unsigned i = cs_start; i <= cs_end; i++)
+      XDrawLine(bx_x_display, win,
+	gc_inv,
+	cursor_x * font_width,
+	cursor_y * font_height + bx_headerbar_y + i,
+	(cursor_x + 1) * font_width - 1,
+	cursor_y * font_height + bx_headerbar_y + i
+      );
     }
 
   XFlush(bx_x_display);
 }
 
+  int
+bx_gui_c::get_clipboard_text(Bit8u **bytes, Bit32s *nbytes)
+{
+  *bytes = (Bit8u *)XFetchBytes (bx_x_display, nbytes);
+  return 1;
+}
+
+  int
+bx_gui_c::set_clipboard_text(char *text_snapshot, Bit32u len)
+{
+  // this writes data to the clipboard.
+  BX_INFO (("storing %d bytes to X windows clipboard", len));
+  XSetSelectionOwner(bx_x_display, XA_PRIMARY, None, CurrentTime);
+  XStoreBytes (bx_x_display, (char *)text_snapshot, len);
+  return 1;
+}
+
 
   void
 bx_gui_c::graphics_tile_update(Bit8u *tile, unsigned x0, unsigned y0)
@@ -1063,8 +1179,8 @@ bx_gui_c::graphics_tile_update(Bit8u *tile, unsigned x0, unsigned y0)
             }
           break;
         default:
-          bx_panic("X_graphics_tile_update: bits_per_pixel %u not implemented\n",
-            (unsigned) imBPP);
+          BX_PANIC(("X_graphics_tile_update: bits_per_pixel %u not implemented",
+            (unsigned) imBPP));
           break;
         }
       }
@@ -1086,7 +1202,7 @@ bx_gui_c::palette_change(unsigned index, unsigned red, unsigned green, unsigned
   color.green = green << 8;
   color.blue  = blue << 8;
 
-  if (bx_options.private_colormap) {
+  if (bx_options.Oprivate_colormap->get ()) {
     color.pixel = index;
     XStoreColor(bx_x_display, default_cmap, &color);
     return(0); // no screen update needed
@@ -1101,8 +1217,19 @@ bx_gui_c::palette_change(unsigned index, unsigned red, unsigned green, unsigned
 
 
   void
-bx_gui_c::dimension_update(unsigned x, unsigned y)
+bx_gui_c::dimension_update(unsigned x, unsigned y, unsigned fheight)
 {
+  if (fheight > 0) {
+    font_height_orig = fheight;
+    rows = y / fheight;
+    columns = x / 8;
+    if (fheight != font_height) {
+      y = rows * font_height;
+    }
+    if (font_width != 8) {
+      x = columns * font_width;
+    }
+  }
   if ( (x != dimension_x) || (y != (dimension_y-bx_headerbar_y)) ) {
     XSizeHints hints;
     long supplied_return;
@@ -1124,15 +1251,24 @@ bx_gui_c::dimension_update(unsigned x, unsigned y)
 bx_gui_c::show_headerbar(void)
 {
   unsigned xorigin;
+  int xleft, xright;
 
   // clear header bar area to white
   XFillRectangle(bx_x_display, win, gc_headerbar_inv, 0,0, dimension_x, bx_headerbar_y);
 
+  xleft = 0;
+  xright = dimension_x;
+
   for (unsigned i=0; i<bx_headerbar_entries; i++) {
-    if (bx_headerbar_entry[i].alignment == BX_GRAVITY_LEFT)
+    if (bx_headerbar_entry[i].alignment == BX_GRAVITY_LEFT) {
       xorigin = bx_headerbar_entry[i].xorigin;
-    else
+      xleft += bx_headerbar_entry[i].xdim;
+      }
+    else {
       xorigin = dimension_x - bx_headerbar_entry[i].xorigin;
+      xright = xorigin;
+      }
+    if (xright < xleft) break;
     XCopyPlane(bx_x_display, bx_headerbar_entry[i].bitmap, win, gc_headerbar,
       0,0, bx_headerbar_entry[i].xdim, bx_headerbar_entry[i].ydim,
               xorigin, 0, 1);
@@ -1144,7 +1280,7 @@ bx_gui_c::show_headerbar(void)
 bx_gui_c::create_bitmap(const unsigned char *bmap, unsigned xdim, unsigned ydim)
 {
   if (bx_bitmap_entries >= BX_MAX_PIXMAPS) {
-    bx_panic("x: too many pixmaps, increase BX_MAX_PIXMAPS\n");
+    BX_PANIC(("x: too many pixmaps, increase BX_MAX_PIXMAPS"));
     }
 
   bx_bitmaps[bx_bitmap_entries].bmap =
@@ -1152,7 +1288,7 @@ bx_gui_c::create_bitmap(const unsigned char *bmap, unsigned xdim, unsigned ydim)
   bx_bitmaps[bx_bitmap_entries].xdim = xdim;
   bx_bitmaps[bx_bitmap_entries].ydim = ydim;
   if (!bx_bitmaps[bx_bitmap_entries].bmap) {
-    bx_panic("x: could not create bitmap\n");
+    BX_PANIC(("x: could not create bitmap"));
     }
   bx_bitmap_entries++;
   return(bx_bitmap_entries-1); // return index as handle
@@ -1165,7 +1301,7 @@ bx_gui_c::headerbar_bitmap(unsigned bmap_id, unsigned alignment, void (*f)(void)
   unsigned hb_index;
 
   if ( (bx_headerbar_entries+1) > BX_MAX_HEADERBAR_ENTRIES )
-    bx_panic("x: too many headerbar entries, increase BX_MAX_HEADERBAR_ENTRIES\n");
+    BX_PANIC(("x: too many headerbar entries, increase BX_MAX_HEADERBAR_ENTRIES"));
 
   bx_headerbar_entries++;
   hb_index = bx_headerbar_entries - 1;
@@ -1227,12 +1363,15 @@ headerbar_click(int x, int y)
   void
 bx_gui_c::exit(void)
 {
-  bx_printf("Note: X11: bx_gui_c::exit() not implemented yet.\n");
+  XCloseDisplay (bx_x_display);
+  BX_INFO(("Exit."));
 }
 
 static void warp_cursor (int dx, int dy)
 {
-      if (bx_options.mouse_enabled) {
+      if (bx_options.Omouse_enabled->get () &&
+	  (warp_dx || warp_dy || dx || dy)
+	  ) {
 	    warp_dx = dx;
 	    warp_dy = dy;
 	    XWarpPointer(bx_x_display, None, None, 0, 0, 0, 0, dx, dy);
@@ -1287,3 +1426,92 @@ static void enable_cursor ()
 {
       XUndefineCursor(bx_x_display, win);
 }
+
+/* convertStringToXKeysym is a keymap callback
+ * used when reading the keymap file.
+ * It converts a Symblic String to a GUI Constant
+ *
+ * It returns a Bit32u constant or BX_KEYMAP_UNKNOWN if it fails
+ */
+static Bit32u convertStringToXKeysym (const char *string)
+{
+    KeySym keysym=XStringToKeysym(string);
+
+    // failure, return unknown
+    if(keysym==NoSymbol) return BX_KEYMAP_UNKNOWN;
+
+    return((Bit32u)keysym);
+}
+
+#if BX_USE_IDLE_HACK
+
+/* BX_USE_IDLE_HACK: a small idle hack by
+ * Roland.Mainz@informatik.med.uni-giessen.de to prevent bochs
+ * from consuming 100% CPU time even when it is not required (for
+ * example, the OS in the emulator calls HLT to wait for an interupt)
+ * pro:
+ * - no more 100% CPU usage
+ * contra:
+ * - we're sleeping too long
+ * - bochs still consumes ~10%-20% CPU time while executing an idle 
+ *   linux kernel
+ * - this is an hack
+ */
+
+/* XPeekEvent() with timeout 
+ * (adopted from mozilla/gfx/src/xprint/xprintutil_printtofile.c#XNextEventTimeout())
+ */
+static
+Bool XPeekEventTimeout( Display *display, XEvent *event_return, struct timeval *timeout ) 
+{
+    int      res;
+    fd_set   readfds;
+    int      display_fd = XConnectionNumber(display);
+
+    /* small shortcut... */
+    if( timeout == NULL )
+    {
+      XPeekEvent(display, event_return);
+      return(True);
+    }
+    
+    FD_ZERO(&readfds);
+    FD_SET(display_fd, &readfds);
+
+    /* Note/bug: In the case of internal X events (like used to trigger callbacks 
+     * registered by XpGetDocumentData()&co.) select() will return with "new info" 
+     * - but XNextEvent() below processes these _internal_ events silently - and 
+     * will block if there are no other non-internal events.
+     * The workaround here is to check with XEventsQueued() if there are non-internal 
+     * events queued - if not select() will be called again - unfortunately we use 
+     * the old timeout here instead of the "remaining" time... (this only would hurt 
+     * if the timeout would be really long - but for current use with values below
+     * 1/2 secs it does not hurt... =:-)
+     */
+    while( XEventsQueued(display, QueuedAfterFlush) == 0 )
+    {
+      res = select(display_fd+1, &readfds, NULL, NULL, timeout);
+    
+      switch(res)
+      {
+        case -1: /* select() error - should not happen */ 
+            perror("XPeekEventTimeout: select() failure"); 
+            return(False);
+        case  0: /* timeout */
+          return(False);
+      }
+    }
+    
+    XPeekEvent(display, event_return); 
+    return(True);
+}
+
+
+void bx_gui_c::sim_is_idle () {
+  XEvent dummy;
+  struct timeval   timeout;   
+  timeout.tv_sec  = 0;
+  timeout.tv_usec = 1000; /* 1/1000 s */  
+  XPeekEventTimeout(bx_x_display, &dummy, &timeout);
+}
+#endif /* BX_USE_IDLE_HACK */  
