@@ -1,4 +1,7 @@
-/*
+/************************************************************************
+ * $Id: plugin.c,v 1.2 2002/07/21 23:14:55 DemonLord Exp $
+ ************************************************************************
+ *
  *  plex86: run multiple x86 operating systems concurrently
  *  Copyright (C) 1999-2000  The plex86 developers team
  *
@@ -29,7 +32,7 @@
 #include "plex86.h"
 #include "user.h"
 #include "plugin.h"
-
+#include "bxemu.h"
 
 
 /************************************************************************/
@@ -38,6 +41,20 @@
 
 static void plugin_init_one(plugin_t *plugin);
 
+/* Pointers to IO register function, used to implement interception to IO calls.
+ * Debugging feature.
+ * Following are the real register functions.
+ */
+int (*pluginRegisterIOReadHandler)(void *thisPtr, ioReadHandler_t callback,
+                                   unsigned base, const char *name, unsigned len);
+int (*pluginRegisterIOWriteHandler)(void *thisPtr, ioWriteHandler_t callback,
+                                    unsigned base, const char *name, unsigned len);
+
+int pluginRegisterIOReadHandlerReal(void *thisPtr, ioReadHandler_t callback,
+                                    unsigned base, const char *name, unsigned len);
+int pluginRegisterIOWriteHandlerReal(void *thisPtr, ioWriteHandler_t callback,
+                                     unsigned base, const char *name, unsigned len);
+
 /* Some builtin (default pass-through) Write-Cache functions, when
  * no Write-Cache plugins are loaded to override these.  These are
  * strict pass-through functions
@@ -74,6 +91,13 @@ unsigned (* pluginRegisterDMAChannel)(
                 void (* dma_read)(Bit8u *data_byte),
                 void (* dma_write)(Bit8u *data_byte)
                 ) = 0;
+unsigned (* pluginRegisterDMA16Channel)(
+                unsigned channel,
+                void (* dma_read)(Bit16u *data_word),
+                void (* dma_write)(Bit16u *data_word)
+                ) = 0;
+unsigned (* pluginUnregisterDMAChannel)(unsigned channel) = 0;
+
 void     (* pluginDMASetDRQ)(unsigned channel, unsigned val) = 0;
 unsigned (* pluginDMAGetTC)(void) = 0;
 
@@ -85,7 +109,6 @@ void     (* pluginVGARedrawArea)(unsigned x0, unsigned y0,
                 unsigned width, unsigned height) = 0;
 
 bx_debug_t   bx_dbg; // xxx Fix me, bochs hack
-bx_options_t bx_options; // xxx Fix me, bochs hack
 
 // xxx Fix me, bochs hack
   void
@@ -309,7 +332,7 @@ plugin_fini_all (void)
 
 
   int
-pluginRegisterIOReadHandler(void *thisPtr, ioReadHandler_t callback,
+pluginRegisterIOReadHandlerReal(void *thisPtr, ioReadHandler_t callback,
                             unsigned base, const char *name, unsigned len)
 {
   int id;
@@ -332,7 +355,7 @@ pluginRegisterIOReadHandler(void *thisPtr, ioReadHandler_t callback,
 
 
   int
-pluginRegisterIOWriteHandler(void *thisPtr, ioWriteHandler_t callback,
+pluginRegisterIOWriteHandlerReal(void *thisPtr, ioWriteHandler_t callback,
                              unsigned base, const char *name, unsigned len)
 {
   int id;
@@ -354,22 +377,14 @@ pluginRegisterIOWriteHandler(void *thisPtr, ioWriteHandler_t callback,
 }
 
 
-#if 0
 #define BX_MAX_IRQS 16
 #define BX_NO_IRQ  -1
 const char *irq_handler_name[BX_MAX_IRQS];
 
-  for (i=0; i < BX_MAX_IRQS; i++) {
-    irq_handler_name[i] = NULL;
-    }
-#endif
-
   void
 pluginRegisterIRQ(unsigned irq, const char *name)
 {
-#warning "Finish pluginRegisterIRQ"
 
-#if 0
   if (irq >= BX_MAX_IRQS) {
     bx_panic("IO device %s registered with IRQ=%d above %u\n",
              name, irq, (unsigned) BX_MAX_IRQS-1);
@@ -379,14 +394,11 @@ pluginRegisterIRQ(unsigned irq, const char *name)
       irq_handler_name[irq], name);
     }
   irq_handler_name[irq] = name;
-#endif
 }
 
   void
 pluginUnregisterIRQ(unsigned irq, const char *name)
 {
-#warning "Finish pluginUnregisterIRQ"
-#if 0
   if (irq >= BX_MAX_IRQS) {
     bx_panic("IO device %s tried to unregister IRQ %d above %u\n",
              name, irq, (unsigned) BX_MAX_IRQS-1);
@@ -404,7 +416,6 @@ pluginUnregisterIRQ(unsigned irq, const char *name)
     return;
     }
   irq_handler_name[irq] = NULL;
-#endif
 }
 
 
@@ -597,6 +608,7 @@ plugin_handle_periodic(void)
 plugin_startup(void)
 {
   ZERO_ARRAY(ints);
+  ZERO_ARRAY(irq_handler_name);
 
   /* Register builtin Write-Cache functions, so we have some defaults
    * until overloaded by plugins
@@ -609,6 +621,8 @@ plugin_startup(void)
   pluginMouseMotion  = builtinMouseMotion;
   pluginGenScancode  = builtinGenScancode;
   pluginRegisterDMAChannel = builtinRegisterDMAChannel;
+  pluginRegisterDMA16Channel = builtinRegisterDMA16Channel;
+  pluginUnregisterDMAChannel = builtinUnregisterDMAChannel;
   pluginDMASetDRQ    = builtinDMASetDRQ;
   pluginDMAGetTC     = builtinDMAGetTC;
 
@@ -616,6 +630,13 @@ plugin_startup(void)
   pluginHDWriteHandler = builtinHDWriteHandler;
 
   pluginVGARedrawArea  = builtinVGARedrawArea;
+  
+  pluginRegisterIRQ(2, "cascade");
+
+  pluginRegisterIOReadHandler = pluginRegisterIOReadHandlerReal;
+  pluginRegisterIOWriteHandler = pluginRegisterIOWriteHandlerReal;
+  
+  bxemu_init();
 }
 
   Bit32u
@@ -669,6 +690,28 @@ builtinRegisterDMAChannel(
   return 0;
 }
 
+  unsigned
+builtinRegisterDMA16Channel(
+  unsigned channel,
+  void (* dma_read)(Bit16u *data_word),
+  void (* dma_write)(Bit16u *data_word)
+  )
+{
+  fprintf(stderr, "builtinRegisterDMA16Channel called, not overloaded by DMA"
+                  " plugin?\n");
+  vm_abort();
+  return 0;
+}
+
+  unsigned
+builtinUnregisterDMAChannel(unsigned channel)
+{
+  fprintf(stderr, "builtinUnregisterDMAChannel called, not overloaded by DMA"
+                  " plugin?\n");
+  vm_abort();
+  return 0;
+}
+
   void
 builtinDMASetDRQ(unsigned channel, unsigned val)
 {
@@ -747,6 +790,9 @@ int plugin_emulate_outport(int port, int op_size, int count, void *loc)
   vm_abort();
 }
 
+// old IRQ handling routines (disabled)
+
+#if 0
   void
 pluginTriggerIRQ(unsigned irq)
 {
@@ -778,6 +824,49 @@ pluginUntriggerIRQ(unsigned irq)
     vm_abort();
     }
 }
+#endif
+
+// new IRQ handling routines
+
+  void
+pluginRaiseIRQ(unsigned irq)
+{
+  int ret;
+  irqMsg_t irqMsg;
+
+  irqMsg.irq = irq;
+  irqMsg.on  = 1;
+
+  if (!irq_handler_name[irq]) {
+    bx_panic("IRQ line %u not registered", irq);
+    vm_abort();
+    }
+  ret = ioctl(filehdl, PLEX86_IRQ, &irqMsg);
+  if (ret == -1) {
+    perror ("ioctl PLEX86_IRQ: ");
+    vm_abort();
+    }
+}
+
+  void
+pluginLowerIRQ(unsigned irq)
+{
+  int ret;
+  irqMsg_t irqMsg;
+
+  irqMsg.irq = irq;
+  irqMsg.on  = 0;
+
+  if (!irq_handler_name[irq]) {
+    bx_panic("IRQ line %u not registered", irq);
+    vm_abort();
+    }
+  ret = ioctl(filehdl, PLEX86_IRQ, &irqMsg);
+  if (ret == -1) {
+    perror ("ioctl PLEX86_IRQ: ");
+    vm_abort();
+    }
+}
 
   void
 pluginSetHRQ(unsigned val)
