@@ -1,4 +1,10 @@
-//  Copyright (C) 2000  MandrakeSoft S.A.
+/////////////////////////////////////////////////////////////////////////
+// $Id: ne2k.cc,v 1.2 2002/07/21 23:14:55 DemonLord Exp $
+/////////////////////////////////////////////////////////////////////////
+// Based on Bochs file <Id: ne2k.cc,v 1.34 2002/05/03 13:52:53 cbothamy Exp >
+/////////////////////////////////////////////////////////////////////////
+//
+//  Copyright (C) 2002  MandrakeSoft S.A.
 //
 //    MandrakeSoft S.A.
 //    43, rue d'Aboukir
@@ -25,15 +31,23 @@
 
 
 #include "bochs.h"
+#include "system.h"
+#include "eth.h"
+#include "ne2k.h"
+#define LOG_THIS bx_ne2k.
+
+bx_ne2k_c bx_ne2k;
 
 #if BX_USE_NE2K_SMF
-bx_ne2k_c bx_ne2k;
-#define this NULL
+#define this (&bx_ne2k)
 #endif
 
 bx_ne2k_c::bx_ne2k_c(void)
 {
-  // nothing for now
+	put("NE2K");
+	settype(NE2KLOG);
+	BX_DEBUG(("Init $Id: ne2k.cc,v 1.2 2002/07/21 23:14:55 DemonLord Exp $"));
+	// nothing for now
 }
 
 
@@ -48,6 +62,7 @@ bx_ne2k_c::~bx_ne2k_c(void)
 void
 bx_ne2k_c::reset_device(void)
 {
+  BX_DEBUG (("reset_device"));
   // Zero out registers and memory
   memset( & BX_NE2K_THIS s.CR,  0, sizeof(BX_NE2K_THIS s.CR) );
   memset( & BX_NE2K_THIS s.ISR, 0, sizeof(BX_NE2K_THIS s.ISR));
@@ -84,7 +99,7 @@ bx_ne2k_c::reset_device(void)
   
   // Set power-up conditions
   BX_NE2K_THIS s.CR.stop      = 1;
-  BX_NE2K_THIS s.CR.rdma_cmd  = 1;
+    BX_NE2K_THIS s.CR.rdma_cmd  = 4;
   BX_NE2K_THIS s.ISR.reset    = 1;
   BX_NE2K_THIS s.DCR.longaddr = 1;
 }
@@ -96,20 +111,25 @@ bx_ne2k_c::reset_device(void)
 Bit32u
 bx_ne2k_c::read_cr(void)
 {
-  return (((BX_NE2K_THIS s.CR.pgsel    & 0x03) << 6) |
+  Bit32u val = 
+         (((BX_NE2K_THIS s.CR.pgsel    & 0x03) << 6) |
 	  ((BX_NE2K_THIS s.CR.rdma_cmd & 0x07) << 3) |
 	  (BX_NE2K_THIS s.CR.tx_packet << 2) |
 	  (BX_NE2K_THIS s.CR.start     << 1) |
 	  (BX_NE2K_THIS s.CR.stop));
+  BX_DEBUG(("read CR returns 0x%08x", val));
+  return val;
 }
 
 void
 bx_ne2k_c::write_cr(Bit32u value)
 {
+    BX_DEBUG(("wrote 0x%02x to CR", value));
   // Validate remote-DMA
   if ((value & 0x38) == 0x00)
-    bx_panic("ne2k: CR write - invalide rDMA value 0");
-  
+    return;
+  //  BX_PANIC(("CR write - invalid rDMA value 0"));
+
   // Check for s/w reset
   if (value & 0x01) {
     BX_NE2K_THIS s.ISR.reset = 1;
@@ -129,34 +149,49 @@ bx_ne2k_c::write_cr(Bit32u value)
   BX_NE2K_THIS s.CR.start = ((value & 0x02) == 0x02);
   BX_NE2K_THIS s.CR.pgsel = (value & 0xc0) >> 6;
 
+    // Check for send-packet command
+    if (BX_NE2K_THIS s.CR.rdma_cmd == 3) {
+	// Set up DMA read from receive ring
+	BX_NE2K_THIS s.remote_start = BX_NE2K_THIS s.remote_dma = BX_NE2K_THIS s.bound_ptr * 256;
+	BX_NE2K_THIS s.remote_bytes = *((Bit16u*) & BX_NE2K_THIS s.mem[BX_NE2K_THIS s.bound_ptr * 256 + 2 - BX_NE2K_MEMSTART]);
+	BX_INFO(("Sending buffer #x%x length %d",
+		  BX_NE2K_THIS s.remote_start,
+		  BX_NE2K_THIS s.remote_bytes));
+    }
+
   // Check for start-tx
-  if (value & 0x04) {
+    if ((value & 0x04) && BX_NE2K_THIS s.TCR.loop_cntl) {
+	if (BX_NE2K_THIS s.TCR.loop_cntl != 1) {
+	    BX_INFO(("Loop mode %d not supported.", BX_NE2K_THIS s.TCR.loop_cntl));
+	} else {
+	    rx_frame (& BX_NE2K_THIS s.mem[BX_NE2K_THIS s.tx_page_start*256 - BX_NE2K_MEMSTART],
+		      BX_NE2K_THIS s.tx_bytes);
+	}
+    } else if (value & 0x04) {
     if (BX_NE2K_THIS s.CR.stop || !BX_NE2K_THIS s.CR.start)
-      bx_panic("ne2k: CR write - tx start, dev in reset");
+      BX_PANIC(("CR write - tx start, dev in reset"));
     
     if (BX_NE2K_THIS s.tx_bytes == 0)
-      bx_panic("ne2k: CR write - tx start, tx bytes == 0");
+      BX_PANIC(("CR write - tx start, tx bytes == 0"));
 
 #ifdef notdef    
     // XXX debug stuff
-    printf("packet tx (%d bytes):\n\t", BX_NE2K_THIS s.tx_bytes);
+    printf("packet tx (%d bytes):\t", BX_NE2K_THIS s.tx_bytes);
     for (int i = 0; i < BX_NE2K_THIS s.tx_bytes; i++) {
       printf("%02x ", BX_NE2K_THIS s.mem[BX_NE2K_THIS s.tx_page_start*256 - 
 				BX_NE2K_MEMSTART + i]);
       if (i && (((i+1) % 16) == 0)) 
-	printf("\n\t");
+	printf("\t");
     }
-    printf("\n");
+    printf("");
 #endif    
 
     // Send the packet to the system driver
-    ethdev->sendpkt(& BX_NE2K_THIS s.mem[BX_NE2K_THIS s.tx_page_start*256 - 
-				BX_NE2K_MEMSTART],
-		    BX_NE2K_THIS s.tx_bytes);
+    BX_NE2K_THIS ethdev->sendpkt(& BX_NE2K_THIS s.mem[BX_NE2K_THIS s.tx_page_start*256 - BX_NE2K_MEMSTART], BX_NE2K_THIS s.tx_bytes);
 
     // some more debug
     if (BX_NE2K_THIS s.tx_timer_active)
-      bx_panic("ne2k: CR write, tx timer still active");
+      BX_PANIC(("CR write, tx timer still active"));
     
     // Schedule a timer to trigger a tx-complete interrupt
     // The number of microseconds is the bit-time / 10.
@@ -177,7 +212,7 @@ bx_ne2k_c::write_cr(Bit32u value)
       BX_NE2K_THIS s.remote_bytes == 0) {
     BX_NE2K_THIS s.ISR.rdma_done = 1;
     if (BX_NE2K_THIS s.IMR.rdma_inte) {
-      BX_NE2K_THIS devices->pic->trigger_irq(BX_NE2K_THIS s.base_irq);
+      pluginRaiseIRQ(BX_NE2K_THIS s.base_irq);
     }
   }
 }
@@ -196,7 +231,7 @@ bx_ne2k_c::chipmem_read(Bit32u address, unsigned int io_len)
   Bit32u retval = 0;
 
   if ((io_len == 2) && (address & 0x1)) 
-    bx_panic("ne2k: unaligned chipmem word read");
+    BX_PANIC(("unaligned chipmem word read"));
 
   // ROM'd MAC address
   if ((address >=0) && (address <= 31)) {
@@ -207,7 +242,7 @@ bx_ne2k_c::chipmem_read(Bit32u address, unsigned int io_len)
     return (retval);
   }
 
-  if ((address >= BX_NE2K_MEMSTART) && (address <= BX_NE2K_MEMEND)) {
+  if ((address >= BX_NE2K_MEMSTART) && (address < BX_NE2K_MEMEND)) {
     retval = BX_NE2K_THIS s.mem[address - BX_NE2K_MEMSTART];
     if (io_len == 2) {
       retval |= (BX_NE2K_THIS s.mem[address - BX_NE2K_MEMSTART + 1] << 8);
@@ -215,7 +250,7 @@ bx_ne2k_c::chipmem_read(Bit32u address, unsigned int io_len)
     return (retval);
   }
 
-  bx_printf("ne2k: out-of-bounds chipmem read, %04X\n", address);
+  BX_DEBUG(("out-of-bounds chipmem read, %04X", address));
 
   return (0xff);
 }
@@ -224,14 +259,14 @@ void
 bx_ne2k_c::chipmem_write(Bit32u address, Bit32u value, unsigned io_len)
 {
   if ((io_len == 2) && (address & 0x1)) 
-    bx_panic("ne2k: unaligned chipmem word write");
+    BX_PANIC(("unaligned chipmem word write"));
 
-  if ((address >= BX_NE2K_MEMSTART) && (address <= BX_NE2K_MEMEND)) {
+  if ((address >= BX_NE2K_MEMSTART) && (address < BX_NE2K_MEMEND)) {
     BX_NE2K_THIS s.mem[address - BX_NE2K_MEMSTART] = value & 0xff;
     if (io_len == 2)
       BX_NE2K_THIS s.mem[address - BX_NE2K_MEMSTART + 1] = value >> 8;
   } else
-    bx_printf("ne2k: out-of-bounds chipmem read, %04X\n", address);
+    BX_DEBUG(("out-of-bounds chipmem write, %04X", address));
 }
 
 //
@@ -251,22 +286,40 @@ bx_ne2k_c::asic_read(Bit32u offset, unsigned int io_len)
 
   switch (offset) {
   case 0x0:  // Data register
-    // 
-    // The device must have been set up to perform DMA in
-    // the same size as is being requested (the WTS bit
-    // in the DCR), a read remote-DMA command must have
-    // been issued, and the source-address and length
-    // registers must have been initialised.
     //
-    if (io_len != (1 + BX_NE2K_THIS s.DCR.wdsize))
-      bx_panic("ne2k: dma read, wrong size %d", io_len);
+    // A read remote-DMA command must have been issued,
+    // and the source-address and length registers must  
+    // have been initialised.
+    //
+    if (io_len > BX_NE2K_THIS s.remote_bytes)
+      {BX_ERROR(("ne2K: dma read underrun iolen=%d remote_bytes=%d",io_len,BX_NE2K_THIS s.remote_bytes));
+       return 0;
+      }
 
-    if (BX_NE2K_THIS s.remote_bytes == 0)
-      bx_panic("ne2K: dma read, byte count 0");
-    
+    //BX_INFO(("ne2k read DMA: addr=%4x remote_bytes=%d",BX_NE2K_THIS s.remote_dma,BX_NE2K_THIS s.remote_bytes));
     retval = chipmem_read(BX_NE2K_THIS s.remote_dma, io_len);
-    BX_NE2K_THIS s.remote_dma   += io_len;
-    BX_NE2K_THIS s.remote_bytes -= io_len;
+    //
+    // The 8390 bumps the address and decreases the byte count
+    // by the selected word size after every access, not by
+    // the amount of data requested by the host (io_len).
+    //
+    BX_NE2K_THIS s.remote_dma += (BX_NE2K_THIS s.DCR.wdsize + 1);
+    if (BX_NE2K_THIS s.remote_dma == BX_NE2K_THIS s.page_stop << 8) {
+      BX_NE2K_THIS s.remote_dma = BX_NE2K_THIS s.page_start << 8;
+    }
+    // keep s.remote_bytes from underflowing
+    if (BX_NE2K_THIS s.remote_bytes > 1)
+      BX_NE2K_THIS s.remote_bytes -= (BX_NE2K_THIS s.DCR.wdsize + 1);
+    else
+      BX_NE2K_THIS s.remote_bytes = 0;
+
+	// If all bytes have been written, signal remote-DMA complete
+	if (BX_NE2K_THIS s.remote_bytes == 0) {
+	    BX_NE2K_THIS s.ISR.rdma_done = 1;
+	    if (BX_NE2K_THIS s.IMR.rdma_inte) {
+		pluginRaiseIRQ(BX_NE2K_THIS s.base_irq);
+	    }
+	}
     break;
 
   case 0xf:  // Reset register
@@ -274,7 +327,7 @@ bx_ne2k_c::asic_read(Bit32u offset, unsigned int io_len)
     break;
 
   default:
-    bx_printf("ne2k: asic read invalid address %04x", (unsigned) offset);
+    BX_INFO(("asic read invalid address %04x", (unsigned) offset));
     break;
   }
 
@@ -284,22 +337,35 @@ bx_ne2k_c::asic_read(Bit32u offset, unsigned int io_len)
 void
 bx_ne2k_c::asic_write(Bit32u offset, Bit32u value, unsigned io_len)
 {
+  BX_DEBUG(("asic write addr=0x%02x, value=0x%04x", (unsigned) offset, (unsigned) value));
   switch (offset) {
   case 0x0:  // Data register - see asic_read for a description
 
-    if (io_len != (1 + BX_NE2K_THIS s.DCR.wdsize))
-      bx_panic("ne2k: dma write, wrong size %d", io_len);
+    if ((io_len == 2) && (BX_NE2K_THIS s.DCR.wdsize == 0)) {
+      BX_PANIC(("dma write length 2 on byte mode operation"));
+      break;
+    }
 
     if (BX_NE2K_THIS s.remote_bytes == 0)
-      bx_panic("ne2K: dma write, byte count 0");
-    
+      BX_PANIC(("ne2K: dma write, byte count 0"));
+
     chipmem_write(BX_NE2K_THIS s.remote_dma, value, io_len);
+    // is this right ??? asic_read uses DCR.wordsize
     BX_NE2K_THIS s.remote_dma   += io_len;
+    if (BX_NE2K_THIS s.remote_dma == BX_NE2K_THIS s.page_stop << 8) {
+      BX_NE2K_THIS s.remote_dma = BX_NE2K_THIS s.page_start << 8;
+    }
+
     BX_NE2K_THIS s.remote_bytes -= io_len;
+    if (BX_NE2K_THIS s.remote_bytes > BX_NE2K_MEMSIZ)
+      BX_NE2K_THIS s.remote_bytes = 0;
 
     // If all bytes have been written, signal remote-DMA complete
     if (BX_NE2K_THIS s.remote_bytes == 0) {
       BX_NE2K_THIS s.ISR.rdma_done = 1;
+      if (BX_NE2K_THIS s.IMR.rdma_inte) {
+	  pluginRaiseIRQ(BX_NE2K_THIS s.base_irq);
+      }
     }
     break;
 
@@ -307,8 +373,8 @@ bx_ne2k_c::asic_write(Bit32u offset, Bit32u value, unsigned io_len)
     reset_device();
     break;
 
-  default:
-    bx_panic("ne2k: asic write invalid address %04x", (unsigned) offset);
+  default: // this is invalid, but happens under win95 device detection
+    BX_INFO(("asic write invalid address %04x, ignoring", (unsigned) offset));
     break ;
   }
 }
@@ -320,9 +386,14 @@ bx_ne2k_c::asic_write(Bit32u offset, Bit32u value, unsigned io_len)
 Bit32u
 bx_ne2k_c::page0_read(Bit32u offset, unsigned int io_len)
 {
-  if (io_len > 1)
-    bx_panic("ne2k: page 0 read from port %04x, len=%u\n", (unsigned) offset,
-             (unsigned) io_len);
+  BX_DEBUG(("page 0 read from port %04x, len=%u", (unsigned) offset,
+	   (unsigned) io_len));
+  if (io_len > 1) {
+    BX_ERROR(("bad length! page 0 read from port %04x, len=%u", (unsigned) offset,
+             (unsigned) io_len)); /* encountered with win98 hardware probe */
+	return 0;
+  }
+
 
   switch (offset) {
   case 0x0:  // CR
@@ -358,7 +429,7 @@ bx_ne2k_c::page0_read(Bit32u offset, unsigned int io_len)
   case 0x6:  // FIFO
     return (BX_NE2K_THIS s.fifo);
     break;
-    
+
   case 0x7:  // ISR
     return ((BX_NE2K_THIS s.ISR.reset     << 7) |
 	    (BX_NE2K_THIS s.ISR.rdma_done << 6) |
@@ -379,12 +450,12 @@ bx_ne2k_c::page0_read(Bit32u offset, unsigned int io_len)
     break;
 
   case 0xa:  // reserved
-    bx_printf("ne2k: reserved read - page 0, 0xa\n");
+    BX_INFO(("reserved read - page 0, 0xa"));
     return (0xff);
     break;
 
   case 0xb:  // reserved
-    bx_printf("ne2k: reserved read - page 0, 0xb\n");
+    BX_INFO(("reserved read - page 0, 0xb"));
     return (0xff);
     break;
     
@@ -412,7 +483,7 @@ bx_ne2k_c::page0_read(Bit32u offset, unsigned int io_len)
     break;
 
   default:
-    bx_panic("ne2k: page 0 offset %04x out of range", (unsigned) offset);
+    BX_PANIC(("page 0 offset %04x out of range", (unsigned) offset));
   }
 
   return(0);
@@ -421,9 +492,15 @@ bx_ne2k_c::page0_read(Bit32u offset, unsigned int io_len)
 void
 bx_ne2k_c::page0_write(Bit32u offset, Bit32u value, unsigned io_len)
 {
-  if (io_len > 1)
-    bx_panic("ne2k: page 0 write to port %04x, len=%u\n", (unsigned) offset,
-             (unsigned) io_len);
+  BX_DEBUG(("page 0 write to port %04x, len=%u", (unsigned) offset,
+	   (unsigned) io_len));
+    if (io_len != 1) {
+	BX_ERROR(("bad length! page 0 write to port %04x, len=%u",
+		  (unsigned) offset,
+            (unsigned) io_len));
+    return;
+  }
+	
   
   switch (offset) {
   case 0x0:  // CR
@@ -435,6 +512,7 @@ bx_ne2k_c::page0_write(Bit32u offset, Bit32u value, unsigned io_len)
     break;
 
   case 0x2:  // PSTOP
+	// BX_INFO(("Writing to PSTOP: %02x", value));
     BX_NE2K_THIS s.page_stop = value;
     break;
 
@@ -468,6 +546,22 @@ bx_ne2k_c::page0_write(Bit32u offset, Bit32u value, unsigned io_len)
     BX_NE2K_THIS s.ISR.overwrite &= ~((value & 0x10) == 0x10);
     BX_NE2K_THIS s.ISR.cnt_oflow &= ~((value & 0x20) == 0x20);
     BX_NE2K_THIS s.ISR.rdma_done &= ~((value & 0x40) == 0x40);
+    value = ((BX_NE2K_THIS s.ISR.rdma_done << 6) |
+             (BX_NE2K_THIS s.ISR.cnt_oflow << 5) |
+             (BX_NE2K_THIS s.ISR.overwrite << 4) |
+             (BX_NE2K_THIS s.ISR.tx_err    << 3) |
+             (BX_NE2K_THIS s.ISR.rx_err    << 2) |
+             (BX_NE2K_THIS s.ISR.pkt_tx    << 1) |
+             (BX_NE2K_THIS s.ISR.pkt_rx));
+    value &= ((BX_NE2K_THIS s.IMR.rdma_inte << 6) |
+              (BX_NE2K_THIS s.IMR.cofl_inte << 5) |
+              (BX_NE2K_THIS s.IMR.overw_inte << 4) |
+              (BX_NE2K_THIS s.IMR.txerr_inte << 3) |
+              (BX_NE2K_THIS s.IMR.rxerr_inte << 2) |
+              (BX_NE2K_THIS s.IMR.tx_inte << 1) |
+              (BX_NE2K_THIS s.IMR.rx_inte));
+    if (value == 0)
+      pluginLowerIRQ(BX_NE2K_THIS s.base_irq);
     break;
 
   case 0x8:  // RSAR0
@@ -499,7 +593,7 @@ bx_ne2k_c::page0_write(Bit32u offset, Bit32u value, unsigned io_len)
   case 0xc:  // RCR
     // Check if the reserved bits are set
     if (value & 0xc0)
-      bx_printf("ne2k: RCR write, reserved bits set\n");
+      BX_INFO(("RCR write, reserved bits set"));
 
     // Set all other bit-fields
     BX_NE2K_THIS s.RCR.errors_ok = ((value & 0x01) == 0x01);
@@ -511,29 +605,29 @@ bx_ne2k_c::page0_write(Bit32u offset, Bit32u value, unsigned io_len)
 
     // Monitor bit is a little suspicious...
     if (value & 0x20)
-      bx_printf("ne2k: RCR write, monitor bit set!\n");
+      BX_INFO(("RCR write, monitor bit set!"));
     break;
 
   case 0xd:  // TCR
     // Check reserved bits
     if (value & 0xe0)
-      bx_panic("ne2k: TCR write, reserved bits set");
+      BX_PANIC(("TCR write, reserved bits set"));
 
     // Test loop mode (not supported)
     if (value & 0x06) {
-      bx_printf("ne2k: TCR write, loop mode not supported");
       BX_NE2K_THIS s.TCR.loop_cntl = (value & 0x6) >> 1;
+	    BX_INFO(("TCR write, loop mode %d not supported", BX_NE2K_THIS s.TCR.loop_cntl));
     } else {
       BX_NE2K_THIS s.TCR.loop_cntl = 0;
     }
 
     // Inhibit-CRC not supported.
     if (value & 0x01)
-      bx_panic("ne2k: TCR write, inhibit-CRC not supported");
+      BX_PANIC(("TCR write, inhibit-CRC not supported"));
 
     // Auto-transmit disable very suspicious
-    if (value & 0x04)
-      bx_panic("ne2k: TCR write, auto transmit disable not supported");
+    if (value & 0x08)
+      BX_PANIC(("TCR write, auto transmit disable not supported"));
 
     // Allow collision-offset to be set, although not used
     BX_NE2K_THIS s.TCR.coll_prio = ((value & 0x08) == 0x08);
@@ -541,15 +635,18 @@ bx_ne2k_c::page0_write(Bit32u offset, Bit32u value, unsigned io_len)
 
   case 0xe:  // DCR
     // Don't allow loopback mode to be set
-    if (!(value & 0x08))
-      bx_panic("ne2k: DCR write, loopback mode selected");
+    if (!(value & 0x08)) {
+      BX_ERROR(("DCR write, loopback mode selected"));
+	  // XXX This is a HACK, lets fix this right!
+	  value -= 8;
+	}
 
     // It is questionable to set longaddr and auto_rx, since they
     // aren't supported on the ne2000. Print a warning and continue
     if (value & 0x04)
-      bx_printf("ne2k: DCR write - LAS set ???\n");
+      BX_INFO(("DCR write - LAS set ???"));
     if (value & 0x10)
-      bx_printf("ne2k: DCR write - AR set ???\n");
+      BX_INFO(("DCR write - AR set ???"));
 
     // Set other values.
     BX_NE2K_THIS s.DCR.wdsize   = ((value & 0x01) == 0x01);
@@ -562,7 +659,7 @@ bx_ne2k_c::page0_write(Bit32u offset, Bit32u value, unsigned io_len)
   case 0xf:  // IMR
     // Check for reserved bit
     if (value & 0x80)
-      bx_panic("ne2k: IMR write, reserved bit set");
+      BX_PANIC(("IMR write, reserved bit set"));
 
     // Set other values
     BX_NE2K_THIS s.IMR.rx_inte    = ((value & 0x01) == 0x01);
@@ -575,7 +672,7 @@ bx_ne2k_c::page0_write(Bit32u offset, Bit32u value, unsigned io_len)
     break;
 
   default:
-    bx_panic("ne2k: page 0 write, bad offset %0x", offset);
+    BX_PANIC(("page 0 write, bad offset %0x", offset));
   }
 }
 
@@ -587,9 +684,11 @@ bx_ne2k_c::page0_write(Bit32u offset, Bit32u value, unsigned io_len)
 Bit32u
 bx_ne2k_c::page1_read(Bit32u offset, unsigned int io_len)
 {
+  BX_DEBUG(("page 1 read from port %04x, len=%u", (unsigned) offset,
+	   (unsigned) io_len));
   if (io_len > 1)
-    bx_panic("ne2k: page 1 read from port %04x, len=%u\n", (unsigned) offset,
-             (unsigned) io_len);
+    BX_PANIC(("bad length! page 1 read from port %04x, len=%u", (unsigned) offset,
+             (unsigned) io_len));
 
   switch (offset) {
   case 0x0:  // CR
@@ -606,6 +705,7 @@ bx_ne2k_c::page1_read(Bit32u offset, unsigned int io_len)
     break;
 
   case 0x7:  // CURR
+      BX_DEBUG(("returning current page: %02x", (BX_NE2K_THIS s.curr_page)));
     return (BX_NE2K_THIS s.curr_page);
 
   case 0x8:  // MAR0-7
@@ -620,7 +720,7 @@ bx_ne2k_c::page1_read(Bit32u offset, unsigned int io_len)
     break;
 
   default:
-    bx_panic("ne2k: page 1 r offset %04x out of range", (unsigned) offset);
+    BX_PANIC(("page 1 r offset %04x out of range", (unsigned) offset));
   }
 
   return (0);
@@ -629,6 +729,7 @@ bx_ne2k_c::page1_read(Bit32u offset, unsigned int io_len)
 void
 bx_ne2k_c::page1_write(Bit32u offset, Bit32u value, unsigned io_len)
 {
+  BX_DEBUG(("page 1 w offset %04x", (unsigned) offset));
   switch (offset) {
   case 0x0:  // CR
     write_cr(value);
@@ -659,7 +760,7 @@ bx_ne2k_c::page1_write(Bit32u offset, Bit32u value, unsigned io_len)
     break;
 
   default:
-    bx_panic("ne2k: page 1 w offset %04x out of range", (unsigned) offset);
+    BX_PANIC(("page 1 w offset %04x out of range", (unsigned) offset));
   }  
 }
 
@@ -671,9 +772,10 @@ bx_ne2k_c::page1_write(Bit32u offset, Bit32u value, unsigned io_len)
 Bit32u
 bx_ne2k_c::page2_read(Bit32u offset, unsigned int io_len)
 {
+  BX_DEBUG(("page 2 read from port %04x, len=%u", (unsigned) offset, (unsigned) io_len));
+
   if (io_len > 1)
-    bx_panic("ne2k: page 2 read from port %04x, len=%u\n", (unsigned) offset,
-             (unsigned) io_len);
+    BX_PANIC(("bad length!  page 2 read from port %04x, len=%u", (unsigned) offset, (unsigned) io_len));
 
   switch (offset) {
   case 0x0:  // CR
@@ -712,7 +814,7 @@ bx_ne2k_c::page2_read(Bit32u offset, unsigned int io_len)
   case 0x9:
   case 0xa:
   case 0xb:
-    bx_printf("ne2k: reserved read - page 2, 0x%02x\n", (unsigned) offset);
+    BX_ERROR(("reserved read - page 2, 0x%02x", (unsigned) offset));
     return (0xff);
     break;
 
@@ -752,7 +854,7 @@ bx_ne2k_c::page2_read(Bit32u offset, unsigned int io_len)
     break;
 
   default:
-    bx_panic("ne2k: page 2 offset %04x out of range", (unsigned) offset);    
+    BX_PANIC(("page 2 offset %04x out of range", (unsigned) offset));
   }
 
   return (0);
@@ -761,11 +863,11 @@ bx_ne2k_c::page2_read(Bit32u offset, unsigned int io_len)
 void
 bx_ne2k_c::page2_write(Bit32u offset, Bit32u value, unsigned io_len)
 {
-  // Maybe all writes here should be bx_panic'd, since they
+  // Maybe all writes here should be BX_PANIC()'d, since they
   // affect internal operation, but let them through for now
   // and print a warning.
   if (offset != 0)
-    bx_printf("ne2k: page 2 write ?\n");
+    BX_ERROR(("page 2 write ?"));
 
   switch (offset) {
   case 0x0:  // CR
@@ -789,7 +891,7 @@ bx_ne2k_c::page2_write(Bit32u offset, Bit32u value, unsigned io_len)
     break;
 
   case 0x4:
-    bx_panic("ne2k: page 2 write to reserved offset 4");
+    BX_PANIC(("page 2 write to reserved offset 4"));
     break;
 
   case 0x5:  // Local Next-packet pointer
@@ -816,11 +918,11 @@ bx_ne2k_c::page2_write(Bit32u offset, Bit32u value, unsigned io_len)
   case 0xd:
   case 0xe:
   case 0xf:
-    bx_panic("ne2k: page 2 write to reserved offset %0x", offset);
+    BX_PANIC(("page 2 write to reserved offset %0x", offset));
     break;
    
   default:
-    bx_panic("ne2k: page 2 write, illegal offset %0x", offset);
+    BX_PANIC(("page 2 write, illegal offset %0x", offset));
     break;
   }
 }
@@ -831,14 +933,14 @@ bx_ne2k_c::page2_write(Bit32u offset, Bit32u value, unsigned io_len)
 Bit32u
 bx_ne2k_c::page3_read(Bit32u offset, unsigned int io_len)
 {
-  bx_panic("ne2k: page 3 read attempted");
+  BX_PANIC(("page 3 read attempted"));
   return (0);
 }
 
 void
 bx_ne2k_c::page3_write(Bit32u offset, Bit32u value, unsigned io_len)
 {
-  bx_panic("ne2k: page 3 write attempted");
+  BX_PANIC(("page 3 write attempted"));
 }
 
 //
@@ -855,11 +957,12 @@ bx_ne2k_c::tx_timer_handler(void *this_ptr)
 void
 bx_ne2k_c::tx_timer(void)
 {
+  BX_DEBUG(("tx_timer"));
   BX_NE2K_THIS s.TSR.tx_ok = 1;
   // Generate an interrupt if not masked and not one in progress
   if (BX_NE2K_THIS s.IMR.tx_inte && !BX_NE2K_THIS s.ISR.pkt_tx) {
     BX_NE2K_THIS s.ISR.pkt_tx = 1;
-    BX_NE2K_THIS devices->pic->trigger_irq(BX_NE2K_THIS s.base_irq);
+    pluginRaiseIRQ(BX_NE2K_THIS s.base_irq);
   }
   BX_NE2K_THIS s.tx_timer_active = 0;
 }
@@ -885,7 +988,8 @@ bx_ne2k_c::read(Bit32u address, unsigned io_len)
 #else
   UNUSED(this_ptr);
 #endif  // !BX_USE_NE2K_SMF
-  Bit32u retval;
+  BX_DEBUG(("read addr %x, len %d", address, io_len));
+  Bit32u retval = 0;
   int offset = address - BX_NE2K_THIS s.base_address;
 
   if (offset >= 0x10) {
@@ -909,8 +1013,8 @@ bx_ne2k_c::read(Bit32u address, unsigned io_len)
       break;
 
     default:
-      bx_panic("ne2K: unknown value of pgsel in read - %d",
-	       BX_NE2K_THIS s.CR.pgsel);
+      BX_PANIC(("ne2K: unknown value of pgsel in read - %d",
+	       BX_NE2K_THIS s.CR.pgsel));
     }
   }
 
@@ -938,6 +1042,7 @@ bx_ne2k_c::write(Bit32u address, Bit32u value, unsigned io_len)
 #else
   UNUSED(this_ptr);
 #endif  // !BX_USE_NE2K_SMF
+  BX_DEBUG(("write with length %d", io_len));
   int offset = address - BX_NE2K_THIS s.base_address;
 
   //
@@ -967,8 +1072,8 @@ bx_ne2k_c::write(Bit32u address, Bit32u value, unsigned io_len)
       break;
 
     default:
-      bx_panic("ne2K: unknown value of pgsel in write - %d",
-	       BX_NE2K_THIS s.CR.pgsel);
+      BX_PANIC(("ne2K: unknown value of pgsel in write - %d",
+	       BX_NE2K_THIS s.CR.pgsel));
     }
   }
 }
@@ -1007,6 +1112,7 @@ bx_ne2k_c::mcast_index(const void *dst)
 void
 bx_ne2k_c::rx_handler(void *arg, const void *buf, unsigned len)
 {
+    // BX_DEBUG(("rx_handler with length %d", len));
   bx_ne2k_c *class_ptr = (bx_ne2k_c *) arg;
   
   class_ptr->rx_frame(buf, len);
@@ -1032,15 +1138,19 @@ bx_ne2k_c::rx_frame(const void *buf, unsigned io_len)
   unsigned char *startptr;
   static unsigned char bcast_addr[6] = {0xff,0xff,0xff,0xff,0xff,0xff};
 
-  if ((BX_NE2K_THIS s.CR.start == 0) ||
+  BX_DEBUG(("rx_frame with length %d", io_len));
+
+
+  if ((BX_NE2K_THIS s.CR.stop != 0) ||
       (BX_NE2K_THIS s.page_start == 0) ||
       (BX_NE2K_THIS s.TCR.loop_cntl != 0)) {
+
     return;
   }
 
   // Add the pkt header + CRC to the length, and work
   // out how many 256-byte pages the frame would occupy
-  pages = (io_len + 4 + 4)/256 + 1;
+  pages = (io_len + 4 + 4 + 255)/256;
 
   if (BX_NE2K_THIS s.curr_page < BX_NE2K_THIS s.bound_ptr) {
     avail = BX_NE2K_THIS s.bound_ptr - BX_NE2K_THIS s.curr_page;    
@@ -1058,35 +1168,51 @@ bx_ne2k_c::rx_frame(const void *buf, unsigned io_len)
   }
 
   if ((io_len < 60) && !BX_NE2K_THIS s.RCR.runts_ok) {
+    BX_DEBUG(("rejected small packet, length %d", io_len));
     return;
   }
 
   // Do address filtering if not in promiscuous mode
-  if (!BX_NE2K_THIS s.RCR.promisc) {
+  if (! BX_NE2K_THIS s.RCR.promisc) {
     if (!memcmp(buf, bcast_addr, 6)) {
       if (!BX_NE2K_THIS s.RCR.broadcast) {
 	return;
       }
     } else if (pktbuf[0] & 0x01) {
+	if (! BX_NE2K_THIS s.RCR.multicast) {
+	    return;
+	}
       idx = mcast_index(buf);
       if (!(BX_NE2K_THIS s.mchash[idx >> 3] & (1 << (idx & 0x7)))) {
 	return;
       }
-    } else if (memcmp(buf, BX_NE2K_THIS s.physaddr, 6)) {
+    } else if (0 != memcmp(buf, BX_NE2K_THIS s.physaddr, 6)) {
       return;
     }
+  } else {
+      BX_DEBUG(("rx_frame promiscuous receive"));
   }
 
+//    BX_INFO(("rx_frame %d to %x:%x:%x:%x:%x:%x from %x:%x:%x:%x:%x:%x",
+//  	   io_len,
+//  	   pktbuf[0], pktbuf[1], pktbuf[2], pktbuf[3], pktbuf[4], pktbuf[5],
+//  	   pktbuf[6], pktbuf[7], pktbuf[8], pktbuf[9], pktbuf[10], pktbuf[11]));
+
   nextpage = BX_NE2K_THIS s.curr_page + pages;
   if (nextpage >= BX_NE2K_THIS s.page_stop) {
     nextpage -= BX_NE2K_THIS s.page_stop - BX_NE2K_THIS s.page_start;
   }
 
   // Setup packet header
-  pkthdr[0] = 0;	// rx status
+  pkthdr[0] = 0;	// rx status - old behavior
+  pkthdr[0] = 1;        // Probably better to set it all the time
+                        // rather than set it to 0, which is clearly wrong.
+  if (pktbuf[0] & 0x01) {
+    pkthdr[0] |= 0x20;  // rx status += multicast packet
+  }
   pkthdr[1] = nextpage;	// ptr to next packet
-  pkthdr[2] = (io_len + 8) & 0xff;	// length-low
-  pkthdr[3] = (io_len + 8) >> 8;	// length-hi
+  pkthdr[2] = (io_len + 4) & 0xff;	// length-low
+  pkthdr[3] = (io_len + 4) >> 8;	// length-hi
 
   // copy into buffer, update curpage, and signal interrupt if config'd
   startptr = & BX_NE2K_THIS s.mem[BX_NE2K_THIS s.curr_page * 256 -
@@ -1116,7 +1242,7 @@ bx_ne2k_c::rx_frame(const void *buf, unsigned io_len)
   BX_NE2K_THIS s.ISR.pkt_rx = 1;
 
   if (BX_NE2K_THIS s.IMR.rx_inte) {
-    BX_NE2K_THIS devices->pic->trigger_irq(BX_NE2K_THIS s.base_irq);
+    pluginRaiseIRQ(BX_NE2K_THIS s.base_irq);
   }
 
 }
@@ -1124,21 +1250,22 @@ bx_ne2k_c::rx_frame(const void *buf, unsigned io_len)
 void
 bx_ne2k_c::init(bx_devices_c *d)
 {
+  BX_DEBUG(("Init $Id: ne2k.cc,v 1.2 2002/07/21 23:14:55 DemonLord Exp $"));
   BX_NE2K_THIS devices = d;
 
-  if (bx_options.ne2k.valid) {
+
+  if (bx_options.ne2k.Ovalid->get ()) {
     // Bring the register state into power-up state
     reset_device();
 
     // Read in values from config file
-    BX_NE2K_THIS s.base_address = bx_options.ne2k.ioaddr;
-    BX_NE2K_THIS s.base_irq     = bx_options.ne2k.irq;
-    memcpy(BX_NE2K_THIS s.physaddr, bx_options.ne2k.macaddr, 6);
+    BX_NE2K_THIS s.base_address = bx_options.ne2k.Oioaddr->get ();
+    BX_NE2K_THIS s.base_irq     = bx_options.ne2k.Oirq->get ();
+    memcpy(BX_NE2K_THIS s.physaddr, bx_options.ne2k.Omacaddr->getptr (), 6);
 
     BX_NE2K_THIS s.tx_timer_index =
       pluginRegisterTimer(this, tx_timer_handler, 0,
 				  0,0); // one-shot, inactive
-
     // Register the IRQ and i/o port addresses
     pluginRegisterIRQ(BX_NE2K_THIS s.base_irq,
 				       "ne2000 ethernet NIC");
@@ -1155,6 +1282,15 @@ bx_ne2k_c::init(bx_devices_c *d)
 						      addr, 
 						      "ne2000 NIC", 1);
     }
+	BX_INFO(("port 0x%x/32 irq %d mac %02x:%02x:%02x:%02x:%02x:%02x",
+				BX_NE2K_THIS s.base_address,
+				BX_NE2K_THIS s.base_irq,
+				BX_NE2K_THIS s.physaddr[0],
+				BX_NE2K_THIS s.physaddr[1],
+				BX_NE2K_THIS s.physaddr[2],
+				BX_NE2K_THIS s.physaddr[3],
+				BX_NE2K_THIS s.physaddr[4],
+				BX_NE2K_THIS s.physaddr[5]));
     
     // Initialise the mac address area by doubling the physical address
     BX_NE2K_THIS s.macaddr[0]  = BX_NE2K_THIS s.physaddr[0];
@@ -1175,22 +1311,273 @@ bx_ne2k_c::init(bx_devices_c *d)
       BX_NE2K_THIS s.macaddr[i] = 0x57;
     
     // Attach to the simulated ethernet dev
-    BX_NE2K_THIS ethdev = eth_locator_c::create(bx_options.ne2k.ethmod, 
-						bx_options.ne2k.ethdev,
-				    (const char *) bx_options.ne2k.macaddr,
+    BX_NE2K_THIS ethdev = eth_locator_c::create(bx_options.ne2k.Oethmod->getptr (), 
+						bx_options.ne2k.Oethdev->getptr (),
+				    (const char *) bx_options.ne2k.Omacaddr->getptr (),
 						rx_handler, 
 						this);
     
     if (BX_NE2K_THIS ethdev == NULL) {
-      bx_printf("ne2k: could not find eth module %s - using null instead\n",
-		bx_options.ne2k.ethmod);
+      BX_PANIC(("could not find eth module %s"));
+      // if they continue, use null.
+      BX_INFO(("could not find eth module %s - using null instead",
+		bx_options.ne2k.Oethmod->getptr ()));
       
       BX_NE2K_THIS ethdev = eth_locator_c::create("null", NULL,
-				    (const char *) bx_options.ne2k.macaddr,
+				    (const char *) bx_options.ne2k.Omacaddr->getptr (),
 						  rx_handler, 
 						  this);
       if (BX_NE2K_THIS ethdev == NULL)
-	bx_panic("ne2k: could not locate null module\n");
+	BX_PANIC(("could not locate null module"));
     }
   }
 }
+
+#if BX_DEBUGGER
+
+/*
+ * this implements the info ne2k commands in the debugger.
+ * info ne2k - shows all registers
+ * info ne2k page N - shows all registers in a page
+ * info ne2k page N reg M - shows just one register
+ */
+
+#define SHOW_FIELD(reg,field) do { \
+  if (n>0 && !(n%5)) fprintf (fp, "\n  "); \
+  fprintf(fp, "%s=%d ", #field, BX_NE2K_THIS s.reg.field); \
+  n++; \
+} while (0);
+#define BX_HIGH_BYTE(x) ((0xff00 & (x)) >> 8)
+#define BX_LOW_BYTE(x) (0x00ff & (x))
+#define BX_DUPLICATE(n) if (brief && num!=n) break;
+
+void
+bx_ne2k_c::print_info (FILE *fp, int page, int reg, int brief)
+{
+  int i;
+  int n = 0;
+  if (page < 0) {
+    for (page=0; page<=2; page++)
+      BX_NE2K_THIS print_info (fp, page, reg, 1);
+    // tell them how to use this command
+    fprintf (fp, "\nHow to use the info ne2k command:\n");
+    fprintf (fp, "info ne2k - show all registers\n");
+    fprintf (fp, "info ne2k page N - show registers in page N\n");
+    fprintf (fp, "info ne2k page N reg M - show just one register\n");
+    return;
+  }
+  if (page > 2) {
+    fprintf (fp, "NE2K has only pages 0, 1, and 2.  Page %d is out of range.\n", page);
+    return;
+  }
+  if (reg < 0) {
+    fprintf (fp, "NE2K registers, page %d\n", page);
+    fprintf (fp, "----------------------\n");
+    for (reg=0; reg<=15; reg++)
+      BX_NE2K_THIS print_info (fp, page, reg, 1);
+    fprintf (fp, "----------------------\n");
+    return;
+  }
+  if (reg > 15) {
+    fprintf (fp, "NE2K has only registers 0-15 (0x0-0xf).  Register %d is out of range.\n", reg);
+    return;
+  }
+  if (!brief) {
+    fprintf (fp, "NE2K Info - page %d, register 0x%02x\n", page, reg);
+    fprintf (fp, "----------------------------------\n");
+  }
+  int num = page*0x100 + reg;
+  switch (num) {
+    case 0x0000:
+    case 0x0100:
+    case 0x0200:
+      fprintf (fp, "CR (Command register):\n  ");
+      SHOW_FIELD (CR, stop);
+      SHOW_FIELD (CR, start);
+      SHOW_FIELD (CR, tx_packet);
+      SHOW_FIELD (CR, rdma_cmd);
+      SHOW_FIELD (CR, pgsel);
+      fprintf (fp, "\n");
+      break;
+    case 0x0003:
+      fprintf (fp, "BNRY = Boundary Pointer = 0x%02x\n", BX_NE2K_THIS s.bound_ptr);
+      break;
+    case 0x0004:
+      fprintf (fp, "TSR (Transmit Status Register), read-only:\n  ");
+      SHOW_FIELD (TSR, tx_ok);
+      SHOW_FIELD (TSR, reserved);
+      SHOW_FIELD (TSR, collided);
+      SHOW_FIELD (TSR, aborted);
+      SHOW_FIELD (TSR, no_carrier);
+      SHOW_FIELD (TSR, fifo_ur);
+      SHOW_FIELD (TSR, cd_hbeat);
+      SHOW_FIELD (TSR, ow_coll);
+      fprintf (fp, "\n");
+      // fall through into TPSR, no break line.
+    case 0x0204:
+      fprintf (fp, "TPSR = Transmit Page Start = 0x%02x\n", BX_NE2K_THIS s.tx_page_start);
+      break;
+    case 0x0005:
+    case 0x0006:  BX_DUPLICATE(0x0005);
+      fprintf (fp, "NCR = Number of Collisions Register (read-only) = 0x%02x\n", BX_NE2K_THIS s.num_coll);
+      fprintf (fp, "TBCR1,TBCR0 = Transmit Byte Count = %02x %02x\n", 
+	  BX_HIGH_BYTE (BX_NE2K_THIS s.tx_bytes),
+	  BX_LOW_BYTE (BX_NE2K_THIS s.tx_bytes));
+      fprintf (fp, "FIFO = %02x\n", BX_NE2K_THIS s.fifo);
+      break;
+    case 0x0007:
+      fprintf (fp, "ISR (Interrupt Status Register):\n  ");
+      SHOW_FIELD (ISR, pkt_rx);
+      SHOW_FIELD (ISR, pkt_tx);
+      SHOW_FIELD (ISR, rx_err);
+      SHOW_FIELD (ISR, tx_err);
+      SHOW_FIELD (ISR, overwrite);
+      SHOW_FIELD (ISR, cnt_oflow);
+      SHOW_FIELD (ISR, rdma_done);
+      SHOW_FIELD (ISR, reset);
+      fprintf (fp, "\n");
+      break;
+    case 0x0008:
+    case 0x0009:  BX_DUPLICATE(0x0008);
+      fprintf (fp, "CRDA1,0 = Current remote DMA address = %02x %02x\n", 
+	  BX_HIGH_BYTE (BX_NE2K_THIS s.remote_dma),
+	  BX_LOW_BYTE (BX_NE2K_THIS s.remote_dma));
+      fprintf (fp, "RSAR1,0 = Remote start address = %02x %02x\n", 
+	  BX_HIGH_BYTE(s.remote_start),
+	  BX_LOW_BYTE(s.remote_start));
+      break;
+    case 0x000a:
+    case 0x000b:  BX_DUPLICATE(0x000a);
+      fprintf (fp, "RCBR1,0 = Remote byte count = %02x\n", BX_NE2K_THIS s.remote_bytes);
+      break;
+    case 0x000c:
+      fprintf (fp, "RSR (Receive Status Register), read-only:\n  ");
+      SHOW_FIELD (RSR, rx_ok);
+      SHOW_FIELD (RSR, bad_crc);
+      SHOW_FIELD (RSR, bad_falign);
+      SHOW_FIELD (RSR, fifo_or);
+      SHOW_FIELD (RSR, rx_missed);
+      SHOW_FIELD (RSR, rx_mbit);
+      SHOW_FIELD (RSR, rx_disabled);
+      SHOW_FIELD (RSR, deferred);
+      fprintf (fp, "\n");
+      // fall through into RCR
+    case 0x020c:
+      fprintf (fp, "RCR (Receive Configuration Register):\n  ");
+      SHOW_FIELD (RCR, errors_ok);
+      SHOW_FIELD (RCR, runts_ok);
+      SHOW_FIELD (RCR, broadcast);
+      SHOW_FIELD (RCR, multicast);
+      SHOW_FIELD (RCR, promisc);
+      SHOW_FIELD (RCR, monitor);
+      SHOW_FIELD (RCR, reserved);
+      fprintf (fp, "\n");
+      break;
+    case 0x000d:
+      fprintf (fp, "CNTR0 = Tally Counter 0 (Frame alignment errors) = %02x\n",
+	  BX_NE2K_THIS s.tallycnt_0);
+      // fall through into TCR
+    case 0x020d:
+      fprintf (fp, "TCR (Transmit Configuration Register):\n  ");
+      SHOW_FIELD (TCR, crc_disable);
+      SHOW_FIELD (TCR, loop_cntl);
+      SHOW_FIELD (TCR, ext_stoptx);
+      SHOW_FIELD (TCR, coll_prio);
+      SHOW_FIELD (TCR, reserved);
+      fprintf (fp, "\n");
+      break;
+    case 0x000e:
+      fprintf (fp, "CNTR1 = Tally Counter 1 (CRC Errors) = %02x\n",
+	  BX_NE2K_THIS s.tallycnt_1);
+      // fall through into DCR
+    case 0x020e:
+      fprintf (fp, "DCR (Data Configuration Register):\n  ");
+      SHOW_FIELD (DCR, wdsize);
+      SHOW_FIELD (DCR, endian);
+      SHOW_FIELD (DCR, longaddr);
+      SHOW_FIELD (DCR, loop);
+      SHOW_FIELD (DCR, auto_rx);
+      SHOW_FIELD (DCR, fifo_size);
+      fprintf (fp, "\n");
+      break;
+    case 0x000f:
+      fprintf (fp, "CNTR2 = Tally Counter 2 (Missed Packet Errors) = %02x\n",
+	  BX_NE2K_THIS s.tallycnt_2);
+      // fall through into IMR
+    case 0x020f:
+      fprintf (fp, "IMR (Interrupt Mask Register)\n  ");
+      SHOW_FIELD (IMR, rx_inte);
+      SHOW_FIELD (IMR, tx_inte);
+      SHOW_FIELD (IMR, rxerr_inte);
+      SHOW_FIELD (IMR, txerr_inte);
+      SHOW_FIELD (IMR, overw_inte);
+      SHOW_FIELD (IMR, cofl_inte);
+      SHOW_FIELD (IMR, rdma_inte);
+      SHOW_FIELD (IMR, reserved);
+      fprintf (fp, "\n");
+      break;
+    case 0x0101:
+    case 0x0102:  BX_DUPLICATE(0x0101);
+    case 0x0103:  BX_DUPLICATE(0x0101);
+    case 0x0104:  BX_DUPLICATE(0x0101);
+    case 0x0105:  BX_DUPLICATE(0x0101);
+    case 0x0106:  BX_DUPLICATE(0x0101);
+      fprintf (fp, "MAC address registers are located at page 1, registers 1-6.\n");
+      fprintf (fp, "The MAC address is ");
+      for (i=0; i<=5; i++) 
+	fprintf (fp, "%02x%c", BX_NE2K_THIS s.physaddr[i], i<5?':' : '\n');
+      break;
+    case 0x0107:
+      fprintf (fp, "Current page is 0x%02x\n", BX_NE2K_THIS s.curr_page);
+      break;
+    case 0x0108:
+    case 0x0109:  BX_DUPLICATE(0x0108);
+    case 0x010A:  BX_DUPLICATE(0x0108);
+    case 0x010B:  BX_DUPLICATE(0x0108);
+    case 0x010C:  BX_DUPLICATE(0x0108);
+    case 0x010D:  BX_DUPLICATE(0x0108);
+    case 0x010E:  BX_DUPLICATE(0x0108);
+    case 0x010F:  BX_DUPLICATE(0x0108);
+      fprintf (fp, "MAR0-7 (Multicast address registers 0-7) are set to:\n");
+      for (i=0; i<8; i++) fprintf (fp, "%02x ", BX_NE2K_THIS s.mchash[i]);
+      fprintf (fp, "\nMAR0 is listed first.\n");
+      break;
+    case 0x0001:
+    case 0x0002:  BX_DUPLICATE(0x0001);
+    case 0x0201:  BX_DUPLICATE(0x0001);
+    case 0x0202:  BX_DUPLICATE(0x0001);
+      fprintf (fp, "PSTART = Page start register = %02x\n", BX_NE2K_THIS s.page_start);
+      fprintf (fp, "PSTOP = Page stop register = %02x\n", BX_NE2K_THIS s.page_stop);
+      fprintf (fp, "Local DMA address = %02x %02x\n", 
+	  BX_HIGH_BYTE(BX_NE2K_THIS s.local_dma),
+	  BX_LOW_BYTE(BX_NE2K_THIS s.local_dma));
+      break;
+    case 0x0203:
+      fprintf (fp, "Remote Next Packet Pointer = %02x\n", BX_NE2K_THIS s.rempkt_ptr);
+      break;
+    case 0x0205:
+      fprintf (fp, "Local Next Packet Pointer = %02x\n", BX_NE2K_THIS s.localpkt_ptr);
+      break;
+    case 0x0206:
+    case 0x0207:  BX_DUPLICATE(0x0206);
+      fprintf (fp, "Address Counter= %02x %02x\n", 
+	 BX_HIGH_BYTE(BX_NE2K_THIS s.address_cnt),
+	 BX_LOW_BYTE(BX_NE2K_THIS s.address_cnt));
+      break;
+    case 0x0208:
+    case 0x0209:  BX_DUPLICATE(0x0208);
+    case 0x020A:  BX_DUPLICATE(0x0208);
+    case 0x020B:  BX_DUPLICATE(0x0208);
+      if (!brief) fprintf (fp, "Reserved\n");
+    case 0xffff:
+      fprintf (fp, "IMR (Interrupt Mask Register):\n  ");
+      fprintf (fp, "\n");
+      break;
+    default:
+      fprintf (fp, "NE2K info: sorry, page %d register %d cannot be displayed.\n", page, reg);
+  }
+  if (!brief)
+    fprintf (fp, "\n");
+}
+
+#endif
