@@ -1,4 +1,10 @@
-//  Copyright (C) 2000  MandrakeSoft S.A.
+/////////////////////////////////////////////////////////////////////////
+// $Id: sb16.cc,v 1.1.1.3 2002/07/20 18:31:21 DemonLord Exp $
+/////////////////////////////////////////////////////////////////////////
+// Based on Bochs file <Id: sb16.cc,v 1.18 2002/01/29 17:20:12 vruppert Exp >
+/////////////////////////////////////////////////////////////////////////
+//
+//  Copyright (C) 2002  MandrakeSoft S.A.
 //
 //    MandrakeSoft S.A.
 //    43, rue d'Aboukir
@@ -24,12 +30,39 @@
 
 
 #include "bochs.h"
+#include "sb16.h"
+
+#if BX_USE_SB16_SMF
+#define this (&bx_sb16)
+bx_sb16_c bx_sb16;
+#define LOG_THIS bx_sb16.
+#else
+#define LOG_THIS sb16->
+#endif
+
+
+  int
+plugin_init(plugin_t *plugin, int argc, char *argv[])
+{
+  bx_sb16_c      *sb16;
+
+#if BX_USE_SB16_SMF
+  sb16 = &bx_sb16;
+#else
+  sb16 = new bx_cmos_c();
+#endif
+  sb16->init();
+
+  return(0); // Success
+}
+
+  void
+plugin_fini(void)
+{
+}
 
 // some shortcuts to save typing
 #define LOGFILE         BX_SB16_THIS logfile
-#define BOTHLOG(x)      (x)
-#define MIDILOG(x)      ((bx_options.sb16.midimode>0?x:0x7f))
-#define WAVELOG(x)      ((bx_options.sb16.wavemode>0?x:0x7f))
 #define MIDIDATA        BX_SB16_THIS midifile
 #define WAVEDATA        BX_SB16_THIS wavefile
 #define MPU             BX_SB16_THIS mpu401
@@ -40,19 +73,18 @@
 
 #define BX_SB16_OUTPUT  BX_SB16_THIS output
 
-#if BX_USE_SB16_SMF
-bx_sb16_c bx_sb16;
-#define this NULL
-#endif
+// here's a safe way to print out null pointeres
+#define MIGHT_BE_NULL(x)  ((x==NULL)? "(null)" : x)
 
 bx_sb16_c::bx_sb16_c(void)
 {
-  // nothing for now
+  put("SB16");
+  settype(SB16LOG);
 }
 
 bx_sb16_c::~bx_sb16_c(void)
 {
-  switch (bx_options.sb16.midimode)
+  switch (bx_options.sb16.Omidimode->get ())
     {
     case 2:
       finishmidifile();
@@ -67,7 +99,7 @@ bx_sb16_c::~bx_sb16_c(void)
       break;
     }
 
-  switch (bx_options.sb16.wavemode)
+  switch (bx_options.sb16.Owavemode->get ())
     {
     case 2:
       finishvocfile();
@@ -84,28 +116,26 @@ bx_sb16_c::~bx_sb16_c(void)
 
   delete(BX_SB16_OUTPUT);
 
-  delete(DSP.dma.chunk);
+  delete[] DSP.dma.chunk;
 
-  if (bx_options.sb16.loglevel > 0)
+  if ((bx_options.sb16.Ologlevel->get () > 0) && LOGFILE)
     fclose(LOGFILE);
 }
 
-void bx_sb16_c::init(bx_devices_c *d)
+void bx_sb16_c::init()
 {
-  BX_SB16_THIS devices = d;
   unsigned addr;
 
-  if ( (bx_options.sb16.logfile == NULL) ||
-       (strlen(bx_options.sb16.logfile) < 1) )
-    bx_options.sb16.loglevel = 0;
+  if ( (strlen(bx_options.sb16.Ologfile->getptr ()) < 1) )
+    bx_options.sb16.Ologlevel->set (0);
 
-  if (bx_options.sb16.loglevel > 0)
+  if (bx_options.sb16.Ologlevel->get () > 0)
     {
-      LOGFILE = fopen(bx_options.sb16.logfile,"w"); // logfile for errors etc.
+      LOGFILE = fopen(bx_options.sb16.Ologfile->getptr (),"w"); // logfile for errors etc.
       if (LOGFILE == NULL)
 	{
-	  bx_printf("#SB16: Error opening file %s. Logging disabled.", bx_options.sb16.logfile);
-	  bx_options.sb16.loglevel = 0;
+	  BX_ERROR(("Error opening file %s. Logging disabled.", bx_options.sb16.Ologfile->getptr ()));
+	  bx_options.sb16.Ologlevel->set (0);
 	}
     }
 
@@ -115,33 +145,33 @@ void bx_sb16_c::init(bx_devices_c *d)
   if (BX_SB16_OUTPUT == NULL)
     {
       writelog( MIDILOG(2), "Couldn't initialize output devices. Output disabled.");
-      bx_options.sb16.midimode = 0;
-      bx_options.sb16.wavemode = 0;
+      bx_options.sb16.Omidimode->set (0);
+      bx_options.sb16.Owavemode->set (0);
     }
 
-  if ( (bx_options.sb16.midimode == 2) ||
-       (bx_options.sb16.midimode == 3) )
+  if ( (bx_options.sb16.Omidimode->get () == 2) ||
+       (bx_options.sb16.Omidimode->get () == 3) )
     {
-      MIDIDATA = fopen(bx_options.sb16.midifile,"w");
+      MIDIDATA = fopen(bx_options.sb16.Omidifile->getptr (),"wb");
       if (MIDIDATA == NULL)
 	{
-	  writelog (MIDILOG(2), "Error opening file %s. Midimode disabled.", bx_options.sb16.midifile);
-	  bx_options.sb16.midimode = 0;
+	  writelog (MIDILOG(2), "Error opening file %s. Midimode disabled.", bx_options.sb16.Omidifile->getptr ());
+	  bx_options.sb16.Omidimode->set (0);
 	}
-      else if (bx_options.sb16.midimode == 2)
+      else if (bx_options.sb16.Omidimode->get () == 2)
 	initmidifile();
     }
 
-  if ( (bx_options.sb16.wavemode == 2) ||
-       (bx_options.sb16.wavemode == 3) )
+  if ( (bx_options.sb16.Owavemode->get () == 2) ||
+       (bx_options.sb16.Owavemode->get () == 3) )
     {
-      WAVEDATA = fopen(bx_options.sb16.wavefile,"w");
+      WAVEDATA = fopen(bx_options.sb16.Owavefile->getptr (),"wb");
       if (WAVEDATA == NULL)
 	{
-	  writelog (WAVELOG(2), "Error opening file %s. Wavemode disabled.", bx_options.sb16.wavefile);
-	  bx_options.sb16.wavemode = 0;
+	  writelog (WAVELOG(2), "Error opening file %s. Wavemode disabled.", bx_options.sb16.Owavefile);
+	  bx_options.sb16.Owavemode->set (0);
 	}
-      else if (bx_options.sb16.wavemode == 2)
+      else if (bx_options.sb16.Owavemode->get () == 2)
 	initvocfile();
     }
 
@@ -153,14 +183,14 @@ void bx_sb16_c::init(bx_devices_c *d)
   if (DSP.dma.chunk == NULL)
     {
       writelog( WAVELOG(2), "Couldn't allocate wave buffer - wave output disabled.");
-      bx_options.sb16.wavemode = 0;
+      bx_options.sb16.Owavemode->set (0);
     }
 
-  bx_printf("#SB16 options: midi=%d,%s  wave=%d,%s  log=%d,%s  dmatimer=%d\n",
-	    bx_options.sb16.midimode, bx_options.sb16.midifile,
-	    bx_options.sb16.wavemode, bx_options.sb16.wavefile,
-	    bx_options.sb16.loglevel, bx_options.sb16.logfile,
-	    bx_options.sb16.dmatimer);
+  BX_INFO(("midi=%d,%s  wave=%d,%s  log=%d,%s  dmatimer=%d",
+	    bx_options.sb16.Omidimode->get (), MIGHT_BE_NULL(bx_options.sb16.Omidifile->getptr ()),
+	    bx_options.sb16.Owavemode->get (), MIGHT_BE_NULL(bx_options.sb16.Owavefile->getptr ()),
+	    bx_options.sb16.Ologlevel->get (), MIGHT_BE_NULL(bx_options.sb16.Ologfile->getptr ()),
+	    bx_options.sb16.Odmatimer->get ()));
 
   // allocate the FIFO buffers - except for the MPUMIDICMD buffer
   // these sizes are generous, 16 or 8 would probably be sufficient
@@ -189,6 +219,11 @@ void bx_sb16_c::init(bx_devices_c *d)
 
   BX_SB16_IRQ = -1; // will be initialized later by the mixer reset
 
+  MIXER.reg[0x80] = 2;  // IRQ 5
+  MIXER.reg[0x81] = 2;  // 8-bit DMA 1, no 16-bit DMA
+  MIXER.reg[0x82] = 0;  // no IRQ pending
+  set_irq_dma();        // set the IRQ and DMA
+
   // call the mixer reset
   mixer_writeregister(0x00);
   mixer_writedata(0x00);
@@ -200,24 +235,23 @@ void bx_sb16_c::init(bx_devices_c *d)
   // Allocate the IO addresses, 2x0..2xf, 3x0..3x4 and 388..38b
   for (addr=BX_SB16_IO; addr<BX_SB16_IO+BX_SB16_IOLEN; addr++) {
     pluginRegisterIOReadHandler(this,
-       read_handler, addr, "SB16");
+       read_handler, addr, "SB16",1);
     pluginRegisterIOWriteHandler(this,
-       write_handler, addr, "SB16");
+       write_handler, addr, "SB16",1);
+
     }
   for (addr=BX_SB16_IOMPU; addr<BX_SB16_IOMPU+BX_SB16_IOMPULEN; addr++) {
     pluginRegisterIOReadHandler(this,
-       read_handler, addr, "SB16");
+       read_handler, addr, "SB16",1);
     pluginRegisterIOWriteHandler(this,
-       write_handler, addr, "SB16");
+       write_handler, addr, "SB16",1);
     }
-  /* Uncomment this if you know the consequences...
   for (addr=BX_SB16_IOADLIB; addr<BX_SB16_IOADLIB+BX_SB16_IOADLIBLEN; addr++) {
     pluginRegisterIOReadHandler(this,
-       read_handler, addr, "SB16");
+       read_handler, addr, "SB16",1);
     pluginRegisterIOWriteHandler(this,
-       write_handler, addr, "SB16");
+       write_handler, addr, "SB16",1);
     }
-  */
 
   writelog(BOTHLOG(3),
 	   "driver initialised, IRQ %d, IO %03x/%03x/%03x, DMA %d/%d",
@@ -257,11 +291,15 @@ void bx_sb16_c::dsp_dmatimer (void *this_ptr)
   // However, don't do this if the next byte/word will fill up the
   // output buffer and the output functions are not ready yet.
 
-  if ( (bx_options.sb16.wavemode != 1) ||
+  if ( (bx_options.sb16.Owavemode->get () != 1) ||
        ( (This->dsp.dma.chunkindex + 1 < BX_SOUND_OUTPUT_WAVEPACKETSIZE) &&
 	 (This->dsp.dma.count > 0) ) ||
-       (This->output->waveready() == BX_SOUND_OUTPUT_OK) )
-    bx_pc_system.set_DRQ(BX_SB16_DMAL, 1);
+       (This->output->waveready() == BX_SOUND_OUTPUT_OK) ) {
+    if (DSP.dma.bits == 8)
+      pluginDMASetDRQ(BX_SB16_DMAL, 1);
+    else
+      pluginDMASetDRQ(BX_SB16_DMAH, 1);
+    }
 }
 
 void bx_sb16_c::opl_timer (void *this_ptr)
@@ -309,7 +347,7 @@ void bx_sb16_c::dsp_reset(Bit32u value)
 
       if (DSP.irqpending != 0)
 	{
-	  BX_SB16_THIS devices->pic->untrigger_irq(BX_SB16_IRQ);
+	  pluginLowerIRQ(BX_SB16_IRQ);
 	  writelog(WAVELOG(4), "DSP reset: IRQ untriggered");
 	}
       if (DSP.dma.mode != 0)
@@ -659,7 +697,7 @@ void bx_sb16_c::dsp_datawrite(Bit32u value)
 	  // 1,2: lo(silence) hi(silence) (len in samples)
 	  DSP.datain.getw(&length);
 	  // only handled for VOC output so far
-	  if (bx_options.sb16.wavemode == 2)
+	  if (bx_options.sb16.Owavemode->get () == 2)
 	    {
 	      Bit8u temparray[3] = { length & 0xff, length >> 8, DSP.dma.timeconstant >> 8 };
 	      writevocblock(3, 3, temparray, 0, NULL);
@@ -843,7 +881,7 @@ void bx_sb16_c::dsp_datawrite(Bit32u value)
 	  DSP.dataout.put(0xaa);
 	  DSP.irqpending = 1;
 	  MIXER.reg[0x82] |= 1; // reg 82 shows the kind of IRQ
-	  BX_SB16_THIS devices->pic->trigger_irq(BX_SB16_IRQ);
+	  pluginRaiseIRQ(BX_SB16_IRQ);
 	  break;
 
 	  // unknown command
@@ -906,7 +944,7 @@ void bx_sb16_c::dsp_dma(Bit8u command, Bit8u mode, Bit16u length, Bit8u comp)
   DSP.dma.chunkcount = 0;
 
   Bit32u sampledatarate = (Bit32u) DSP.dma.samplerate * (Bit32u) DSP.dma.bps;
-  DSP.dma.timer = (Bit32u) bx_options.sb16.dmatimer / sampledatarate;
+  DSP.dma.timer = (Bit32u) bx_options.sb16.Odmatimer->get () / sampledatarate;
 
   writelog( WAVELOG(5), "DMA is %db, %dHz, %s, %s, mode %d, %s, %s, %d bps, %d us/b",
 	    DSP.dma.bits, DSP.dma.samplerate, (DSP.dma.stereo != 0)?"stereo":"mono",
@@ -920,13 +958,13 @@ void bx_sb16_c::dsp_dma(Bit8u command, Bit8u mode, Bit16u length, Bit8u comp)
 	// write the output to the device/file
   if (DSP.dma.output == 1)
     {
-      if (bx_options.sb16.wavemode == 1)
+      if (bx_options.sb16.Owavemode->get () == 1)
 	{
 	  if (DSP.outputinit == 0)
 	    {
-	      if (BX_SB16_OUTPUT->openwaveoutput(bx_options.sb16.wavefile) != BX_SOUND_OUTPUT_OK)
+	      if (BX_SB16_OUTPUT->openwaveoutput(bx_options.sb16.Owavefile->getptr ()) != BX_SOUND_OUTPUT_OK)
 		{
-		  bx_options.sb16.wavemode = 0;
+		  bx_options.sb16.Owavemode->set (0);
 		  writelog( WAVELOG(2), "Error: Could not open wave output device.");
 		}
 	      else
@@ -981,11 +1019,12 @@ Bit32u bx_sb16_c::dsp_status()
   // read might be to acknowledge IRQ
   if ( DSP.irqpending != 0 )
     {
-      DSP.irqpending = 0;
       MIXER.reg[0x82] &= (~0x01);
       writelog( WAVELOG(4), "8-bit DMA or SBMIDI IRQ acknowledged");
-      if (MIXER.reg[0x82] == 0)
-	BX_SB16_THIS devices->pic->untrigger_irq(BX_SB16_IRQ);
+      if (MIXER.reg[0x82] == 0) {
+        DSP.irqpending = 0;
+        pluginLowerIRQ(BX_SB16_IRQ);
+      }
     }
 
   // if buffer is not empty, there is data to be read
@@ -1004,10 +1043,11 @@ Bit32u bx_sb16_c::dsp_irq16ack()
 
   if ( DSP.irqpending != 0 )
     {
-      DSP.irqpending = 0;
       MIXER.reg[0x82] &= (~0x02);
-      if (MIXER.reg[0x82] == 0)
-	BX_SB16_THIS devices->pic->untrigger_irq(BX_SB16_IRQ);
+      if (MIXER.reg[0x82] == 0) {
+        DSP.irqpending = 0;
+        pluginLowerIRQ(BX_SB16_IRQ);
+      }
       writelog( WAVELOG(4), "16-bit DMA IRQ acknowledged");
     }
   else
@@ -1053,7 +1093,7 @@ void bx_sb16_c::dsp_getwavepacket()
 // write a wave packet to the output device
 void bx_sb16_c::dsp_sendwavepacket()
 {
-  switch (bx_options.sb16.wavemode)
+  switch (bx_options.sb16.Owavemode->get ())
     {
     case 1:
       BX_SB16_OUTPUT->sendwavepacket(DSP.dma.chunkindex, DSP.dma.chunk);
@@ -1115,13 +1155,13 @@ void bx_sb16_c::dsp_dmadone()
     {
       dsp_sendwavepacket();  // flush the output
 
-      if (bx_options.sb16.wavemode == 1)
+      if (bx_options.sb16.Owavemode->get () == 1)
 	{
 	  if (DSP.dma.mode != 2)
 	    BX_SB16_OUTPUT->stopwaveplayback();  // don't stop if Auto-DMA
 	}
 
-      else if (bx_options.sb16.wavemode == 2)
+      else if (bx_options.sb16.Owavemode->get () == 2)
 	fflush(WAVEDATA);
     }
 
@@ -1131,7 +1171,7 @@ void bx_sb16_c::dsp_dmadone()
  else
     MIXER.reg[0x82] |= 2;
 
-  BX_SB16_THIS devices->pic->trigger_irq(BX_SB16_IRQ);
+  pluginRaiseIRQ(BX_SB16_IRQ);
   DSP.irqpending = 1;
 
   //if auto-DMA, reinitialize
@@ -1152,7 +1192,7 @@ void bx_sb16_c::dsp_dmadone()
 // and write = from soundcard to application (input)
 void bx_sb16_c::dma_read8(Bit8u *data_byte)
 {
-  bx_pc_system.set_DRQ(BX_SB16_DMAL, 0);  // the timer will raise it again
+  pluginDMASetDRQ(BX_SB16_DMAL, 0);  // the timer will raise it again
 
   if (DSP.dma.count % 100 == 0) // otherwise it's just too many lines of log
     writelog( WAVELOG(5), "Received 8-bit DMA %2x, %d remaining ",
@@ -1167,7 +1207,7 @@ void bx_sb16_c::dma_read8(Bit8u *data_byte)
 
 void bx_sb16_c::dma_write8(Bit8u *data_byte)
 {
-  bx_pc_system.set_DRQ(BX_SB16_DMAL, 0);  // the timer will raise it again
+  pluginDMASetDRQ(BX_SB16_DMAL, 0);  // the timer will raise it again
 
   DSP.dma.count--;
 
@@ -1183,13 +1223,13 @@ void bx_sb16_c::dma_write8(Bit8u *data_byte)
 
 void bx_sb16_c::dma_read16(Bit16u *data_word)
 {
-  bx_pc_system.set_DRQ(BX_SB16_DMAH, 0);  // the timer will raise it again
+  pluginDMASetDRQ(BX_SB16_DMAH, 0);  // the timer will raise it again
 
   if (DSP.dma.count % 100 == 0) // otherwise it's just too many lines of log
     writelog( WAVELOG(5), "Received 16-bit DMA %4x, %d remaining ",
 	      *data_word, DSP.dma.count);
 
-  DSP.dma.count -= 2;
+  DSP.dma.count--;
 
   dsp_getsamplebyte(*data_word & 0xff);
   dsp_getsamplebyte(*data_word >> 8);
@@ -1202,9 +1242,9 @@ void bx_sb16_c::dma_write16(Bit16u *data_word)
 {
   Bit8u byte1, byte2;
 
-  bx_pc_system.set_DRQ(BX_SB16_DMAH, 0);  // the timer will raise it again
+  pluginDMASetDRQ(BX_SB16_DMAH, 0);  // the timer will raise it again
 
-  DSP.dma.count -= 2;
+  DSP.dma.count--;
 
   byte1 = dsp_putsamplebyte();
   byte2 = dsp_putsamplebyte();
@@ -1238,12 +1278,8 @@ void bx_sb16_c::mixer_writedata(Bit32u value)
     case 0:  // initialize mixer
 
       writelog(BOTHLOG(4), "Initializing mixer...");
-      for (i=0; i<BX_SB16_MIX_REG; i++)
+      for (i=0; i<0x80; i++)
 	MIXER.reg[i] = 0;
-      MIXER.reg[0x80] = 2;  // IRQ 5
-      MIXER.reg[0x81] = 2;  // 8-bit DMA 1, no 16-bit DMA
-
-      set_irq_dma();        // set the IRQ and DMA
 
       MIXER.regindex = 0;   // next mixer register read is register 0
       break;
@@ -1276,7 +1312,9 @@ void bx_sb16_c::mixer_writeregister(Bit32u value)
 
 void bx_sb16_c::set_irq_dma()
 {
+  static Boolean isInitialized=0;
   int newirq;
+  int oldDMA8, oldDMA16;
 
   // set the IRQ according to the value in mixer register 0x80
   switch (MIXER.reg[0x80]) 
@@ -1309,6 +1347,7 @@ void bx_sb16_c::set_irq_dma()
     }
 
   // set the 8 bit DMA
+  oldDMA8=BX_SB16_DMAL;
   switch (MIXER.reg[0x81] & 0x0f)
     {
     case 1: 
@@ -1328,7 +1367,20 @@ void bx_sb16_c::set_irq_dma()
       MIXER.reg[0x81] |= (1 << BX_SB16_DMAL);
     }
 
+  // Unregister the previous DMA if initialized
+  if ( (isInitialized) && (oldDMA8 != BX_SB16_DMAL) ) {
+    pluginUnregisterDMAChannel(oldDMA8);
+    BX_INFO(("Unregistering 8bits DMA Channel %d for sb16",oldDMA8));
+  }
+
+  // And register the new 8bits DMA Channel
+  if ( (!isInitialized) || (oldDMA8 != BX_SB16_DMAL) ) {
+    pluginRegisterDMAChannel(BX_SB16_DMAL, bx_sb16.dma_read8, bx_sb16.dma_write8);
+    BX_INFO(("Registering 8bits DMA Channel %d for sb16",BX_SB16_DMAL));
+  }
+
   // and the 16 bit DMA
+  oldDMA16=BX_SB16_DMAH;
   switch (MIXER.reg[0x81] >> 4)
     {
     case 0:
@@ -1352,6 +1404,23 @@ void bx_sb16_c::set_irq_dma()
       // no default 16 bit channel!
     }
 
+  // Unregister the previous DMA if initialized
+  if ( (isInitialized) && (oldDMA16 != 0) && (oldDMA16 != BX_SB16_DMAH) ) {
+    pluginUnregisterDMAChannel(oldDMA16);
+    BX_INFO(("Unregistering 16bits DMA Channel %d for sb16",oldDMA16));
+  }
+
+  // And register the new 16bits DMA Channel
+  if ( (BX_SB16_DMAH != 0) && (oldDMA16 != BX_SB16_DMAH) ) {
+    pluginRegisterDMA16Channel(BX_SB16_DMAH, bx_sb16.dma_read16, bx_sb16.dma_write16);
+    BX_INFO(("Registering 16bits DMA Channel %d for sb16",BX_SB16_DMAH));
+  }
+
+  // If not already initialized
+  if(!isInitialized) {
+    isInitialized=1;
+  }
+
   writelog(BOTHLOG(4), "Resources set to I%d D%d H%d", 
 	   BX_SB16_IRQ, BX_SB16_DMAL, BX_SB16_DMAH);
 
@@ -1369,7 +1438,7 @@ Bit32u bx_sb16_c::mpu_status()
   Bit32u result = 0;
 
   if ( (MPU.datain.full() == 1) ||
-       ( (bx_options.sb16.midimode == 1) &&
+       ( (bx_options.sb16.Omidimode->get () == 1) &&
 	 (BX_SB16_OUTPUT->midiready() == BX_SOUND_OUTPUT_ERR) ) )
     result |= 0x40;	// output not ready
   if (MPU.dataout.empty() == 1)
@@ -1420,7 +1489,7 @@ void bx_sb16_c::mpu_command(Bit32u value)
 	  if (BX_SB16_IRQMPU != -1)
 	    {
 	      MIXER.reg[0x82] |= 4;
-	      BX_SB16_THIS devices->pic->trigger_irq(BX_SB16_IRQMPU);
+	      pluginRaiseIRQ(BX_SB16_IRQMPU);
 	    }
 	  break;
 
@@ -1483,7 +1552,7 @@ Bit32u bx_sb16_c::mpu_dataread()
 	MPU.irqpending = 0;
 	MIXER.reg[0x82] &= (~4);
 	if (MIXER.reg[0x82] == 0)
-	  BX_SB16_THIS devices->pic->untrigger_irq(BX_SB16_IRQMPU);
+	  pluginLowerIRQ(BX_SB16_IRQMPU);
 	writelog(MIDILOG(4), "MPU IRQ acknowledged");
       }
 
@@ -1742,27 +1811,27 @@ void bx_sb16_c::emul_write(Bit32u value)
 	    }
 	  break;
 	case 6: // close midi and wave files and/or output
-	  if ( (bx_options.sb16.midimode == 2) ||
-	       (bx_options.sb16.midimode == 3) )
+	  if ( (bx_options.sb16.Omidimode->get () == 2) ||
+	       (bx_options.sb16.Omidimode->get () == 3) )
 	    {
-	      if (bx_options.sb16.midimode == 2)
+	      if (bx_options.sb16.Omidimode->get () == 2)
 		finishmidifile();
 	      fclose(MIDIDATA);
 	    }
-	  else if (bx_options.sb16.midimode == 1)
+	  else if (bx_options.sb16.Omidimode->get () == 1)
 	    BX_SB16_OUTPUT->closemidioutput();
-	  bx_options.sb16.midimode = 0;
+	  bx_options.sb16.Omidimode->set (0);
 	  
-	  if ( (bx_options.sb16.wavemode == 2) ||
-	       (bx_options.sb16.wavemode == 3) )
+	  if ( (bx_options.sb16.Owavemode->get () == 2) ||
+	       (bx_options.sb16.Owavemode->get () == 3) )
 	    {
-	      if (bx_options.sb16.wavemode == 2)
+	      if (bx_options.sb16.Owavemode->get () == 2)
 		finishvocfile();
 	      fclose(WAVEDATA);
 	    }
 	  else
 	    BX_SB16_OUTPUT->closewaveoutput();
-	  bx_options.sb16.wavemode = 0;
+	  bx_options.sb16.Owavemode->set (0);
 
 	  break;
 	case 7: // clear bank/program mappings
@@ -2251,7 +2320,7 @@ break_here:
       if (channum != -1)
 	{
 	  int needchange = 0;
-	  if ((OPL.oper[OPL.chan[channum].opnum[0]][4] & 1) != (value & 1))
+	  if ((OPL.oper[OPL.chan[channum].opnum[0]][4] & 1) != (int)(value & 1))
 	    needchange = 1;
 
 	  opl_changeop(channum, OPL.chan[channum].opnum[0], 4, value & 0x3f);
@@ -2536,11 +2605,11 @@ void bx_sb16_c::initmidifile()
     Bit16u timecode;  // 0x80 + deltatimesperquarter << 8
     } midiheader = 
 #ifdef BX_LITTLE_ENDIAN
-      { "MTh", 0x06000000, 0x0100, 0x0100, 0x8001 };
+      { "MTh", 0x06000000, 0, 0x0100, 0x8001 };
 #else
-      { "MTh", 6, 1, 1, 0x180 };
+      { "MTh", 6, 0, 1, 0x180 };
 #endif
-   midiheader.chunk[3] = 'k';
+   midiheader.chunk[3] = 'd';
 
   struct {
     Bit8u chunk[4];
@@ -2569,28 +2638,28 @@ void bx_sb16_c::writemidicommand(int command, int length, Bit8u data[])
   int deltatime = currentdeltatime();
 
   /* Initialize output device if necessary and not done yet */
-  if (bx_options.sb16.midimode == 1)
+  if (bx_options.sb16.Omidimode->get () == 1)
     {
       if (MPU.outputinit != 1)
 	{
 	  writelog( MIDILOG(4), "Initializing Midi output.");
-	  if (BX_SB16_OUTPUT->openmidioutput(bx_options.sb16.midifile) == BX_SOUND_OUTPUT_OK)
+	  if (BX_SB16_OUTPUT->openmidioutput(bx_options.sb16.Omidifile->getptr ()) == BX_SOUND_OUTPUT_OK)
 	    MPU.outputinit = 1;
 	  else 
 	    MPU.outputinit = 0;
 	  if (MPU.outputinit != 1)
 	    {
 	      writelog( MIDILOG(2), "Error: Couldn't open midi output. Midi disabled.");
-	      bx_options.sb16.midimode = 0;
+	      bx_options.sb16.Omidimode->set (0);
 	    }
 	}
       BX_SB16_OUTPUT->sendmidicommand(deltatime, command, length, data); 
       return;
     }
-  else if (bx_options.sb16.midimode < 2)
+  else if (bx_options.sb16.Omidimode->get () < 2)
     return;
 
-  if (bx_options.sb16.midimode == 2)
+  if (bx_options.sb16.Omidimode->get () == 2)
     writedeltatime(deltatime);
 
   fputc(command, MIDIDATA);
@@ -2776,7 +2845,6 @@ void bx_sb16_c::writedeltatime(Bit32u deltatime)
 
 void bx_sb16_c::finishmidifile()
 {
-  fpos_t tracklen;
   struct { 
     Bit8u delta, statusbyte, metaevent, length;
   } metatrackend = { 0, 0xff, 0x2f, 0 };
@@ -2784,19 +2852,20 @@ void bx_sb16_c::finishmidifile()
      // Meta event track end (0xff 0x2f 0x00) plus leading delta time
   fwrite(&metatrackend, 1, sizeof metatrackend, MIDIDATA );
 
-  fgetpos(MIDIDATA, &tracklen);
+  Bit32u tracklen = ftell(MIDIDATA);
+  if (tracklen < 0)
+    BX_PANIC (("ftell failed in finishmidifile"));
+  if (tracklen < 22)
+    BX_PANIC (("finishmidifile with track length too short"));
   tracklen -= 22;    // subtract the midi file and track header
-
   fseek(MIDIDATA, 22 - 4, SEEK_SET);
-
-     // value has to be in big endian
+  // value has to be in big endian
 #ifdef BX_LITTLE_ENDIAN
   tracklen = (tracklen << 24) | (tracklen >> 24) |
-        ((tracklen & 0x00ff0000) >> 8) |
-        ((tracklen & 0x0000ff00) << 8);
+	((tracklen & 0x00ff0000) >> 8) |
+	((tracklen & 0x0000ff00) << 8);
 #endif
   fwrite(&tracklen, 4, 1, MIDIDATA);
-
   return;
 }
 
@@ -3130,12 +3199,12 @@ void bx_sb16_c::write(Bit32u address, Bit32u value, unsigned io_len)
 void bx_sb16_c::writelog(int loglevel, const char *str, ...)
 {
 // append a line to the log file, if desired
-  if ( (int) bx_options.sb16.loglevel >= loglevel)
+  if ( (int) bx_options.sb16.Ologlevel->get () >= loglevel)
     {
 	time_t timep = time(NULL);
-	tm *t = localtime(&timep);
-	fprintf(LOGFILE, "SB16 %02d:%02d:%02d (%i): ",
-		t->tm_hour, t->tm_min, t->tm_sec, loglevel);
+        tm *t = localtime(&timep);
+        fprintf(LOGFILE, "SB16 %02d:%02d:%02d (%i): ",
+			-               t->tm_hour, t->tm_min, t->tm_sec, loglevel);
 	va_list ap;
 	va_start(ap, str);
 	vfprintf(LOGFILE, str, ap);
@@ -3178,7 +3247,7 @@ void bx_sb16_buffer::reset()
 bx_sb16_buffer::~bx_sb16_buffer(void)
 {
   if (buffer != NULL)
-    delete buffer;
+    delete[] buffer;
 
   buffer = NULL;
   length = 0;
@@ -3227,7 +3296,7 @@ Boolean bx_sb16_buffer::puts(char *data, ...)
   va_end(ap);
 
   if ( (int) strlen(string) >= length)
-    bx_panic("bx_sb16_buffer: puts() too long!\n");
+    BX_PANIC(("bx_sb16_buffer: puts() too long!"));
 
   while (string[index] != 0)
     {
