@@ -1,4 +1,10 @@
-//  Copyright (C) 2000  MandrakeSoft S.A.
+/////////////////////////////////////////////////////////////////////////
+// $Id: sb16.h,v 1.1.1.2 2002/07/20 18:28:27 DemonLord Exp $
+/////////////////////////////////////////////////////////////////////////
+// Based on Bochs file <Id: sb16.h,v 1.6 2001/10/03 13:10:38 bdenney Exp >
+/////////////////////////////////////////////////////////////////////////
+//
+//  Copyright (C) 2001  MandrakeSoft S.A.
 //
 //    MandrakeSoft S.A.
 //    43, rue d'Aboukir
@@ -22,6 +28,8 @@
 
 // This file (SB16.H) written and donated by Josef Drexler
 
+#ifndef _BX_SB16_H_
+#define _BX_SB16_H_
 
 #if BX_USE_SB16_SMF
 #  define BX_SB16_SMF   static
@@ -30,7 +38,7 @@
 #else
 #  define BX_SB16_SMF
 #  define BX_SB16_THIS  this->
-#  define BX_SB16_THISU (this)
+#  define BX_SB16_THISP (this)
 #endif
 
 #if BX_USE_SOUND_VIRTUAL
@@ -42,7 +50,8 @@
 #endif
 
 // If the buffer commands are to be inlined:
-#define BX_SB16_BUFINL inline
+#define BX_SB16_BUFINL BX_CPP_INLINE
+// BX_CPP_INLINE is defined to the inline keyword for the C++ compiler.
 
 // maximum number of patch translations
 #define BX_SB16_PATCHTABLESIZE 1024
@@ -54,10 +63,10 @@
 #define BX_SB16_IOMPULEN 4          // number of addresses covered
 #define BX_SB16_IOADLIB 0x388       // equivalent to 0x220..0x223 and 0x228..0x229
 #define BX_SB16_IOADLIBLEN 4        // number of addresses covered
-#define BX_SB16_IRQ     BX_SB16_THIS currentirq
+#define BX_SB16_IRQ     bx_sb16.currentirq
 #define BX_SB16_IRQMPU  BX_SB16_IRQ // IRQ for the MPU401 part - same value
-#define BX_SB16_DMAL    BX_SB16_THIS currentdma8
-#define BX_SB16_DMAH    BX_SB16_THIS currentdma16
+#define BX_SB16_DMAL    bx_sb16.currentdma8
+#define BX_SB16_DMAH    bx_sb16.currentdma16
 
 /*
    A few notes:
@@ -184,12 +193,13 @@ private:
 class BX_SOUND_OUTPUT_C_DEF;
 
 // The actual emulator class, emulating the sound blaster ports
-class bx_sb16_c {
+class bx_sb16_c : public logfunctions {
 public:
 
   bx_sb16_c(void);
   ~bx_sb16_c(void);
-  BX_SB16_SMF void init(bx_devices_c *);
+  //BX_SB16_SMF void init(bx_devices_c *);
+  BX_SB16_SMF void init(void);
 
       /* DMA input and output, 8 and 16 bit, have to be public */
   BX_SB16_SMF void   dma_write8(Bit8u *data_byte);
@@ -297,7 +307,6 @@ private:
     int remaps;
   } emuldata;
 
-  bx_devices_c *devices;
 
       /* the MPU 401 part of the emulator */
   BX_SB16_SMF Bit32u mpu_status();                   // read status port   3x1
@@ -305,7 +314,7 @@ private:
   BX_SB16_SMF Bit32u mpu_dataread();                 // read data port     3x0
   BX_SB16_SMF void   mpu_datawrite(Bit32u value);    // write data port    3x0
   BX_SB16_SMF void   mpu_mididata(Bit32u value);     // get a midi byte
-  BX_SB16_SMF void   mpu_timer (void *);
+  static void   mpu_timer (void *);
 
       /* The DSP part */
   BX_SB16_SMF void   dsp_reset(Bit32u value);        // write to reset port 2x6
@@ -323,7 +332,7 @@ private:
   BX_SB16_SMF void   dsp_dmadone();		     // stop a DMA transfer
   BX_SB16_SMF void   dsp_enabledma();		     // enable the transfer
   BX_SB16_SMF void   dsp_disabledma();		     // temporarily disable DMA
-  BX_SB16_SMF void   dsp_dmatimer (void *);
+  static void   dsp_dmatimer (void *);
 
       /* The mixer part */
   BX_SB16_SMF Bit32u mixer_readdata(void);
@@ -340,7 +349,7 @@ private:
   BX_SB16_SMF Bit32u opl_status(int chipid);
   BX_SB16_SMF void   opl_index(Bit32u value, int chipid);
   BX_SB16_SMF void   opl_data(Bit32u value, int chipid);
-  BX_SB16_SMF void   opl_timer(void *);
+  static void   opl_timer(void *);
   BX_SB16_SMF void   opl_changeop(int channum, int opernum, int byte, int value);
   BX_SB16_SMF void   opl_settimermask(int value, int chipid);
   BX_SB16_SMF void   opl_set4opmode(int new4opmode);
@@ -369,8 +378,8 @@ private:
 
       /* The port IO multiplexer functions */
 
-  BX_SB16_SMF Bit32u read_handler(void *this_ptr, Bit32u address, unsigned io_len);
-  BX_SB16_SMF void   write_handler(void *this_ptr, Bit32u address, Bit32u value, unsigned io_len);
+  static Bit32u read_handler(void *this_ptr, Bit32u address, unsigned io_len);
+  static void   write_handler(void *this_ptr, Bit32u address, Bit32u value, unsigned io_len);
 
 #if !BX_USE_SB16_SMF
   Bit32u read(Bit32u address, unsigned io_len);
@@ -379,7 +388,7 @@ private:
 };
 
 // The class with the output functions
-class bx_sound_output_c {
+class bx_sound_output_c : public logfunctions {
 public:
 
       /* These functions are the sound output functions, sending
@@ -404,9 +413,14 @@ public:
   BX_SOUND_VIRTUAL int    closewaveoutput();
 };
 
-#if BX_USE_SB16_SMF
 extern bx_sb16_c bx_sb16;
-#endif
+
+#define WRITELOG        sb16->writelog
+#define BOTHLOG(x)      (x)
+#define MIDILOG(x)      ((bx_options.sb16.Omidimode->get ()>0?x:0x7f))
+#define WAVELOG(x)      ((bx_options.sb16.Owavemode->get ()>0?x:0x7f))
 
 #include "soundlnx.h"
 #include "soundwin.h"
+
+#endif // _BX_SB16_H_
