@@ -1,4 +1,10 @@
-//  Copyright (C) 2000  MandrakeSoft S.A.
+/////////////////////////////////////////////////////////////////////////
+// $Id: serial.cc,v 1.2 2002/07/21 23:14:55 DemonLord Exp $
+/////////////////////////////////////////////////////////////////////////
+// Based on Bochs file <Id: serial.cc,v 1.21 2002/06/04 21:35:08 vruppert Exp >
+/////////////////////////////////////////////////////////////////////////
+//
+//  Copyright (C) 2002  MandrakeSoft S.A.
 //
 //    MandrakeSoft S.A.
 //    43, rue d'Aboukir
@@ -33,6 +39,7 @@
 // to the serial port /AM
 
 #include "bochs.h"
+#define LOG_THIS bx_serial.
 #include "serial.h"
 
 #if USE_RAW_SERIAL
@@ -47,8 +54,11 @@
 #endif
 #endif
 
+#if defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__linux__) || defined(__GNU__)
+#define SERIAL_ENABLE
+#endif
 
-#ifdef __FreeBSD__
+#ifdef SERIAL_ENABLE
 extern "C" {
 #include <termios.h>
 };
@@ -59,7 +69,7 @@ bx_serial_c bx_serial;
 #define this (&bx_serial)
 #endif
 
-#ifdef __FreeBSD__
+#ifdef SERIAL_ENABLE
 static struct termios term_orig, term_new;
 #endif
 
@@ -89,8 +99,37 @@ plugin_fini(void)
 
 bx_serial_c::bx_serial_c(void)
 {
-#ifdef __FreeBSD__
-  tcgetattr(0, &term_orig);
+  put("SER");
+  settype(SERLOG);
+}
+
+bx_serial_c::~bx_serial_c(void)
+{
+#ifdef SERIAL_ENABLE
+  if (bx_options.com1.Opresent->get ())
+    tcsetattr(tty_id, TCSAFLUSH, &term_orig);
+#endif
+  // nothing for now
+}
+
+
+  void
+#ifdef PLEX86
+bx_serial_c::init(void)
+#else
+bx_serial_c::init(bx_devices_c *d)
+#endif
+{
+  if (!bx_options.com1.Opresent->get ())
+    return;
+
+#ifdef SERIAL_ENABLE
+  tty_id = open(bx_options.com1.Odev->getptr (), O_RDWR|O_NONBLOCK,600);
+  if (tty_id < 0)
+    BX_PANIC(("open of %s (%s) failed\n",
+	      "com1", bx_options.com1.Odev->getptr ()));
+  BX_DEBUG(("tty_id: %d",tty_id));
+  tcgetattr(tty_id, &term_orig);
   bcopy((caddr_t) &term_orig, (caddr_t) &term_new, sizeof(struct termios));
   cfmakeraw(&term_new);
   term_new.c_oflag |= OPOST | ONLCR;  // Enable NL to CR-NL translation
@@ -103,41 +142,35 @@ bx_serial_c::bx_serial_c(void)
   // ctl-C will be delivered to the serial port
   term_new.c_iflag |= IGNBRK;
   term_new.c_iflag &= ~BRKINT;
-#endif
-  term_new.c_iflag |= IXOFF;
-  tcsetattr(0, TCSAFLUSH, &term_new);
-#endif
+#endif    /* !def TRUE_CTLC */
+  term_new.c_iflag = 0;
+  term_new.c_oflag = 0;
+  term_new.c_cflag = CS8|CREAD|CLOCAL;
+  term_new.c_lflag = 0;
+  term_new.c_cc[VMIN] = 1;
+  term_new.c_cc[VTIME] = 0;
+  //term_new.c_iflag |= IXOFF;
+  tcsetattr(tty_id, TCSAFLUSH, &term_new);
+#endif   /* def SERIAL_ENABLE */
   // nothing for now
 #if USE_RAW_SERIAL
-  raw = new serial_raw("/dev/cua0", SIGUSR1);
+  this->raw = new serial_raw("/dev/cua0", SIGUSR1);
 #endif // USE_RAW_SERIAL
-}
 
-bx_serial_c::~bx_serial_c(void)
-{
-#ifdef __FreeBSD__
-  tcsetattr(0, TCSAFLUSH, &term_orig);
+#ifndef PLEX86
+  BX_SER_THIS devices = d;
 #endif
-  // nothing for now
-#if USE_RAW_SERIAL
-  delete raw;
-#endif // USE_RAW_SERIAL
-}
 
-
-  void
-bx_serial_c::init(void)
-{
   pluginRegisterIRQ(4, "Serial Port 1");
 
 #if defined (USE_TTY_HACK)
   tty_id = tty_alloc("Bx Serial Console, Your Window to the 8250");
   if (tty_id > 0)
-    bx_printf("TTY Allocated fd = %d\n", tty_get_fd(tty_id));
+    BX_INFO(("TTY allocated fd = %d", tty_get_fd(tty_id)));
   else
-    bx_printf("TTY allocation failed\n");
+    BX_INFO(("TTY allocation failed"));
 #else
-  bx_printf("TTY not used, serial port is not connected\n");
+  //BX_INFO(("TTY not used, serial port is not connected"));
 #endif
 
   /*
@@ -147,6 +180,12 @@ bx_serial_c::init(void)
     /* internal state */
     BX_SER_THIS s[i].rx_empty = 1;
     BX_SER_THIS s[i].tx_empty = 1;
+    BX_SER_THIS s[i].ls_ipending = 0;
+    BX_SER_THIS s[i].ms_ipending = 0;
+    BX_SER_THIS s[i].rx_ipending = 0;
+    BX_SER_THIS s[i].tx_ipending = 0;
+    BX_SER_THIS s[i].ls_interrupt = 0;
+    BX_SER_THIS s[i].ms_interrupt = 0;
     BX_SER_THIS s[i].rx_interrupt = 0;
     BX_SER_THIS s[i].tx_interrupt = 0;
 
@@ -219,6 +258,7 @@ bx_serial_c::init(void)
   }
 
   for (unsigned addr=0x03F8; addr<=0x03FF; addr++) {
+	BX_DEBUG(("register read/write: 0x%x",addr));
     pluginRegisterIOReadHandler(this,
        read_handler,
        addr, "Serial Port 1", 1);
@@ -227,6 +267,8 @@ bx_serial_c::init(void)
        addr, "Serial Port 1", 1);
     }
 
+  BX_INFO(( "com1 at 0x3f8/8 irq 4" ));
+
 }
 
 
@@ -250,46 +292,46 @@ bx_serial_c::read(Bit32u address, unsigned io_len)
 #else
   UNUSED(this_ptr);
 #endif  // !BX_USE_SER_SMF
-  UNUSED(address);
+  //UNUSED(address);
   Bit8u val;
 
   /* SERIAL PORT 1 */
 
   if (io_len > 1)
-    bx_panic("serial: io read from port %04x, bad len=%u\n",
+    BX_PANIC(("io read from port %04x, bad len=%u",
 	     (unsigned) address,
-             (unsigned) io_len);
+             (unsigned) io_len));
 
-  if (bx_dbg.serial)
-    bx_printf("serial register read from address 0x%x - ",
-	      (unsigned) address);
+  BX_DEBUG(("register read from address 0x%x - ", (unsigned) address));
 
   switch (address) {
     case 0x03F8: /* receive buffer, or divisor latch LSB if DLAB set */
       if (BX_SER_THIS s[0].line_cntl.dlab) {
-	val = BX_SER_THIS s[0].divisor_lsb;
+        val = BX_SER_THIS s[0].divisor_lsb;
       } else {
-	val = BX_SER_THIS s[0].rxbuffer;
-	BX_SER_THIS s[0].line_status.rxdata_ready = 0;
-	BX_SER_THIS s[0].rx_empty = 1;
+        val = BX_SER_THIS s[0].rxbuffer;
+        BX_SER_THIS s[0].line_status.rxdata_ready = 0;
+        BX_SER_THIS s[0].rx_empty = 1;
 
-	/* If there are no more ints pending, clear the irq */
-	if ((BX_SER_THIS s[0].rx_interrupt == 1) &&
-	    (BX_SER_THIS s[0].tx_interrupt == 0)) {
-	  pluginUntriggerIRQ(4);
-	}
-	BX_SER_THIS s[0].rx_interrupt = 0;
+        /* If there are no more ints pending, clear the irq */
+        if ((BX_SER_THIS s[0].tx_interrupt == 0) &&
+            (BX_SER_THIS s[0].ls_interrupt == 0) &&
+            (BX_SER_THIS s[0].ms_interrupt == 0)) {
+          pluginLowerIRQ(4);
+        }
+        BX_SER_THIS s[0].rx_interrupt = 0;
+        BX_SER_THIS s[0].rx_ipending = 0;
       }
       break;
 
     case 0x03F9: /* interrupt enable register, or div. latch MSB */
       if (BX_SER_THIS s[0].line_cntl.dlab) {
-	val = BX_SER_THIS s[0].divisor_msb;
+        val = BX_SER_THIS s[0].divisor_msb;
       } else {
-	val = BX_SER_THIS s[0].int_enable.rxdata_enable     |
-	  (BX_SER_THIS s[0].int_enable.txhold_enable  << 1) |
-	  (BX_SER_THIS s[0].int_enable.rxlstat_enable << 2) |
-	  (BX_SER_THIS s[0].int_enable.modstat_enable << 3);	
+        val = BX_SER_THIS s[0].int_enable.rxdata_enable |
+              (BX_SER_THIS s[0].int_enable.txhold_enable  << 1) |
+              (BX_SER_THIS s[0].int_enable.rxlstat_enable << 2) |
+              (BX_SER_THIS s[0].int_enable.modstat_enable << 3);
       }
       break;
 
@@ -297,27 +339,36 @@ bx_serial_c::read(Bit32u address, unsigned io_len)
       /*
        * Set the interrupt ID based on interrupt source
        */
-      if (BX_SER_THIS s[0].rx_interrupt) {
-	BX_SER_THIS s[0].int_ident.int_ID = 0x2;
-	BX_SER_THIS s[0].int_ident.ipending = 0;
+      if (BX_SER_THIS s[0].ls_interrupt) {
+        BX_SER_THIS s[0].int_ident.int_ID = 0x3;
+        BX_SER_THIS s[0].int_ident.ipending = 0;
+      } else if (BX_SER_THIS s[0].rx_interrupt) {
+        BX_SER_THIS s[0].int_ident.int_ID = 0x2;
+        BX_SER_THIS s[0].int_ident.ipending = 0;
       } else if (BX_SER_THIS s[0].tx_interrupt) {
-	BX_SER_THIS s[0].int_ident.int_ID = 0x1;
-	BX_SER_THIS s[0].int_ident.ipending = 0;
+        BX_SER_THIS s[0].int_ident.int_ID = 0x1;
+        BX_SER_THIS s[0].int_ident.ipending = 0;
+      } else if (BX_SER_THIS s[0].ms_interrupt) {
+        BX_SER_THIS s[0].int_ident.int_ID = 0x0;
+        BX_SER_THIS s[0].int_ident.ipending = 0;
       } else {
-	BX_SER_THIS s[0].int_ident.int_ID = 0x0;
-	BX_SER_THIS s[0].int_ident.ipending = 1;
+        BX_SER_THIS s[0].int_ident.int_ID = 0x0;
+        BX_SER_THIS s[0].int_ident.ipending = 1;
+      }
+
+      /* If there are no more ints pending, clear the irq */
+      if ((BX_SER_THIS s[0].rx_interrupt == 0) &&
+          (BX_SER_THIS s[0].ls_interrupt == 0) &&
+          (BX_SER_THIS s[0].ms_interrupt == 0)) {
+        pluginLowerIRQ(4);
       }
 
       BX_SER_THIS s[0].tx_interrupt = 0;
-
-      /* no more ints pending, clear the irq */
-      if (BX_SER_THIS s[0].int_ident.int_ID == 0x1) {
-	pluginUntriggerIRQ(4);
-      }
+      BX_SER_THIS s[0].tx_ipending = 0;
 
       val = BX_SER_THIS s[0].int_ident.ipending  |
-	(BX_SER_THIS s[0].int_ident.int_ID << 1) |
-	(BX_SER_THIS s[0].int_ident.fifo_enabled << 6);
+            (BX_SER_THIS s[0].int_ident.int_ID << 1) |
+            (BX_SER_THIS s[0].int_ident.fifo_enabled << 6);
       break;
 
     case 0x03FB: /* Line control register */
@@ -331,11 +382,11 @@ bx_serial_c::read(Bit32u address, unsigned io_len)
       break;
 
     case 0x03FC: /* MODEM control register */
-      val = BX_SER_THIS s[0].modem_cntl.dtr     |
-	(BX_SER_THIS s[0].modem_cntl.rts << 1)  |
-	(BX_SER_THIS s[0].modem_cntl.out1 << 2) |
-	(BX_SER_THIS s[0].modem_cntl.out2 << 3) |
-	(BX_SER_THIS s[0].modem_cntl.local_loopback);
+      val = BX_SER_THIS s[0].modem_cntl.dtr |
+            (BX_SER_THIS s[0].modem_cntl.rts << 1) |
+            (BX_SER_THIS s[0].modem_cntl.out1 << 2) |
+            (BX_SER_THIS s[0].modem_cntl.out2 << 3) |
+            (BX_SER_THIS s[0].modem_cntl.local_loopback << 4);
       break;
 
     case 0x03FD: /* Line status register */
@@ -347,7 +398,16 @@ bx_serial_c::read(Bit32u address, unsigned io_len)
 	(BX_SER_THIS s[0].line_status.txhold_empty   << 5) |
 	(BX_SER_THIS s[0].line_status.txtransm_empty << 6) |
 	(BX_SER_THIS s[0].line_status.fifo_error     << 7);
+      BX_SER_THIS s[0].line_status.overrun_error = 0;
       BX_SER_THIS s[0].line_status.break_int = 0;
+      /* If there are no more ints pending, clear the irq */
+      if ((BX_SER_THIS s[0].rx_interrupt == 0) &&
+          (BX_SER_THIS s[0].tx_interrupt == 0) &&
+          (BX_SER_THIS s[0].ms_interrupt == 0)) {
+        pluginLowerIRQ(4);
+      }
+      BX_SER_THIS s[0].ls_interrupt = 0;
+      BX_SER_THIS s[0].ls_ipending = 0;
       break;
 
     case 0x03FE: /* MODEM status register */
@@ -359,22 +419,32 @@ bx_serial_c::read(Bit32u address, unsigned io_len)
 	(BX_SER_THIS s[0].modem_status.dsr          << 5) |
 	(BX_SER_THIS s[0].modem_status.ri           << 6) |
 	(BX_SER_THIS s[0].modem_status.dcd          << 7);
+      BX_SER_THIS s[0].modem_status.delta_cts = 0;
+      BX_SER_THIS s[0].modem_status.delta_dsr = 0;
+      BX_SER_THIS s[0].modem_status.ri_trailedge = 0;
+      BX_SER_THIS s[0].modem_status.delta_dcd = 0;
+      /* If there are no more ints pending, clear the irq */
+      if ((BX_SER_THIS s[0].rx_interrupt == 0) &&
+          (BX_SER_THIS s[0].tx_interrupt == 0) &&
+          (BX_SER_THIS s[0].ls_interrupt == 0)) {
+        pluginLowerIRQ(4);
+      }
+      BX_SER_THIS s[0].ms_interrupt = 0;
+      BX_SER_THIS s[0].ms_ipending = 0;
       break;
 
     case 0x03FF: /* scratch register */
-      val = BX_SER_THIS s[0].scratch = 0;
+      val = BX_SER_THIS s[0].scratch;
       break;
 
     default:
       val = 0; // keep compiler happy
-      bx_panic("unsupported serial io read from address=%0x%x!\n",
-        (unsigned) address);
+      BX_PANIC(("unsupported io read from address=%0x%x!",
+        (unsigned) address));
       break;
   }
 
-  if (bx_dbg.serial)
-    bx_printf("val =  0x%x\n",
-	      (unsigned) val);
+  BX_DEBUG(("val =  0x%x", (unsigned) val));
 
   return(val);
 }
@@ -383,7 +453,7 @@ bx_serial_c::read(Bit32u address, unsigned io_len)
   // static IO port write callback handler
   // redirects to non-static class handler to avoid virtual functions
 
-  void
+void
 bx_serial_c::write_handler(void *this_ptr, Bit32u address, Bit32u value, unsigned io_len)
 {
 #if !BX_USE_SER_SMF
@@ -392,26 +462,29 @@ bx_serial_c::write_handler(void *this_ptr, Bit32u address, Bit32u value, unsigne
   class_ptr->write(address, value, io_len);
 }
 
-  void
+void
 bx_serial_c::write(Bit32u address, Bit32u value, unsigned io_len)
 {
 #else
   UNUSED(this_ptr);
 #endif  // !BX_USE_SER_SMF
+  Boolean prev_cts, prev_dsr, prev_ri, prev_dcd;
+  Boolean gen_int = 0;
+
+  BX_DEBUG(("write: 0x%x <- %d",address,value));
 
   /* SERIAL PORT 1 */
 
   if (io_len > 1)
-    bx_panic("serial: io write to address %08x len=%u\n",
-             (unsigned) address, (unsigned) io_len);
+    BX_PANIC(("io write to address %08x len=%u",
+             (unsigned) address, (unsigned) io_len));
 
-  if (bx_dbg.serial)
-    bx_printf("serial: write to address: 0x%x = 0x%x\n",
-	      (unsigned) address, (unsigned) value);
+  BX_DEBUG(("write to address: 0x%x = 0x%x",
+	      (unsigned) address, (unsigned) value));
 
   switch (address) {
     case 0x03F8: /* transmit buffer, or divisor latch LSB if DLAB set */
-      if (BX_SER_THIS s[0].line_cntl.dlab) {	
+      if (BX_SER_THIS s[0].line_cntl.dlab) {
 	BX_SER_THIS s[0].divisor_lsb = value;
 
 	if (value != 0) {
@@ -423,24 +496,26 @@ bx_serial_c::write(Bit32u address, Bit32u value, unsigned io_len)
 #endif // USE_RAW_SERIAL
 	}
       } else {
-	if (BX_SER_THIS s[0].tx_empty) {
-	  BX_SER_THIS s[0].tx_empty = 0;
-	  BX_SER_THIS s[0].line_status.txtransm_empty = 0;
-	  BX_SER_THIS s[0].line_status.txhold_empty = 0;
-	  /* If there are no more ints pending, clear the irq */
-	  if ((BX_SER_THIS s[0].tx_interrupt == 1) &&
-	      (BX_SER_THIS s[0].rx_interrupt == 0)) {
-	    pluginUntriggerIRQ(4);
-	  }
-	  BX_SER_THIS s[0].tx_interrupt = 0;
-	  BX_SER_THIS s[0].txbuffer = value;
-	  pluginActivateTimer(BX_SER_THIS s[0].tx_timer_index
-,
-		 (int) (1000000.0 / (BX_SER_THIS s[0].baudrate / 8)),
-				      0); /* not continuous */
-	} else {
-	  bx_printf("serial: write to tx hold register when not empty\n");
-	}
+        if (BX_SER_THIS s[0].tx_empty) {
+          BX_SER_THIS s[0].tx_empty = 0;
+          BX_SER_THIS s[0].line_status.txtransm_empty = 0;
+          BX_SER_THIS s[0].line_status.txhold_empty = 0;
+          /* If there are no more ints pending, clear the irq */
+          if ((BX_SER_THIS s[0].rx_interrupt == 0) &&
+              (BX_SER_THIS s[0].ls_interrupt == 0) &&
+              (BX_SER_THIS s[0].ms_interrupt == 0)) {
+            pluginLowerIRQ(4);
+          }
+          BX_SER_THIS s[0].tx_interrupt = 0;
+          BX_SER_THIS s[0].tx_ipending = 0;
+          Bit8u bitmask = 0xff >> (3 - BX_SER_THIS s[0].line_cntl.wordlen_sel);
+          BX_SER_THIS s[0].txbuffer = value & bitmask;
+          pluginActivateTimer(BX_SER_THIS s[0].tx_timer_index,
+                              (int) (1000000.0 / (BX_SER_THIS s[0].baudrate / 8)),
+                              0); /* not continuous */
+        } else {
+          BX_ERROR(("write to tx hold register when not empty"));
+        }
       }
       break;
 
@@ -461,6 +536,32 @@ bx_serial_c::write(Bit32u address, Bit32u value, unsigned io_len)
 	BX_SER_THIS s[0].int_enable.txhold_enable  = (value & 0x02) >> 1;
 	BX_SER_THIS s[0].int_enable.rxlstat_enable = (value & 0x04) >> 2;
 	BX_SER_THIS s[0].int_enable.modstat_enable = (value & 0x08) >> 3;
+        if ((BX_SER_THIS s[0].ms_ipending == 1) &&
+            (BX_SER_THIS s[0].int_enable.modstat_enable == 1)) {
+          BX_SER_THIS s[0].ms_interrupt = 1;
+          BX_SER_THIS s[0].ms_ipending = 0;
+          gen_int = 1;
+        }
+        if ((BX_SER_THIS s[0].tx_ipending == 1) &&
+            (BX_SER_THIS s[0].int_enable.txhold_enable == 1)) {
+          BX_SER_THIS s[0].tx_interrupt = 1;
+          BX_SER_THIS s[0].tx_ipending = 0;
+          gen_int = 1;
+        }
+        if ((BX_SER_THIS s[0].rx_ipending == 1) &&
+            (BX_SER_THIS s[0].int_enable.rxdata_enable == 1)) {
+          BX_SER_THIS s[0].rx_interrupt = 1;
+          BX_SER_THIS s[0].rx_ipending = 0;
+          gen_int = 1;
+        if ((BX_SER_THIS s[0].ls_ipending == 1) &&
+            (BX_SER_THIS s[0].int_enable.rxlstat_enable == 1)) {
+          BX_SER_THIS s[0].ls_interrupt = 1;
+          BX_SER_THIS s[0].ls_ipending = 0;
+          gen_int = 1;
+        }
+        }
+        if (gen_int == 1)
+          pluginRaiseIRQ(4);
       }
       break;
 
@@ -487,7 +588,7 @@ bx_serial_c::write(Bit32u address, Bit32u value, unsigned io_len)
 	  BX_SER_THIS s[0].line_cntl.stick_parity != (value & 0x20) >> 5) {
 	    if (((value & 0x20) >> 5) &&
 		((value & 0x8) >> 3))
-		  bx_panic("[serial] sticky parity set and parity enabled");
+		  BX_PANIC(("sticky parity set and parity enabled"));
 	    BX_SER_THIS raw->set_parity_mode(((value & 0x8) >> 3),
 					     ((value & 0x10) >> 4) ? P_EVEN : P_ODD);
       }
@@ -515,13 +616,11 @@ bx_serial_c::write(Bit32u address, Bit32u value, unsigned io_len)
 		      (int) (1000000.0 / (BX_SER_THIS s[0].baudrate / 4)),
 				      0); /* not continuous */
 	}
-	if (bx_dbg.serial) {
-	  bx_printf("serial: baud rate set - %d\n", BX_SER_THIS s[0].baudrate);
-	}
+	BX_DEBUG(("baud rate set - %d", BX_SER_THIS s[0].baudrate));
       }
       BX_SER_THIS s[0].line_cntl.dlab = (value & 0x80) >> 7;
       break;
- 	
+
     case 0x03FC: /* MODEM control register */
       if ((value & 0x01) == 0) {
 #if USE_RAW_SERIAL
@@ -536,27 +635,50 @@ bx_serial_c::write(Bit32u address, Bit32u value, unsigned io_len)
       BX_SER_THIS s[0].modem_cntl.local_loopback = (value & 0x10) >> 4;
 
       if (BX_SER_THIS s[0].modem_cntl.local_loopback) {
-	BX_SER_THIS s[0].modem_status.cts = BX_SER_THIS s[0].modem_cntl.rts;
-	BX_SER_THIS s[0].modem_status.dsr = BX_SER_THIS s[0].modem_cntl.dtr;
-	BX_SER_THIS s[0].modem_status.ri  = BX_SER_THIS s[0].modem_cntl.out1;
-	BX_SER_THIS s[0].modem_status.dcd = BX_SER_THIS s[0].modem_cntl.out2;
+        prev_cts = BX_SER_THIS s[0].modem_status.cts;
+        prev_dsr = BX_SER_THIS s[0].modem_status.dsr;
+        prev_ri  = BX_SER_THIS s[0].modem_status.ri;
+        prev_dcd = BX_SER_THIS s[0].modem_status.dcd;
+        BX_SER_THIS s[0].modem_status.cts = BX_SER_THIS s[0].modem_cntl.rts;
+        BX_SER_THIS s[0].modem_status.dsr = BX_SER_THIS s[0].modem_cntl.dtr;
+        BX_SER_THIS s[0].modem_status.ri  = BX_SER_THIS s[0].modem_cntl.out1;
+        BX_SER_THIS s[0].modem_status.dcd = BX_SER_THIS s[0].modem_cntl.out2;
+        if (BX_SER_THIS s[0].modem_status.cts != prev_cts) {
+          BX_SER_THIS s[0].modem_status.delta_cts = 1;
+          BX_SER_THIS s[0].ms_ipending = 1;
+        }
+        if (BX_SER_THIS s[0].modem_status.dsr != prev_dsr) {
+          BX_SER_THIS s[0].modem_status.delta_dsr = 1;
+          BX_SER_THIS s[0].ms_ipending = 1;
+        }
+        if (BX_SER_THIS s[0].modem_status.ri != prev_ri)
+          BX_SER_THIS s[0].ms_ipending = 1;
+        if ((BX_SER_THIS s[0].modem_status.ri == 0) && (prev_ri == 1))
+          BX_SER_THIS s[0].modem_status.ri_trailedge = 1;
+        if (BX_SER_THIS s[0].modem_status.dcd != prev_dcd) {
+          BX_SER_THIS s[0].modem_status.delta_dcd = 1;
+          BX_SER_THIS s[0].ms_ipending = 1;
+        }
+        if ((BX_SER_THIS s[0].ms_ipending == 1) &&
+            (BX_SER_THIS s[0].int_enable.modstat_enable == 1)) {
+          BX_SER_THIS s[0].ms_interrupt = 1;
+          BX_SER_THIS s[0].ms_ipending = 0;
+        }
       } else {
-	/* set these to 0 for the time being */
-	BX_SER_THIS s[0].modem_status.cts = 0;
-	BX_SER_THIS s[0].modem_status.dsr = 0;
-	BX_SER_THIS s[0].modem_status.ri  = 0;
-	BX_SER_THIS s[0].modem_status.dcd = 0;
+        /* set these to 0 for the time being */
+        BX_SER_THIS s[0].modem_status.cts = 0;
+        BX_SER_THIS s[0].modem_status.dsr = 0;
+        BX_SER_THIS s[0].modem_status.ri  = 0;
+        BX_SER_THIS s[0].modem_status.dcd = 0;
       }
       break;
 
     case 0x03FD: /* Line status register */
-      /* XXX ignore ?  */
-      bx_panic("serial: write to line status register\n");
+      BX_ERROR(("write to line status register ignored"));
       break;
 
     case 0x03FE: /* MODEM status register */
-      /* XXX ignore ?  */
-      bx_panic("serial: write to MODEM status register\n");
+      BX_ERROR(("write to MODEM status register ignored"));
       break;
 
     case 0x03FF: /* scratch register */
@@ -564,8 +686,8 @@ bx_serial_c::write(Bit32u address, Bit32u value, unsigned io_len)
       break;
 
     default:
-      bx_panic("unsupported serial io write to address=0x%x, value = 0x%x!\n",
-        (unsigned) address, (unsigned) value);
+      BX_PANIC(("unsupported io write to address=0x%x, value = 0x%x!",
+        (unsigned) address, (unsigned) value));
       break;
   }
 }
@@ -590,30 +712,48 @@ bx_serial_c::tx_timer(void)
   if (BX_SER_THIS s[0].int_enable.txhold_enable) {
     gen_int = 1;
     BX_SER_THIS s[0].tx_interrupt = 1;
+  } else {
+    BX_SER_THIS s[0].tx_ipending = 1;
   }
 
   if (BX_SER_THIS s[0].modem_cntl.local_loopback) {
+    if (BX_SER_THIS s[0].rx_empty == 0)
+      BX_SER_THIS s[0].line_status.overrun_error = 1;
     BX_SER_THIS s[0].rxbuffer = BX_SER_THIS s[0].txbuffer;
     BX_SER_THIS s[0].line_status.rxdata_ready = 1;
     BX_SER_THIS s[0].rx_empty = 0;
-    if (BX_SER_THIS s[0].int_enable.rxdata_enable) {
-      gen_int = 1;
-      BX_SER_THIS s[0].rx_interrupt = 1;
+    if (BX_SER_THIS s[0].modem_cntl.out2) {
+      if (BX_SER_THIS s[0].int_enable.rxdata_enable) {
+        gen_int = 1;
+        BX_SER_THIS s[0].rx_interrupt = 1;
+      } else {
+        BX_SER_THIS s[0].rx_ipending = 1;
+      }
+      if (BX_SER_THIS s[0].int_enable.rxlstat_enable) {
+        gen_int = 1;
+        BX_SER_THIS s[0].ls_interrupt = 1;
+      } else {
+        BX_SER_THIS s[0].ls_ipending = 1;
+      }
     }
   } else {
 #if defined (USE_TTY_HACK)
     tty(tty_id, 0, & BX_SER_THIS s[0].txbuffer);
 #elif USE_RAW_SERIAL
     if (!BX_SER_THIS raw->ready_transmit())
-	  bx_panic("[serial] Not ready to transmit");
+	  BX_PANIC(("Not ready to transmit"));
     BX_SER_THIS raw->transmit(BX_SER_THIS s[0].txbuffer);
-#elif 0
-    write(0, (bx_ptr_t) & BX_SER_THIS s[0].txbuffer, 1);
+#endif
+#if defined(SERIAL_ENABLE)
+    { char *s = (char *)(BX_SER_THIS s[0].txbuffer);
+	BX_DEBUG(("write: '%c'",(bx_ptr_t) & s));
+	}
+    write(tty_id, (bx_ptr_t) & BX_SER_THIS s[0].txbuffer, 1);
 #endif
   }
 
   if (gen_int) {
-    pluginTriggerIRQ(4);
+    pluginRaiseIRQ(4);
   }
 }
 
@@ -630,26 +770,27 @@ bx_serial_c::rx_timer_handler(void *this_ptr)
 void
 bx_serial_c::rx_timer(void)
 {
-  struct timeval tval;
 #if BX_HAVE_SELECT
 #if BX_WITH_BEOS == 0
+  struct timeval tval;
   fd_set fds;
 #endif
 #endif
   int bdrate = BX_SER_THIS s[0].baudrate / 8;
-  unsigned char chbuf;
-
-  tval.tv_sec  = 0;
-  tval.tv_usec = 0;
+  unsigned char chbuf = 0;
 
 #if BX_HAVE_SELECT
 #if BX_WITH_BEOS == 0
+  tval.tv_sec  = 0;
+  tval.tv_usec = 0;
+
 // MacOS: I'm not sure what to do with this, since I don't know
 // what an fd_set is or what FD_SET() or select() do. They aren't
 // declared in the CodeWarrior standard library headers. I'm just
 // leaving it commented out for the moment.
 
-  FD_SET(0, &fds);
+  FD_ZERO(&fds);
+  FD_SET(tty_id, &fds);
 
   if (BX_SER_THIS s[0].line_status.rxdata_ready == 0) {
 #if defined (USE_TTY_HACK)
@@ -661,36 +802,39 @@ bx_serial_c::rx_timer(void)
     if ((rdy = BX_SER_THIS raw->ready_receive())) {
       data = BX_SER_THIS raw->receive();
       if (data == C_BREAK) {
-	    bx_printf("[serial] Got BREAK\n");
+	    BX_DEBUG(("got BREAK"));
 	    BX_SER_THIS s[0].line_status.break_int = 1;
 	    rdy = 0;
       }
     }
     if (rdy) {
 	  chbuf = data;
-#elif 0
-    if (select(1, &fds, NULL, NULL, &tval) == 1) {
-      (void) read(0, &chbuf, 1);
+#elif defined(SERIAL_ENABLE)
+    if (select(tty_id + 1, &fds, NULL, NULL, &tval) == 1) {
+      (void) read(tty_id, &chbuf, 1);
+	  BX_DEBUG(("read: '%c'",chbuf));
 #else
     if (0) {
 #endif
       if (!BX_SER_THIS s[0].modem_cntl.local_loopback) {
 
-	BX_SER_THIS s[0].rxbuffer = chbuf;
-	BX_SER_THIS s[0].line_status.rxdata_ready = 1;
-	BX_SER_THIS s[0].rx_empty = 0;
-	if (BX_SER_THIS s[0].int_enable.rxdata_enable) {
-	  BX_SER_THIS s[0].rx_interrupt = 1;
-	  pluginTriggerIRQ(4);
-	}
+        BX_SER_THIS s[0].rxbuffer = chbuf;
+        BX_SER_THIS s[0].line_status.rxdata_ready = 1;
+        BX_SER_THIS s[0].rx_empty = 0;
+        if (BX_SER_THIS s[0].int_enable.rxdata_enable) {
+          BX_SER_THIS s[0].rx_interrupt = 1;
+          pluginRaiseIRQ(4);
+        } else {
+          BX_SER_THIS s[0].rx_ipending = 1;
+        }
       }
     } else {
-      bdrate = (int) (1000000.0 / 100); // Poll frequency is 100ms
+      bdrate = (int) (1000000.0 / 100000); // Poll frequency is 100ms
     }
   } else {
     // Poll at 4x baud rate to see if the next-char can
     // be read
-    bdrate *= 4; 
+    bdrate *= 4;
   }
 #endif
 #endif
